#!/bin/bash
#---------------------------------------------------------------------------------------------------------
# FOXCLONE BUILD SCRIPT DEC 2020 - BUILDS AN ISO
#
# Useful sources:
# https://help.ubuntu.com/community/LiveCDCustomizationFromScratch
# https://superuser.com/questions/165116/mount-dev-proc-sys-in-a-chroot-environment
# https://askubuntu.com/questions/551195/scripting-chroot-how-to
#
# modified 18 Jan 2022 for foxclone50
# lightdm instead of slim
# ubuntu font not noto
# later palemoon
#---------------------------------------------------------------------------------------------------------
#set -x

#---------------------------------------------------------------------------------------------------------
# SETUP - VARIABLES, FILES & FOLDERS
# where variables are exported it is to make them available in the chroot
# environment
#---------------------------------------------------------------------------------------------------------
clear #the screen
today=$(date +'%a %d %b %Y')
SECONDS=0      #set counter to determine elapsed time for the script

buildver=52-bionic #xx-yy = foxclone version xx, iso version yy used in the iso filename
arch=amd64     #foxclone only supports 64 bit

# get the local user, because the script runs as root can't use $USER. EUID will return 0  and USER=root
# when running as root and RUID (Real User ID) returned empty on linux mint.
# Assumption is that the local user is the first user created, so get the min UID for the system and 
# see if a name exists for it. This is later used to chown the bootstrap archive file from root back to 
# local user so that backup utilities such as backintime do not throw an error. Note - the build folder
# still needs to be excluded from backintime or equivalent.
minUID=`grep ^UID_MIN /etc/login.defs | cut -f2 -d\ `
localuser=`getent passwd 1000 | cut -d: -f1`

# BUILD - these two need to be changed as a pair
export codename=bionic            #e.g. xenial, focal for ubuntu or buster for debian (not tested)
export codenumber="18.04"        #to match the codename.

export kernel=linux-generic-hwe-18.04  #the explicit package to install, could be
                                       #linux-generic-hwe-$codenumber, but this way any
                                       #kernel can be installed in any package. 
                                       #linux-generic-hwe-18.04 for bionic & kernel 5.4
                                       #linux-oem-20.04 for focal & kernel 5.14
initrd_img=initrd.img-*-generic        #initrd.img-5.14.0-1020-oem for 5.14 kernel
vmlinuz_img=vmlinuz-*-generic          #vmlinuz-5.14.0-1020-oem for 5.14 kernel



export foxlocale=en_GB          #sets the locale

#FOLDERS & FILES
foxfolder=$PWD               #current folder where the script is executed
foxbuild=build               #the image will be built in foxfolder/foxbuild
image=image                  #image folder for boot files, it will be in foxfolder/foxbuild
newroot=chroot               #where the linux system will be built, it will be in foxfolder/foxbuild
plymouth=ubuntu-spinner-logo #folder holding the files for the custom plymouth theme, assumed to be in 
                             #foxfolder
isofilename=$PWD/foxclone$buildver.iso #the new iso will be created with this filename in foxfolder    
#cache file - this contains the initial bootstrap, if present it saves some time
cachefile=$foxfolder/$codename-base.tar.gz 

#deb files for installation in the iso, need to be in foxfolder
export browserdeb=palemoon_28.13.0-2_amd64.deb #note, tried lastest palemoon
                                               #palemoon_29.4.3-1.1.gtk2_amd64.deb 
                                               #doesnt work unmet dependencies
export foxclonedeb=foxclone52_amd64.deb

echo Current folder= $foxfolder
echo Build folder= $foxbuild
echo Linux variant= $codename
echo arch= $arch
echo Image folder= $image
echo Chroot folder= $newroot
echo Browser= $browserdeb
echo Foxclone= $foxclonedeb
echo "cache file ="$cachefile
echo "kernel to install ="$kernel

# Setup text colour output for terminal
export red=`tput setaf 1`
export grn=`tput setaf 2`
export yel=`tput setaf 226`
export def=`tput sgr0` #default or reset

#check if running as root
if [ "$EUID" -ne 0 ]; then 
    echo $red"The foxclone build script must be run with sudo."$def
    exit
fi

#---------------------------------------------------------------------------------------------------------
# STARTUP - BANNER & INITIAL CHECKS
#---------------------------------------------------------------------------------------------------------

# Banner
echo $yel------------------
echo FOXCLONE ISO BUILD
echo Version $buildver
echo $today
echo $yel------------------$def

# check if root
if [ "$EUID" -ne 0 ]; then
	echo $red 
	echo "the foxbuild script must be run as root, exiting..." $def
	exit 1
fi

# check if the deb files are present
if [ ! -f $browserdeb ]; then
    echo $red
    echo "The deb file to install the browser cannot be found"
    echo $browserdeb" file is not present in "$foxfolder
    echo "exiting..."$def
    echo
    exit 1
fi
if [ ! -f $foxclonedeb ]; then
    echo $red
    echo "The deb file to install foxclone cannot be found"
    echo $foxclonedeb" file is not present in "$foxfolder
    echo "exiting..."$def
    echo
    exit 1
fi

# check if the required packages for the script are installed. 
for i in pigz debootstrap xorriso grub-mkimage ; do
    check=`whereis $i | grep bin`
    if [ ${#check} -eq 0 ]; then #package not installed
        echo $red"package $i is not installed. It is required for the build script."
        read -p "Install (Y/n)?$def" answer 
        if [ -z "$answer" ]; then
            answer=Y
        fi
        case $answer in
            [Yy]* ) if [ "$i" = "grub-mkimage" ]; then
                        apt-get install --yes --no-install-recommends grub-common
                    else
                        apt-get install --yes --no-install-recommends $i
                    fi;;
            [Nn]* ) echo $red"Exiting ..."$def
                    exit 1;;
        esac  
    fi 
done 
echo

#---------------------------------------------------------------------------------------------------------
# FUNCTION INSTALL_SH
# Runs inside chroot to install all the packages to run the system.
#---------------------------------------------------------------------------------------------------------
install_sh() {
    # The thing about chroots and /proc, /sys and /dev/pts is that these three filesystems are provided 
    # by the kernel, so they remain the same whether you mount within the chroot or from without. 
    echo $grn"mounting proc, sys and dev/pts inside chroot ..."$def
    mount none -t proc /proc
    mount none -t sysfs /sys
    mount none -t devpts /dev/pts
    echo
    
    # turn off all apt prompts
    export DEBIAN_FRONTEND=noninteractive
    
    # set home and default language
    export HOME=/root
    export LC_ALL=C
    
    # having added to the repos need to update before installing anything
    echo $grn"running apt update ..."$def
    apt update
    #echo $grn"running apt upgrade ..."$def
    apt-get upgrade --yes
    echo
    
    # D-Bus is a message bus system that provides an easy way for inter-process communication
    # may or may not be installed by default, install anyway
    echo $grn"installing dbus ..."$def
    apt-get install --yes --no-install-recommends dbus
    echo
    
    # automatically installed in bionic, not in focal & caused problems as no update-initramfs
    echo $grn"installing initramfs-tools ..."$def
    apt-get install --yes --no-install-recommends initramfs-tools
    echo
    
    # Install kernel
    echo $grn"installing $kernel ..."$def
    apt-get install --yes --no-install-recommends $kernel
    if [[ $? -ne 0 ]]; then
        echo $red"ERROR: failed to install $kernel ... "$def
        exit 1
    fi
    echo
    
    # Install system packages
    # casper - run a "live" preinstalled system from read-only media
    # lupin-casper - add support for loop-mount installations to casper
    # cifs-utils - provides a means for mounting SMB/CIFS shares on a Linux system
    # gvfs - GNOME Virtual file system userspace virtual filesystem 
    # gvfs-backends - installed because of problems with thunar on other drives leave in with pcmanfm
    # nmap - network mapper
    # network-manager-gnome - network connectivity out of the box
    # smbclient - client to access SMB/CIFS resources on servers
    # time - measuring CPU resource usage
    # xdotool - simulate keyboard input, needed for right click in thunar to toggle hidden files
    # remove xdotool - not using thunar any longer 
    # xinit - not installed in minbase, first process started during booting, ancestor of all processes
    # gnome keyring added so that connecting to wifi it prompts for pwd
    #build for ubuntu 22.04 - lupin-casper not found
    #https://bugs.launchpad.net/ubuntu/+source/lupin/+bug/1960457
    #seems only needed for livecd-rootfs does try and install lupin-casper when 
    #building Ubuntu images but that is cruft but may need to be removed from livecd-rootfs 
    #before the package is removed - this seems to have been fixed, do I need it.
    if [ $codename != 'jammy' ]; then
        apt-get install --yes --no-install-recommends lupin-casper
    fi
    
    echo $grn"installing system packages ..."$def
    apt-get install --yes --no-install-recommends casper \
                                              cifs-utils \
                                              gvfs \
                                              gvfs-backends \
                                              nmap \
                                              network-manager-gnome \
                                              gnome-keyring \
                                              smbclient \
                                              time \
                                              xinit 
    if [[ $? -ne 0 ]]; then
        echo $red"ERROR: failed to install applications ... "$def
        exit 1
    fi
    echo      


    # Install window support
    # openbox - window manager
    # xserver - X window system display server
    # pcmanfm - used as desktop manager
    # yad - used graphical support to startup/shutdown scripts
    # gtk2-engines-pixbuf - pixbuf-based theme for GTK+ 2.x
    # gsettings-desktop-schemas - settings shared by various components of a desktop
    # notify-osd - on screen notifications
    echo $grn"installing GUI ..."$def
    # bionic
    if [ $codename == 'bionic' ]; then
        apt-get install --yes --no-install-recommends xserver-xorg-hwe-$codenumber \
                                                      xserver-xorg-video-vesa-hwe-$codenumber
        if [[ $? -ne 0 ]]; then
            echo $red"ERROR: failed to install $codename xserver packages ... "$def
            exit 1
        fi
    fi  
    
    # for everything else, works with focal
    if [ $codename != 'bionic' ]; then
        apt-get install --yes --no-install-recommends xserver-xorg \
                                                      xserver-xorg-video-all \
                                                      xserver-xorg-video-intel \
                                                      xserver-xorg-video-qxl \
                                                      xserver-xorg-video-mga 
        if [[ $? -ne 0 ]]; then
            echo $red"ERROR: failed to install $codename xserver packages ... "$def
            exit 1
        fi
    fi
    
    #thunderbolt not in bionic 5.4 kernel
    echo $grn"installing thunderbolt ..."$def
    #apt-get install --yes --no-install-recommends linux-modules-extra-$(uname -r)
    #didnt work
    #if [ $codename == 'bionic' ]; then not needed already installed
    #    apt-get install --yes --no-install-recommends linux-modules-extra-5.4.0-150-generic  
    #fi 
    apt-get install --yes --no-install-recommends bolt
    apt-get install --yes --no-install-recommends thunderbolt-tools
    
    # common
    # no ttf-ubuntu-font-family in 22.04 deleted, but where is it now?
    apt-get install --yes --no-install-recommends openbox \
                                              nouveau-firmware \
                                              x11-xserver-utils \
                                              pcmanfm \
                                              gnome-icon-theme \
                                              gnome-brave-icon-theme \
                                              dmz-cursor-theme \
                                              lxmenu-data \
                                              lxpanel \
                                              yad \
                                              gtk2-engines-pixbuf \
                                              gsettings-desktop-schemas \
                                              notify-osd \
                                              notify-osd-icons 
    if [[ $? -ne 0 ]]; then
        echo $red"ERROR: failed to install GUI ... "$def
        exit 1
    fi
    echo
    
    # Install filesystem support and utilities
    echo $grn"installing filesystem support ..."$def
    apt-get install --yes --no-install-recommends util-linux \
                                              dosfstools \
                                              ntfs-3g \
                                              jfsutils \
                                              fuse \
                                              gvfs-fuse \
                                              exfat-fuse \
                                              btrfs-progs \
                                              udisks2-btrfs \
                                              f2fs-tools \
                                              lvm2 \
                                              xfsdump \
                                              xfsprogs \
                                              udftools 
    #no exfat-utils in ubuntu 22.04 replaced by exfatprogs
    if [ $codename == 'jammy' ]; then
        apt-get install --yes --no-install-recommends exfatprogs
    else
        apt-get install --yes --no-install-recommends exfat-utils
    fi

    if [[ $? -ne 0 ]]; then
        echo $red"ERROR: failed to install file support packages ... "$def
        exit 1
    fi
    # bionic only, not present in focal
    if [ $codename == 'bionic' ]; then
        apt-get install --yes --no-install-recommends btrfs-tools
    fi
    if [[ $? -ne 0 ]]; then
        echo $red"ERROR: failed to install $codename btrfs tools ... "$def
        exit 1
    fi
    
    echo

    # Install plymouth for GUI startup and shutdown
    echo $grn"installing plymouth for startup/shudown ..."$def
    apt-get install --yes --no-install-recommends plymouth-x11 plymouth-label 
    if [[ $? -ne 0 ]]; then
        echo $red"ERROR: failed to install plymouth ... "$def
        exit 1
    fi
    echo
    
    # Install GUI applications
    #is lshw installed to support the gui app - NO, is it pulled in? - NO so is it needed?
    # arandr - set display resolution
    # baobab - disk usage analyser
    # gnome-disk-utility - disks, manage drives/partitions
    # gnome-terminal - a terminal emulator, one of many
    # gnome-screenshot - screenshot utility
    # gparted - standard linux partition editor
    # gpicview - view image files
    # lshw-gtk - GUI front end for lshw, showing hardware details
    # mousepad - text editor
    # testdisk - not GUI but put on the menu, file recovery along with photorec - where is that?
    # smartmontools & gsmartcontrol - disk health
    # lm-sensors - needed for CPU temp displayed in lxpanel
    # thunar - file manager
    echo $grn"installing GUI applications ..."$def
    apt-get install --yes --no-install-recommends arandr \
                                              baobab \
                                              gnome-disk-utility \
                                              gnome-terminal \
                                              gnome-screenshot \
                                              gparted \
                                              gpicview \
                                              lshw-gtk \
                                              mousepad \
                                              testdisk \
                                              smartmontools \
                                              gsmartcontrol \
                                              lm-sensors 
    if [[ $? -ne 0 ]]; then
        echo $red"ERROR: failed t--no-install-recommendso install applications ... "$def
        exit 1
    fi
    echo   
    echo $grn"installing grub ..."$def    
    #boot packages
    apt-get install --yes --no-install-recommends efibootmgr \
                                              shim-signed \
                                              grub-efi-amd64-signed \
                                              grub-efi-amd64-bin \
                                              grub-efi-ia32-bin \
                                              grub-pc-bin \
                                              grub2-common
    #troubleshooting
    apt-get install --yes --no-install-recommends inxi

    if [[ $? -ne 0 ]]; then
        echo $red"ERROR: failed to install boot utilities ... "$def
        exit 1
    fi
          
    # Install browser
    echo $grn"Installing browser ..."$def
    apt-get install /$browserdeb -y --no-install-recommends
    if [[ $? -ne 0 ]]; then
        echo $red"ERROR: failed to install $browserdeb ... "$def
        exit 1
    fi
    # finished with it, delete the deb
    rm /$browserdeb
    echo      
    
    #Install foxclone 
    echo $grn"Installing foxclone ..."$def
    apt-get install /$foxclonedeb -y --no-install-recommends
    if [[ $? -ne 0 ]]; then
        echo $red"ERROR: failed to install $foxclonedeb ... "$def
        exit 1
    fi
    # finished with it, delete the deb
    rm /$foxclonedeb
    
    # Remove partclone files not used - filesystems not supported in foxclone
    # hfs = apple, reiser = old and not a nice man, it's a principle!
    rm /usr/sbin/partclone.hfs+
    rm /usr/sbin/partclone.hfsp
    rm /usr/sbin/partclone.hfsplus
    rm /usr/sbin/partclone.reiser4
    rm /usr/sbin/partclone.restore
    # Remove fonts not used, noto installed by foxclone deb - only want the regular font 
    rm /usr/share/fonts/truetype/ubuntu/Ubuntu-C.ttf
    rm /usr/share/fonts/truetype/ubuntu/Ubuntu-L.ttf
    rm /usr/share/fonts/truetype/ubuntu/Ubuntu-LI.ttf
    rm /usr/share/fonts/truetype/ubuntu/Ubuntu-M.ttf
    rm /usr/share/fonts/truetype/ubuntu/Ubuntu-MI.ttf
    echo 
     
    # fonts-ubuntu removal with apt takes out plymouth-label, plymouth shutdown screen doesn't display
    # message and hangs (but enter does shutdown). Delete fonts manually.
    # Dejavu fonts - pulled in by the kernel and some of the system packages
    # apt remove insists on installing ttf-bitstream-vera and some symbols do not display in foxclone
    # keep DejaVuSansMono.ttf and delete the rest. Could have used aliases, too difficult
    # update the fonts cache 
    fc-cache -f -v
    
    # Install the packages that have custom configuration files managed in the
    # source tree, and use non-interactive conflict resolution to choose those
    # configuration files, not package maintainer's versions.  The non-interactive
    # conflict resolutions means its the responsibility of the of the build script
    # maintainer to periodically ensure the version-controlled configuration file
    # hasn't does not become too old and out-of-date compared to the package
    # maintainer's version.
    
    #install display manager, lightdm
    echo $grn"Installing lightdm ..."$def
    apt-get -o Dpkg::Options::="--force-confdef" -o Dpkg::Options::="--force-confold" install --yes --no-install-recommends lightdm
    if [[ $? -ne 0 ]]; then
        echo $red"ERROR: failed to install lightdm ... "$def
        exit 1
    fi

    # Create empty config file for the network-manager service to manage all
    # network devices. Without this running the iso in VB will not connect to the wired network
    # source https://askubuntu.com/a/893614/394984
    touch /etc/NetworkManager/conf.d/10-globally-managed-devices.conf

    # Set plymouth theme    
    echo $grn"set plymouth to use the spinner theme ..."$def
    update-alternatives --install /usr/share/plymouth/themes/default.plymouth default.plymouth /usr/share/plymouth/themes/ubuntu-spinner-logo/ubuntu-spinner-logo.plymouth 100
    update-alternatives --set default.plymouth /usr/share/plymouth/themes/ubuntu-spinner-logo/ubuntu-spinner-logo.plymouth

    #-u     This mode updates an existing initramfs.
    echo $grn"now update initramfs..."$def
    update-initramfs -u
    
    # Set locale and delete unwanted files (think it was 70MB), details here:
    # https://www.linuxquestions.org/questions/blog/bittner-195120/remove-unwanted-locales-on-ubuntu-debian-3281/
    echo $grn"set up locales and delete unwanted files"$def
    echo $grn"creating /var/lib/locales/supported.d"$def
    mkdir -p /var/lib/locales/supported.d
    cat << EOF > /var/lib/locales/supported.d/en
$foxlocale.UTF-8 UTF-8
EOF
    echo $grn"now update /etc/default/locale"$def
    cat << EOF > /etc/default/locale
#  File generated by update-locale
LANG=$foxlocale.UTF-8
LANGUAGE=$foxlocale:en
EOF
    echo $grn"deleting files in  /usr/lib/locale/*"$def
    rm -rf /usr/lib/locale/*
    cat /etc/default/locale
    locale-gen
    # now delete /usr/share/locale
    rm -rf /usr/share/locale/*
    
    echo $grn"run apt-get autoremove..."$def
    apt-get autoremove
    
    # Delete unwanted files to reduce iso size    
    rm /var/lib/dbus/machine-id
    rm /sbin/initctl
    dpkg-divert --rename --remove /sbin/initctl

    rm -r /usr/share/themes/Artwiz-boxed
    rm -r /usr/share/themes/Bear2
    rm -r /usr/share/themes/Breeze-ob
    rm -r /usr/share/themes/Clearlooks
    rm -r /usr/share/themes/Clearlooks-3.4
    rm -r /usr/share/themes/Clearlooks-Olive
    rm -r /usr/share/themes/Default
    rm -r /usr/share/themes/Emacs
    rm -r /usr/share/themes/Mikachu
    rm -r /usr/share/themes/Natura
    rm -r /usr/share/themes/Nightmare
    rm -r /usr/share/themes/Nightmare-01
    rm -r /usr/share/themes/Nightmare-02
    rm -r /usr/share/themes/Nightmare-03
    rm -r /usr/share/themes/Onyx
    rm -r /usr/share/themes/Onyx-Citrus
    rm -r /usr/share/themes/Orang
    rm -r /usr/share/themes/Raleigh
    rm -r /usr/share/themes/Syscrash

    rm -r /usr/share/icons/Humanity
    rm -r /usr/share/icons/DMZ-Black
    rm -r /usr/share/icons/Humanity-Dark
    rm -r /usr/share/icons/locolor
    rm -r /usr/share/icons/ubuntu-mono-dark
    rm -r /usr/share/icons/ubuntu-mono-light
    rm /usr/share/icons/*/icon-theme.cache
    
    rm /usr/share/applications/lxde-x-www-browser.desktop
    rm /usr/share/applications/pcmanfm-desktop-pref.desktop
    rm /usr/share/applications/vim.desktop
    rm /usr/share/applications/pcmanfm.desktop #replaced and renamed
    rm /usr/share/applications/lxde-x-terminal-emulator.desktop
    rm /usr/share/applications/yad-icon-browser.desktop

    rm -r /usr/bin/{rpcclient,smbcacls,smbclient,smbcquotas,smbget,smbspool,smbtar}
    rm -rf /usr/share/doc
    rm -rf /usr/share/man
    rm -rf /etc/network/if-up.d/ntpdate    

    rm -rf /var/cache/apt/* #contains all the debs downloaded
    rm -rf /var/lib/apt/lists/????????*

    rm /etc/resolv.conf

    # From `man apt-get`: "clears out the local repository of retrieved package
    # files. It removes everything but the lock file from /var/cache/apt/archives/
    # and /var/cache/apt/archives/partial/."
    # didnt have this first time around with new iso
    apt-get clean
    
    # remove packages installed but not needed, using dpkg to remove but leave the
    # dependenices installed.

    dpkg -r --force-depends gdisk # 860.0 kB	GPT fdisk text-mode partitioning tool
    dpkg -r --force-depends libgweather-3-16 # 213.0 kB GWeather shared library
    dpkg -r --force-depends libvorbisenc2 # 704.0 kB encoder library for Vorbis General Audio Compression Codec
    dpkg -r --force-depends sound-theme-freedesktop # 537.0 kB	freedesktop.org sound theme
    dpkg -r --force-depends vim-tiny # 1,531.0 kB	Vi IMproved - enhanced vi editor - compact version
    if [ $codename == "focal" ]; then
        dpkg -r --force-depends evolution-data-server # 2,752.0 kB evolution database backend server
        dpkg -r --force-depends evolution-data-server-common # 232.0 kB arch independ files for Evo Data Server

        dpkg -r --force-depends libebackend-1.2-10 # 393.00 KB Utility library for evolution data servers
        dpkg -r --force-depends libebook-1.2-20 # 365.0 kB Client library for evolution address books	
        dpkg -r --force-depends libebook-contacts-1.2-3 # 216.0 kB Client library for evolution contacts books

        dpkg -r --force-depends libecal-2.0-1 # 596.0 kB Client library for evolution calendars
        dpkg -r --force-depends libedata-book-1.2-26 # 810.0 kB Backend library for evolution address books
        dpkg -r --force-depends libedata-cal-2.0-1 # 524.0 kB Backend library for evolution calendars
        dpkg -r --force-depends libedataserver-1.2-24 # 975.0 kB Utility library for evolution data servers

        dpkg -r --force-depends ubuntu-wallpapers # 2,336.0 kB Ubuntu Wallpapers
        dpkg -r --force-depends ubuntu-wallpapers-focal # 11,867.0 kB	Ubuntu 20.04 Wallpapers
        dpkg -r --force-depends hunspell-en-us # 572.0 kB English_american dictionary for hunspell
        dpkg -r --force-depends libaspell15 # 2,268.0 kB	GNU Aspell spell-checker runtime library
        dpkg -r --force-depends libgweather-common # 1,640.0 kB GWeather common files
        dpkg -r --force-depends libhunspell-1.7-0 # 517.0 kB spell checker and morphological analyzer (shared library)
        dpkg -r --force-depends libwebpdemux2 # 36.0 kB	Lossy compression of digital photographic images
    fi
    rm -r /usr/src 
    rm /var/log/*.log
    rm /var/log/apt/*.log
    rm /usr/share/libthai/*
    rm /usr/share/info/*
    rm -r /usr/share/help/*
    rm -r /usr/share/i18n/*
    rm -r /usr/share/bug/*
    rm -r /usr/share/alsa/*
    rm /usr/share/consolefonts/*
    cp /usr/share/gtksourceview-3.0/language-specs/sh.lang /
    rm /usr/share/gtksourceview-3.0/language-specs/*
    cp /sh.lang /usr/share/gtksourceview-3.0/language-specs/ #leave one file in there for mousepad
    rm /sh.lang                                              #otherwise it segfaults
    rm -r /usr/share/file/*
    rm -r /usr/share/libwacom/*
    cp /var/lib/dpkg/info/dp* / #setting the timezone will not work if one or  
    cp /var/lib/dpkg/info/tz* / #both of the dpkg or tz files are not there
    rm /var/lib/dpkg/info/*
    cp /dp* /var/lib/dpkg/info/
    cp /tz* /var/lib/dpkg/info/
    rm /dp* 
    rm /tz* 
    rm /var/lib/dpkg/status-old
    rm /var/cache/debconf/templates.dat-old
    rm /var/cache/debconf/config.dat-old
    if [ $codename == "focal" ]; then
        rm /usr/lib/gcc/x86_64-linux-gnu/9/cc1 #don't need a c compiler
        rm /usr/share/libgphoto2/2.5.25/konica/*
    fi
    if [ $codename == "bionic" ]; then
        rm /usr/lib/gcc/x86_64-linux-gnu/7/cc1 #don't need a c compiler
    fi

}

#---------------------------------------------------------------------------------------------------------
# MAIN BODY
#---------------------------------------------------------------------------------------------------------

# Check if build folder exists
# if exists then delete so that a new build contains only the required files
if [ -d $foxfolder/$foxbuild ]; then
    echo $grn"deleting old build in $foxfolder/$foxbuild ..."$def
    rm -r $foxbuild
fi
# Check if iso exists, if so delete it
if [ -f $isofilename ]; then
    echo $grn"deleting old iso file in $isofilename ..."$def
    rm $isofilename
fi

echo
# make build folder to contain all files for iso
echo $grn"creating $foxfolder/$foxbuild for new build..."$def
mkdir $foxbuild
echo
echo $grn"creating image folder $foxfolder/$foxbuild/$image ..."$def
mkdir $foxbuild/$image
echo
echo $grn"creating root folder $foxfolder/$foxbuild/$newroot ..."$def
mkdir $foxbuild/$newroot
echo

# create initial system
if [ ! -f $cachefile ]; then
    echo $grn"bootstraping the base system ..."$def
    #debootstrap --arch=$arch --verbose --print-debs --variant=minbase $codename $foxbuild/$newroot
    debootstrap --arch=$arch --verbose $codename $foxbuild/$newroot
    echo $grn"creating cache $cachefile ..."$def
    tar -cf $cachefile -I pigz $foxbuild/$newroot
    #to avoid having files owned by root in the foxclone build folder 
    #useful if running backup utilities like backintime
    if [ ! ${#localuser} -eq 0 ]; then #there is a local user
        echo $grn"chowning $cachefile to $localuser ..."$def
        chown $localuser:$localuser $cachefile
        echo
    fi
else
    echo $grn"restoring base from cache $cachefile ..."$def
    tar -I pigz -xf $cachefile -C ./
    echo
fi

# copy over /etc/hosts and /etc/resolv.conf so that chroot has network access to install packages
# -v = verbose
echo $grn"Copying hosts & resolv.conf ..."$def
cp -v /etc/hosts $foxbuild/$newroot/etc/hosts
cp -v /etc/resolv.conf $foxbuild/$newroot/etc/resolv.conf
echo

# copy sources.list and modify for codename so chroot has access to the repositories
# without this couldnt install anything in ubuntu, but was able to install in buster
# need to have a check on file exists and also if building buster
echo $grn"Copying ubuntu-repositories and modifying for $codename ..."$def
cp -v ubuntu-repositories.list $foxbuild/$newroot/etc/apt/sources.list.d
#remove /etc/apt/sources.list produced by debootstrap
rm $foxbuild/$newroot/etc/apt/sources.list
sed -i "s/xxxx/$codename/" $foxbuild/$newroot/etc/apt/sources.list.d/ubuntu-repositories.list
echo

# copy over debs to be installed
echo $grn"Copying browser & foxclone debs ..."$def
cp -v $browserdeb $foxbuild/$newroot
cp -v $foxclonedeb $foxbuild/$newroot
echo

# copy plymouth theme - spinner logo
echo $grn"Copying plymouth theme ..."$def
# -p = no error if existing, make parent directories as needed
mkdir -p $foxbuild/$newroot/usr/share/plymouth/themes/$plymouth
cp -vr $foxfolder/$plymouth $foxbuild/$newroot/usr/share/plymouth/themes
echo

# mount the new system
# /dev is populated by the kernel, but is not a kernel-provided filesystem, so it has to be bind-mounted.
# note seems to work without mounting dev 
echo $grn"Mounting dev ..."$def
mount -o bind /dev $foxbuild/$newroot/dev
echo

#run chroot
echo $yel"ENTERING CHROOT-----------------------------------------------------------------"$def
export -f install_sh      # -f = refer to shell functions
chroot $foxbuild/$newroot /bin/bash -c "install_sh"
install_error=$?
echo $yel"EXITING CHROOT------------------------------------------------------------------"$def
#umount the system
for i in dev/pts proc sys dev
do
    echo $grn"   umounting "$foxfolder/$foxbuild/$newroot/$i $def
    umount $foxfolder/$foxbuild/$newroot/$i
    sleep 2 #included because sometimes the umount fails
done
if [[ $install_error -ne 0 ]]; then
    echo $red"ERROR: install in chroot ... "$def
    exit 1
fi

# Copy in the config files and various other things from the overlay folder
echo $grn"Copying src folder ..."$def
echo $yel"rsync --archive src/livecd/ $foxbuild"$def
rsync --archive src/livecd/ $foxbuild
# Copy the foxclone & uguide desktop files into the Desktop
cp $foxbuild/$newroot/usr/share/applications/fox* $foxbuild/$newroot/home/ubuntu/Desktop

#create folders and copy the linux image into casper
echo $grn"creating folders $foxbuild/$image/casper $foxbuild/$image/isolinux $foxbuild/$image/install"$def
mkdir -p $foxbuild/$image/casper $foxbuild/$image/isolinux $foxbuild/$image/install
echo $grn"copying linux image into casper"$def
cp $foxbuild/$newroot/boot/$vmlinuz_img $foxbuild/$image/casper/vmlinuz
if [[ $? -ne 0 ]]; then
    echo "Error: Failed to copy vmlinuz image."
    exit
fi

#copy initrd.lz
echo $yel cp $foxbuild/$newroot/boot/initrd.img $foxbuild/$image/casper/initrd.lz $def
cp -v $foxbuild/$newroot/boot/$initrd_img $foxbuild/$image/casper/initrd.lz
if [[ $? -ne 0 ]]; then
   echo $red"Error: Failed to copy initrd image for $codename."$def
   exit
fi

# add boot files for legacy boot
echo $grn"copying syslinux files from host system ..."$def
cp -v $foxfolder/syslinux/* $foxbuild/$image/isolinux/
if [[ $? -ne 0 ]]; then
    echo "Error: Failed to copy ISOLINUX files."
    exit 1
fi

# sort boot issues, focal insists on doing a filesystem check on boot and finding
# an error that is not an error, default is to fsck.mode=skip in boot params.
# Do not need to do this with bionic so remove it. Need to do it in 
# /build/image/boot/grub/grub.cfg for UEFI and /build/image/isolinux/isolinux.cfg
#for legacy boot.
if [ $codename == 'bionic' ]; then
    sed -i "s/fsck.mode=skip //" $foxbuild/$image/boot/grub/grub.cfg
    sed -i "s/fsck.mode=skip //" $foxbuild/$image/isolinux/isolinux.cfg
fi

# sort shutdown issues, bionic is okay but focal hangs on shutdown for 90 secs
# think it is trying and failing to unmount /cdrom.
# /usr/sbin/graphical-shutdown.sh has 
# [[ $ret -eq 1 ]] && poweroff      change to poweroff --force
# [[ $ret -eq 2 ]] && reboot        change to reboot --force 
# try reboot --force
if [ $codename != 'bionic' ]; then
    sed -i "s/poweroff #/poweroff --force #/" $foxbuild/$newroot/usr/sbin/graphical-shutdown.sh
    sed -i "s/reboot #/reboot --force #/" $foxbuild/$newroot/usr/sbin/graphical-shutdown.sh
fi
#previously okay but now complaining that don't have permission to execute shutdown
#chmod so all can execute - but why, worked before
chmod a+x $foxbuild/$newroot/usr/sbin/graphical-shutdown.sh
chmod a+x $foxbuild/$newroot/usr/sbin/graphical-shutdown.sh
# copy over memtest
echo $grn"copying memtest from host system ..."
cp -v /boot/memtest86+.bin $foxbuild/$image/install/memtest

# Create manifest 
echo $yel"ENTERING CHROOT-----------------------------------------------------------------"$def
echo $grn"creating filesystem manifest ..."$def

chroot $foxbuild/$newroot dpkg-query -W --showformat='${Package} ${Version}\n' > $foxbuild/$image/casper/filesystem.manifest
echo $yel"EXITING CHROOT------------------------------------------------------------------"$def

#copy the manifest to -desktop, and then clean it up, why????
# -v is verbose
cp -v $foxbuild/$image/casper/filesystem.manifest $foxbuild/$image/casper/filesystem.manifest-desktop
REMOVE='ubiquity ubiquity-frontend-gtk ubiquity-frontend-kde casper lupin-casper live-initramfs user-setup discover xresprobe os-prober libdebian-installer4'
for i in $REMOVE; do
  sed -i "/${i}/d" $foxbuild/$image/casper/filesystem.manifest-desktop
done
# it finds casper, user-setup, discover, os-prober (I need this for foxclone) 

#what is this file used for?
cat << EOF > $foxbuild/$image/README.diskdefines
#define DISKNAME foxclone
#define TYPE binary
#define TYPEbinary 1
#define ARCH $arch
#define ARCH$arch 1
#define DISKNUM 1
#define DISKNUM1 1
#define TOTALNUM 0
#define TOTALNUM0 1
EOF

#what is this used for??
touch $foxbuild/$image/ubuntu #nothing is put in it maybe later??
mkdir $foxbuild/$image/.disk
cd $foxbuild/$image/.disk
touch base_installable
echo "full_cd/single" > cd_type
echo "Ubuntu Remix" > info
echo "https://www.foxclone.com" > release_notes_url
cd ../..
ls
#think we should be in the image folder - no back in build folder
# ../.. means move up two levels

#not needed there is no file to remove as build is recreated every time
#rm -rf image/casper/filesystem.squashfs 
# source destination optinons
# think -e is exclude
# try using pigz for compression????
mksquashfs $newroot $image/casper/filesystem.squashfs -e boot -e /sys -e /proc


printf $(sudo du -sx --block-size=1 $newroot | cut -f1) > $image/casper/filesystem.size
cd $image
find . -type f -print0 | xargs -0 md5sum | grep -v "./md5sum.txt" > md5sum.txt
#why md5sums and not something more recent?? 
#md5sums go into the build/image folder

#--------------------------
# now make a UEFI partition
#--------------------------

# more info on what its trying to do https://superuser.com/questions/1390369/how-to-create-a-file-of-the-exact-size-of-a-directory-and-its-files-to-hold-a
BOOT_IMG_DATA="$PWD"
#currently $PWD should be $image which it is
echo $yel"current folder=$BOOT_IMG_DATA"$def
BOOT_IMG=efi.img
echo $yel"mkdir -p $(dirname $BOOT_IMG)"$def
#this appears as mkdir -p . ??????
# -p make parent folders if needed
mkdir -p $(dirname $BOOT_IMG)
#dirname - strip last component from file name
#Output each NAME with its last non-slash component and trailing slashes
#removed; if NAME contains no /'s, output '.' (meaning the current directory
#e.g. dirname /usr/bin/ -> "/usr"
#so sounds like it does nothing??? as current folder exists 


echo $yel"truncate -s 4M $BOOT_IMG"$def
#Shrink or extend the size of each FILE to the specified size
# -s = size, so making efi.img 4M
truncate -s 4M $BOOT_IMG
#now have a file in image, efi.img = 4.2MB

echo $yel"mkfs.vfat $BOOT_IMG"$def
#28 Jan 2022 change to fat32, change back - wrong syntax
#mkfs.vfat $BOOT_IMG
mkfs.vfat $BOOT_IMG
echo $ye"mkdir -p $BOOT_IMG_DATA/EFI/boot"
mkdir -p $BOOT_IMG_DATA/EFI/boot


#make a bootable image of GRUB
# -C = compression options are xz/none/auto
# -O = format = wide range of os types available
# -p = set prefix directory, what does this mean?
# -o = outputfile, looks like a list of files to create
echo $yel"creating bootable image of grub ..."$def
grub-mkimage \
    -C xz \
    -O x86_64-efi \
    -p /BOOT/grub \
    -o $BOOT_IMG_DATA/EFI/boot/bootx64.efi \
    boot linux search normal configfile \
    part_gpt btrfs fat iso9660 loopback \
    test keystatus gfxmenu regexp probe \
    efi_gop efi_uga all_video gfxterm font \
    echo read ls cat png jpeg halt reboot
#this is an exact copy of the example at https://superuser.com/questions/1390369/how-to-create-a-file-of-the-exact-size-of-a-directory-and-its-files-to-hold-a
# only difference -p /BOOT/grub instead of -p /boot/grub

#this created a /home/andy/foxiso/build/image/EFI/boot/bootx64.efi = 942kb

# mcopy - copy MSDOS files to/from Unix
# -s = recursive
# https://en.wikipedia.org/wiki/Mtools
# Copying a file to floppy image:
# mcopy -i Disk.img file_source ::file_target
# The drive character : (colon) has a special meaning. It is used to access image files which are 
# directly specified on the command line using the -i options
mcopy -i $BOOT_IMG -s $BOOT_IMG_DATA/EFI ::

#-----------------
# now make the ISO
#-----------------
new_files="$PWD"
echo $yel"new files = $new_files"$def
echo $grn"building the iso in $isofilename ..."$def

xorriso -as mkisofs "$new_files" \
    -V "FoxClone" \
    -o "$isofilename" \
    -isohybrid-mbr isolinux/isohdpfx.bin \
    -c isolinux/boot.cat \
    -b isolinux/isolinux.bin \
    -no-emul-boot -boot-load-size 4 -boot-info-table \
    -eltorito-alt-boot \
    -e boot/grub/efi.img \
    -no-emul-boot \
    -isohybrid-gpt-basdat \
    -m grub.cfg
if [[ $? -eq 0 ]]; then
    echo $yel"ISO file created at $isofilename ..."
    echo "ISO file size =" `ls -l --si $isofilename | cut -d " " -f5`B $def
    if [ ${#localuser} -ne 0 ]; then #there is a local user
        echo $grn"chowning $isofilename to $localuser ..."$def
        chown $localuser:$localuser $isofilename
        echo
    fi
else
    echo $red"Failed to create ISO at $isofilename ..."$def
fi

# how long does the script take to run? taken from
# https://stackoverflow.com/questions/8903239/how-to-calculate-time-elapsed-in-bash-script
duration=$SECONDS
echo $yel"ISO creation time: "$(($duration / 60)) minutes and $(($duration % 60))" seconds elapsed."$def

cd $foxfolder




