{ FoxClone
  Copyright (C) 2021 Andy Hardwick
  This source is free software; you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free
  Software Foundation; either version 2 of the License, or (at your option)
  any later version.
  This code is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
  details.
  A copy of the GNU General Public License is available on the World Wide Web
  at <http://www.gnu.org/copyleft/gpl.html>. You can also obtain it by writing
  to the Free Software Foundation, Inc., 51 Franklin Street - Fifth Floor,
  Boston, MA 02110-1335, USA.
}
{-------------------------------------------------------------------------------
UNIT - FDIALOG

A replacement for MessageDlg in Dialogs. Why? MessageDlg has the following limitations:

    Does not wordwrap - means that linefeeds have to be embedded in the string making international
    versions more cumbersome. FoxMsg automatically word wraps.

    Only provides one message string, so making up a string including terminal output is more
    cumbersome. FoxMsg allows for multiple message strings.

    Does not autosize - FoxMsg autosizes, although resizing has some issues.

FoxMsg uses the same TMsgDlgType and TMsgDlgBtn types as MessageDlg.

Useful source of info: https://www.thoughtco.com/communicating-between-forms-4092543

Limitations - only two buttons are provided for.

Note - the utilities unit contains functions GetTextWidth and GetTextHeight. They are identical to the
functions in this unit. The functions in this unit have been renamed xGetTextWidth and xGetTextHeight
and retained so that the unit is standalone.

Note - utilities is used, only because this is where imagefolder is defined.

V1.0 - 11 May 2021
V1.1 - 12 Jun 2021, MsgLbl changed ot MsgText and a TMemo, line spacing is slightly better, cramped with
                    a tlabel and wordwrap on.
V1.2 - 19 Aug 2021, made specific to foxclone, as the font size increases (as set in the settings tab)
                    the width of the form for the message increases.
-------------------------------------------------------------------------------}
unit fdialog;

{$mode objfpc}{$H+}

interface

uses
  //utilities - needed because that is where imagefolder is defined
  Classes, SysUtils, Forms, Controls, Dialogs, Graphics, ComCtrls, StdCtrls, ExtCtrls, utilities;

resourcestring
  rYesStr = 'Yes';
  rNoStr ='No';
  rOkStr = 'OK';
  rCancelStr = 'Cancel';
  rAbortStr = 'Abort';
  rRetryStr = 'Retry';
  rIgnoreStr = 'Ignore';
  rAllStr = 'All';
  rNoAllStr = 'No to All';
  rYesAllStr = 'Yes to All';
  rHelpStr = 'Help';
  rCloseStr = 'Close';

type
  TFoxMsg = class(TForm)
    const
      fwBorder = 15; //default border unit for gaps horizontally in the form, same as panels (pwBorder)
      fhBorder = 15;
      fxWidth = 320; //initial form width
      fwImage = 48;
      ErrorImg = ImageFolder + 'dialog-error.png';
      InfoImg = ImageFolder + 'dialog-information.png';
      ConfirmImg = ImageFolder + 'dialog-question.png';
      WarnImg = ImageFolder + 'dialog-warning.png';
      SysErrorImg = ImageFolder + 'dialog-system-error.png';
     private
      msgImg : TImage;
      MsgText :  TMemo;
      Btn1, Btn2 : TButton;
      function xGetTextWidth(aText: String; AFont: TFont): Integer;
      function xGetTextHeight(aText: String; AFont: TFont): Integer;
      procedure Btn1Click(Sender: TObject);
      procedure Btn2Click(Sender: TObject);
      procedure FoxMsgResize(Sender : TObject);
    public
      constructor CreateNew(TheOwner: TComponent);
  end;
//the function called by other procedures to display a message
function FoxMsg(const aCaption : string; aMsg : array of string; aType : TMsgDlgType;
                aButtons: array of TMsgDlgBtn) : integer;

implementation
uses mainform, settings; //this makes it specific to foxclone
{-------------------------------------------------------------------------------
CONSTRUCTOR TFOXMSG.CREATENEW
Constructor for the dialog form. Creates all the objects and sets their alignments and spacing.

Assignment of captions, loading of images and sizing of the form is done when the function FoxMsg
is called.

Used by:

V1.0 - 13 Mar 2021
V1.1 - 11 May 2021, min width of form set, funny it wasn't being used before and width on creation
                    was okay?
V1.2 - 12 Jun 2021, MsgLbl changed to MsgText and changed from TLabel to TMemo = better line spacing.
V1.3 - 19 Aug 2021, made dependent on mainform and form width gets wider as the font size selected in
                    settings increases.
V1.4 - 10 Dec 2021, set background colour of the tmemo for visual appearance in the iso. Setting the
                    colour did not work without removing enabled=false and using readonly=true
                    instead.
V1.5 - 09 Jan 2022, fsSystemStayOnTop added so messages are always on top
-------------------------------------------------------------------------------}
constructor TFoxMsg.CreateNew(TheOwner: TComponent);
begin
  inherited CreateNew(TheOwner);
  Position:= poMainFormCenter; //poScreenCenter;
  OnResize:= @FoxMsgResize;
  formstyle:=fsSystemStayOnTop; //hopefully above all other windows, inc stayontop
  msgImg:=TImage.Create(Self);
  Font.Size:= MainFm.Font.Size; //this makes the unit specific to foxclone
  case Font.Size of //makes it specific to foxclone, make the window wider as the font increases
       10 : Constraints.MinWidth:= fxWidth;
       11 : Constraints.MinWidth:= fxWidth + 10;
       12 : Constraints.MinWidth:= fxWidth + 20;
       13 : Constraints.MinWidth:= fxWidth + 30;
       14 : Constraints.MinWidth:= fxWidth + 40;
       15 : Constraints.MinWidth:= fxWidth + 50;
       16 : Constraints.MinWidth:= fxWidth + 60;
       17 : Constraints.MinWidth:= fxWidth + 70;
       18 : Constraints.MinWidth:= fxWidth + 80;
  end; //case
  with msgImg do begin
       Parent:=self;
       AnchorSide[akTop].Side:= asrTop;
       AnchorSide[akTop].Control:= self;
       BorderSpacing.top:= fhBorder;
       AnchorSide[akLeft].Side:= asrLeft;
       AnchorSide[akLeft].Control:= self;
       BorderSpacing.left:= fwBorder;
       AutoSize:= true;
  end;
  MsgText:= TMemo.Create(Self);
  with MsgText do begin
       Parent:= Self;
       Color:= clForm;
       BorderStyle:= bsNone;
       Width:= Self.Width - fwImage - 3 * fwBorder;
       AnchorSide[akTop].Side:= asrTop;
       AnchorSide[akTop].Control:= msgImg;
       BorderSpacing.top:= 0;
       AnchorSide[akLeft].Side:= asrLeft;
       AnchorSide[akLeft].Control:= Self;
       BorderSpacing.left:= 2 * fwBorder + fwImage;
       WordWrap:= true;
       ReadOnly:= true;
  end;
  Btn1:= TButton.Create(Self);
  with Btn1 do begin
       Parent:= Self;
       AnchorSide[akTop].Side:= asrBottom;
       AnchorSide[akTop].Control:= MsgText;
       BorderSpacing.top:= 0;
       AnchorSide[akLeft].Side:= asrLeft;
       AnchorSide[akLeft].Control:= Self;
       BorderSpacing.left:= Self.Width - fwBorder - Width;
       OnClick:= @Btn1Click;
  end;
  Btn2:= TButton.Create(Self);
  with Btn2 do begin
       Parent:= Self;
       AnchorSide[akTop].Side:= asrTop;
       AnchorSide[akTop].Control:= Btn1;
       BorderSpacing.top:= 0;
       AnchorSide[akLeft].Side:= asrLeft;
       AnchorSide[akLeft].Control:= Self;
       BorderSpacing.left:= Self.Width - 2 * fhBorder - Width - Btn1.Width;
       OnClick:= @Btn2Click;
       Visible:= false;
       Enabled:= false;
  end;
end; //end constructor TFoxMsg.CreateNew
{-------------------------------------------------------------------------------
PROCEDURE TFOXMSG.BTN2CLICK
Event handler for button 1, it sets the form modalresult as the modalresult set for the button.

Used by:

V1.0 - 13 Mar 2021
-------------------------------------------------------------------------------}
procedure TFoxMsg.Btn1Click(Sender: TObject);
begin
  Self.ModalResult:= Btn1.ModalResult;
end; //end procedure TFoxMsg.Btn1Click
{-------------------------------------------------------------------------------
PROCEDURE TFOXMSG.BTN2CLICK
Event handler for button 2, it sets the form modalresult as the modalresult set for the button.

Used by:

V1.0 - 13 Mar 2021
-------------------------------------------------------------------------------}
procedure TFoxMsg.Btn2Click(Sender: TObject);
begin
  Self.ModalResult:= Btn2.ModalResult;
end; //end procedure TFoxMsg.Btn2Click
{-------------------------------------------------------------------------------
PROCEDURE TFOXMSG.FOXMSGRESIZE
Event handler for resizing.

Can't use the number of lines directly as not available so have to calculate the number of
linefeeds in the label caption.

Used by:

V1.0 - 09 May 2021
V1.1 - 11 May 2021, was explicitly setting form height in the procedure, which causes a loop as
                    resize calls itself. End result, form moves to (0,0) on the screen. Removed. Now
                    min size is set in the constructor and FoxMsg.
V1.2 - 12 Jun 2021, changes as a result of changing from a TLabel to a TMemo
-------------------------------------------------------------------------------}
procedure TFoxMsg.FoxMsgResize(Sender : TObject);
var
  i, LineHeight : byte;
  NoLines : byte = 0;
begin
  //message label
  MsgText.Width:= Self.Width - fwImage - 3 * fwBorder;
  LineHeight:= xGetTextHeight('Ag',MsgText.Font);
  for i:= 0 to MsgText.Lines.Count - 1 do with MsgText do
      if Lines[i] <> '' then NoLines:= NoLines + round(xGetTextWidth(Lines[i],Font)/Width + 0.5)
      else NoLines+= 1;
  MsgText.Height:= NoLines * LineHeight;
  //Buttons
  Btn1.BorderSpacing.left:= Self.Width - fwBorder - Btn1.Width;
  Btn2.BorderSpacing.left:= Self.Width - 2 * fhBorder - Btn2.Width - Btn1.Width;
end; //end procedure TFoxMsg.FoxMsgResize
{-------------------------------------------------------------------------------
FUNCTION TFOXMSG.GETTEXTWIDTH
Return the width of a text string in pixels.
Used in checklistboxes to determine item width and hence turn on/off horizontal scroll bars
Used in some labels to determine alignment as lazarus.

It pastes the text into a bitmap to determine width. Did look at other alternatives but based on
units either very old not straightforward.

Parameters:
  aText - the string to calculate the width of
  aFont - the font being used.

Used by:

V1.0 - 21 Feb 2021
-------------------------------------------------------------------------------}
function TFoxMsg.xGetTextWidth(aText: String; aFont: TFont): Integer;
var
  bmp: TBitmap;
begin
  Result:= 0;
  bmp:= TBitmap.Create;
  try
    bmp.Canvas.Font.Assign(aFont);
    Result:= bmp.Canvas.TextWidth(aText);
  finally
    bmp.Free;
  end;
end; //end TFoxMsg.GetTextWidth
{-------------------------------------------------------------------------------
FUNCTION TFOXMSG.GETTEXTHEIGHT
Return the height of a text string in pixels.
Used in checklistboxes to determine item width and hence turn on/off horizontal scroll bars
Used in some labels to determine alignment as lazarus.

It pastes the text into a bitmap to determine width. Did look at other alternatives but based on
units either very old not straightforward.

Parameters:
  aText - the string to calculate the width of
  aFont - the font being used.

Used by:

V1.0 - 21 Feb 2021
-------------------------------------------------------------------------------}
function TFoxMsg.xGetTextHeight(aText: String; aFont: TFont): Integer;
var
  bmp: TBitmap;
begin
  Result:= 0;
  bmp:= TBitmap.Create;
  try
    bmp.Canvas.Font.Assign(aFont);
    Result:= bmp.Canvas.TextHeight(aText);
  finally
    bmp.Free;
  end;
end; //end function TFoxMsg.GetTextHeight
{-------------------------------------------------------------------------------
FUNCTION FOXMSG
A replacemnt for MessageDlg. It provides:

    Multiple lines of msgs, MessageDlg only provides one.

    Word wrapping on the msgs, with MessageDlg you have to insert linefeeds into the text string, it
    doesn't do word wrap.

    Resizing of the dialog window - MessageDlg is fixed size.

    Centred on the parent form - I was getting messagedlg appearing top and left of the main window - not
    where I wanted it.

I did try CreateMessageDialog and MessageDlgPos but thought the buttons were naff with GTK.

Main issue was setting the TLabel height with a fixed width and wordwrap on. Tried a few solutions from
the web, but delphi based and didn't work. Settled on calculating the text width in pixels (GetTextWidth)
dividing by the width of the TLabel to get a base number of lines. To this is added the number of msgs.
Result is multiplied up by the line height (determined by GetTextHeight). Bit of an approximation as
text will break at spaces and not be exactly at the width of the TLabel.

Limitations - only two buttons.

It uses the same TMsgDlgType and TMsgDlgBtn as MessageDlg.

Returns the same modalresult as MessageDlg, this is set by the modalresult assigned to each button.

Parameters:
    aCaption - title for the message box.
    aMsg - an array of strings to display in the msg box, each one is displayed separated by 2 X LF.
    aType - TMsgDlgType, used to determine the image file to load, e.g. error, information.
    aButtons - TMsgDlgBtn, a max of two buttons are allowed, the first button type is the button on the
               right, the second is the button to the left.

Used by:

V1.0 - 09 May 2021
V1.1 - 12 Jun 2021, modified as a result of using a TMemo instead of a TLabel
V1.2 - 13 Jun 2021, NoLines was not added enough space, missing the empty lines.
V1.3 - 13 Dec 2021, modified to dump to the debug file.
-------------------------------------------------------------------------------}
function FoxMsg(const aCaption : string; aMsg : array of string; aType : TMsgDlgType;
                aButtons: array of TMsgDlgBtn) : integer;
var
  FoxForm : TFoxMSg;
  i, LineHeight : byte;
  NoLines : byte = 0; //should not have more than 255 lines
  MsgString : string = ''; //for debug
begin
  FoxForm:= TFoxMsg.CreateNew(nil);
  //send info to debug if enabled
  for i:= 0 to length(aMsg) - 1 do MsgString+= aMsg[i];
  //writeln('foxmsg string=',MsgString);
  DebugInfo(false,'FoxMsg',aCaption,MsgString);
  with FoxForm do
       begin
         Caption:= aCaption;
         //load the applicable image file, std TMsgDlgType are used, mtCustom is used for system error
         case aType of
              mtWarning      : msgImg.Picture.LoadFromFile(WarnImg);
              mtError        : msgImg.Picture.LoadFromFile(ErrorImg);
              mtCustom       : msgImg.Picture.LoadFromFile(SysErrorImg);
              mtInformation  : msgImg.Picture.LoadFromFile(InfoImg);
              mtConfirmation : msgImg.Picture.LoadFromFile(ConfirmImg);
         end;
         //provides for a max of two buttons, so Btn1 always present, set caption
         //Captions are assigned in the order that they are passed when FoxMsg called
         with Btn1 do Case aButtons[0] of
              mbYes      : begin Caption:= rYesStr; ModalResult:= mrYes; Default:= true; Cancel:= false end;
              mbNo       : begin Caption:= rNoStr; ModalResult:= mrNo; Default:= false; Cancel:= true end;
              mbOK       : begin Caption:= rOkStr; ModalResult:= mrOK; Default:= true; Cancel:= false end;
              mbCancel   : begin Caption:= rCancelStr; ModalResult:= mrCancel; Default:= true; Cancel:= true end;
              mbAbort    : begin Caption:= rAbortStr; ModalResult:= mrAbort; Default:= false; Cancel:= true end;
              mbRetry    : begin Caption:= rRetryStr; ModalResult:= mrRetry; Default:= true; Cancel:= false end;
              mbIgnore   : begin Caption:= rIgnoreStr; ModalResult:= mrIgnore; Default:= false; Cancel:= true end;
              mbAll      : begin Caption:= rAllStr; ModalResult:= mrAll; Default:= true; Cancel:= false end;
              mbNoToAll  : begin Caption:= rNoAllStr; ModalResult:= mrNoToAll; Default:= false; Cancel:= true end;
              mbYesToAll : begin Caption:= rYesAllStr; ModalResult:= mrYesToAll; Default:= true; Cancel:= false end;
              mbHelp     : begin Caption:= rHelpStr; ModalResult:= mrNone end; //doesn't close form
              mbClose    : begin Caption:= rCloseStr; ModalResult:= mrClose; Default:= true; Cancel:= false end;
         end;
         //is there a second button, if so apply caption
         if length(aButtons) = 2 then with Btn2 do begin
            Case aButtons[1] of
                 mbYes      : begin Caption:= rYesStr; ModalResult:= mrYes; Default:= true; Cancel:= false end;
                 mbNo       : begin Caption:= rNoStr; ModalResult:= mrNo; Default:= false; Cancel:= true end;
                 mbOK       : begin Caption:= rOkStr; ModalResult:= mrOK; Default:= true; Cancel:= false end;
                 mbCancel   : begin Caption:= rCancelStr; ModalResult:= mrCancel; Default:= true; Cancel:= true end;
                 mbAbort    : begin Caption:= rAbortStr; ModalResult:= mrAbort; Default:= false; Cancel:= true end;
                 mbRetry    : begin Caption:= rRetryStr; ModalResult:= mrRetry; Default:= true; Cancel:= false end;
                 mbIgnore   : begin Caption:= rIgnoreStr; ModalResult:= mrIgnore; Default:= false; Cancel:= true end;
                 mbAll      : begin Caption:= rAllStr; ModalResult:= mrAll; Default:= true; Cancel:= false end;
                 mbNoToAll  : begin Caption:= rNoAllStr; ModalResult:= mrNoToAll; Default:= false; Cancel:= true end;
                 mbYesToAll : begin Caption:= rYesAllStr; ModalResult:= mrYesToAll; Default:= true; Cancel:= false end;
                 mbHelp     : begin Caption:= rHelpStr; ModalResult:= mrNone end; //doesn't close form
                 mbClose    : begin Caption:= rCloseStr; ModalResult:= mrClose; Default:= true; Cancel:= false end;
            end;
            Visible:= true;
            Enabled:= true;
         end;
         //button heights
         with Btn1 do Height:= xGetTextHeight(Caption,Font) + 14; //adjustment to make is visually better
         Btn2.Height:= Btn1.Height;
         //message string
         //no reliable method for getting label height in lazarus with wordwrap on found so
         //get text width and divide by label width, add in lines for number of strings passed
         for i:= 0 to length(aMsg) - 1 do with MsgText do
              begin
                Lines.Add(aMsg[i]);
                Lines.Add('');
                NoLines:= NoLines + Round(xGetTextWidth(aMsg[i],Font)/Width + 0.5) + 1; //+1 for the empty line
              end;
         LineHeight:= xGetTextHeight('Ag',MsgText.Font);
         MsgText.Height:= NoLines * LineHeight;
         //set form height
         FoxForm.Height:= fhBorder * 3 + MsgText.Height + Btn1.Height;
         //set min constraints - min width is set in the constructor
         FoxForm.Constraints.MinHeight:= fhBorder * 3 + Btn1.Height + length(aMsg) * LineHeight * 2;
       end;
  FoxMsg:= FoxForm.showmodal;
  FoxForm.free;
end; //end function FoxMsg

end.

