{ FoxClone
  Copyright (C) 2021 Andy Hardwick
  This source is free software; you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free
  Software Foundation; either version 2 of the License, or (at your option)
  any later version.
  This code is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
  details.
  A copy of the GNU General Public License is available on the World Wide Web
  at <http://www.gnu.org/copyleft/gpl.html>. You can also obtain it by writing
  to the Free Software Foundation, Inc., 51 Franklin Street - Fifth Floor,
  Boston, MA 02110-1335, USA.
}
{-------------------------------------------------------------------------------
UNIT - INFO

The previous approach of populating the individual elements in TDrive and Tpart was reasonably
structured but suffered in terms of speed. An example - parted was being called three times for each
partiton. This unit contains the data structures to hold the output from parted and sfdisk (to get the
partition table on each drive). Parted output as a TStringList and sfdisk output as an TObjectList of
TStringLists.

Population of TDrives and TParts are now done through calling the methods in this unit to parse the
information in TStringLists for parted or sfdisk, or to run terminal commands. This is a lot faster
than the previous implementation.

Debug also uses this unit to dump the output of parted and the sfdisk partition table output from
each drive in the debug file.

V1.0 - 17 Dec 2021
V1.1 - 18 Dec 2021 dfList added to type definition - to hold output from df - where are parts mounted?
V1.2 - 29 Dec 2021 dfList removed.
V1.3 - 03 Jan 2022 support for exfat added
V1.4 - 23 Oct 2023 errors fixed for a drive with more than 9 partitions, some functions would match
                   sda1 and sda10, solved by added a space into the search criteria.
-------------------------------------------------------------------------------}
unit Info;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, Contnrs, StrUtils, Types, utilities, Dialogs, fdialog, il8nstrings;

type
  TSFlist = class(TObjectList) //a list containing TStringLists of sfdisk output
    private
      function GetItems(aIndex: Integer): TStringList;
    public
      property Items[Index: Integer]: TStringList read GetItems; default;
  end;
  TSysInfo = class(TObject) //holds the output of various commands giving info on the system
    const
      devstr = '/dev/';
      delim = ':';
      exitstr = 'BYT;'; //precedes the start of output on a drive in the output of parted -lm
      comma = ',';
    private
      PartList : TStringList; //container for the output of parted -lm
      FullPartList : TStringList; //parted -l, human readable, needed for primary/extended/logical
      sfIndex : array of string; //contains the device name, e.g. sda and provides index into the TSFlist
      sfList : TSFlist; //a list of sfdisk outputs for each drive
      aList : TStringList; //a TStringList to add to sfList
      function FindPTable(Device: string) : shortint;
      function FindDrive(DeviceP: string) : shortint;
    public
      NoDrives : byte; //number of drives in the system
      NoPartitions : byte; //number of partitions in the system
      constructor create;
      procedure runSFdisk;
      procedure GetDrives(DriveList : TStringList); //populates DriveList with devices, sda, sdb, etc.
      function PartCount(Device : string) : byte; //returns # of partitions on Device
      procedure GetParts(PartNames : TStringList; Device : string);
      //Drive information
      function getModel(Device : string) : string;
      function getLogSectorSize(Device : string) : integer;
      function getPTableType(Device : string) : string;
      function getSerial(Device : string) : string;
      function getFirstPartition(Device : string) : integer;
      function getDriveSize(Device : string; SectorSize : integer) : int64;
      function getDriveUUID(Device : string) : string;
      function getFirstLBA(Device : string) : shortint;
      function getLastLBA(Device : string) : int64;
      //Partition information
      function getFlags(DeviceP : string; Number : string) : string;
      function getGPTname(DeviceP : string; Number : string) : string;
      function getMount(DeviceP : string; Number : string) : string;
      function getPStart(DeviceP : string; Number : string) : int64;
      function getPsize(DeviceP : string; Number : string; SectorSize : integer) : int64;
      function getPtype(DeviceP : string; Number : string) : string;
      function getUsedBlks(DeviceP : string; Number : string; SectorSize : integer) : int64;
      function getPSFtype(DeviceP : string; Number : string) : string;
      function getFS(DeviceP : string; Number : string) : string;
      function getLabel(DeviceP : string; Number : string) : string;
      function getUUID(DeviceP : string; Number : string) : string;
      function getBootFlag(DeviceP : string; Number : string) : string;
      //copy procedures used by debug
      procedure CopyPartedLM(LogList : TStringList);
      procedure CopyParted(LogList : TStringList);
      procedure CopyPtables(LogList : TStringList);
  end;

var
  SysInfo : TSysInfo;

implementation
{-------------------------------------------------------------------------------
CONSTRUCTOR TSYSINFO.CREATE
Creates TSysInfo and dumps the output from parted into PartList. This uses the parted -lm output -
output in machine readable form, previously used parted -l. Format:

       BYT;
       /dev/sda:240GB:scsi:512:512:gpt:ATA WDC WDS240G2G0A-:;
       1:1049kB:556MB:555MB:ntfs:Basic data partition:hidden, diag;
       2:556MB:661MB:105MB:fat32:EFI system partition:boot, esp;
       3:661MB:677MB:16.8MB::Microsoft reserved partition:msftres;
       4:677MB:40.8GB:40.2GB:ntfs:Basic data partition:msftdata;
       5:40.8GB:56.6GB:15.7GB:ext4::;
       6:56.6GB:77.5GB:21.0GB:ext4::;
       7:77.5GB:85.9GB:8389MB:linux-swap(v1)::swap;

The output is parsed to count the number of drives and set NoDrives. sfList (a TObjectList of sfdisk
output) is created but NOT populated.

The order that the content (parted and sfdisk) is populated is so that the startup form can show
progress at appropriate points. parted = finding drives, sfdisk = finding partitions.

Used by:

V1.0 - 17 Dec 2021
V1.1 - 19 Dec 2021, dfList added to get initial mount points, fullpartedlist added, it is the only
                    easy source of is a part extended, logical, primary. Otherwise partlist with
                    parted -lm is more convenient to extract info. might move elsewhere for startup
                    form display timing
V1.2 - 24 Dec 2021, now checks for /dev/mapper before incrementing the number of drives. dfList
                    removed, FullPartList added for the 'human readable' output of parted. This is the
                    only way of finding out if a partition is extended or logical (not in the parted -lm
                    output).
V1.3 - 28 Jan 2022, parted -l output now redirects stderr so that any errors from parted are reported
                    as well.
-------------------------------------------------------------------------------}
constructor TSysInfo.create;
var
  i : byte;
begin
  inherited create;
  //populate partlist, use parted -lm to get the 'machine readable' output
  PartList:= TStringList.Create;
  getText('pkexec parted -lm',PartList);
  //populate fullpartlist from parted -l, only needed to find extended parts
  FullPartList:= TStringList.Create;
  getText('pkexec parted -l 2>&1',FullPartList); //include output from stderr as well
  //count the number of drives in the partfunction
  NoDrives:= 0;
  for i:= 0 to PartList.Count - 1 do if pos('/dev/',PartList[i]) = 1 then
      if pos('mapper',PartList[i]) = 0 then NoDrives+= 1;
  sfList:= TSFlist.Create;
end; //end constructor TSysInfo.create
{-------------------------------------------------------------------------------
FUNCTION TSFLIST.GETITEMS
A wrapper so that instead of accessing an element in aList (sfdisk output as a TStringList) with:
  TStringList(SysInfo[i])[j]

It can be accessed with:
  SysInfo[i][j]

See also TSystem where the same form of wrapper is used.

Used by:

V1.0 - 17 Dec 2021
-------------------------------------------------------------------------------}
function TSFlist.GetItems(aIndex: Integer): TStringList;
begin
  result:= TStringList(get(aIndex));
end; //end function TSFlist.GetItems
{-------------------------------------------------------------------------------
PROCEDURE TSYSINFO.RUNSFDISK
Runs sfdisk for all drives in the system, dumps the output in a TStringList (aList) and adds aList to
sfList. It also populates an array of strings (sfIndex) with the device name, e.g. sda. This is so that
a search on sfIndex can return the index to the sfdisk output in the TStringList in sfList. The sfdisk
ouput can then be parsed for the relevant information.

The procedure counts the number of partitions in the system and populates NoPartitions. Used for display
in the startup form.

Used by:

V1.0 - 17 Dec 2021
V1.1 - 25 Dec 2021, checks for a loop device and increments the number of partitions - no of parts is
                    calculated from the sfdisk output for each drive. No part table on a loop device so
                    it gets the number of parts wrong.
-------------------------------------------------------------------------------}
procedure TSysInfo.runSFdisk;
const
  cLoop = 'loop';
var
  Dname : string; //device name, e.g. sda
  i,j : byte;
begin
  NoPartitions:= 0;
  //get partition tables with sfdisk and add to sfList
  for i:= 0 to PartList.Count - 1 do if pos(devstr,PartList[i]) = 1 then
      begin
        if pos(cLoop,PartList[i]) > 0 then NoPartitions+= 1; //check for a loop device and inc. no parts
        //writeln('partlist output=',partlist[i]);
        Dname:= copy(PartList[i],1,pos(delim,PartList[i])-1);
        //writeln('device name =',Dname);
        aList:= TStringList.Create;
        getText('pkexec sfdisk --dump ' + Dname,aList);
        if aList.Count > 0 then
           begin
             sfList.Add(aList);
             setlength(sfIndex,sfList.Count); //increase array size
             sfIndex[sfList.Count-1]:= replacestr(Dname,devstr,''); //strip out /dev/
           end;
      end;
  //count the partitions
  for i:= 0 to sfList.Count - 1 do
      for j:= 0 to sfList[i].Count - 1 do
             if pos('/dev/',sfList[i][j]) = 1 then NoPartitions+= 1;
  //for i:= 0 to length(sfIndex) - 1 do writeln('sfIndex ',i,' = ',sfIndex[i]);
  //if there are any loop devices this will miss the partition, check for it
end; //end procedure TSysInfo.runSFdisk
{-------------------------------------------------------------------------------
PROCEDURE TSYSINFO.GETDRIVES
Searches the parted output in PartList for lines starting /dev/, strips everything in the line out
except for the device name, e.g. sda, and adds it to a TStringList. This is then used elsewhere to
create a TDrive for each drive found in the system.

Parameters:
    DriveList - a TStringList populated with the device names for drives found in the system.

V1.1 - testing the function to get the part table type, sdb was modified to include an extended
partition. gparted completed without reporting any errors (did seem to take a long time). Subsequently
sdb1 is showing errors in gparted. Standard output from parted shows sdb as normal:

   Model: HITACHI HTS545032A7E380 (scsi)
   Disk /dev/sdb: 320GB
   Sector size (logical/physical): 512B/4096B
   Partition Table: msdos
   Disk Flags:

   Number  Start   End    Size   Type      File system  Flags
    1      1049kB  149GB  149GB  primary   ext4
    2      149GB   320GB  171GB  extended
    5      149GB   320GB  171GB  logical   ext4

but the following has appeared:

   Model: Linux device-mapper (linear) (dm)
   Disk /dev/mapper/osprober-linux-sdb1: 149GB
   Sector size (logical/physical): 512B/4096B
   Partition Table: loop
   Disk Flags:

   Number  Start  End    Size   File system  Flags
    1      0.00B  149GB  149GB  ext4

Used by:

V1.0 - 17 Dec 2021
V1.1 - 19 Dec 2021, if a /dev/mapper device is found it will be added to the list. Modified to exclude.
                    Error message added.
V1.2 - 10 Jan 2022, text about running fsck on finding a dev mapper removed and msg changed to warning
                    not error
V1.3 - 17 Apr 2024, errors out when there is a virtual device, e.g. LUKS, problem is the foxmsg, first
                    time okay, second time

                    The program 'foxclone52' received an X Window System error.
                    This probably reflects a bug in the program.
                    The error was 'BadWindow (invalid Window parameter)'.
                      (Details: serial 6774 error_code 3 request_code 12 minor_code 0)
                      (Note to programmers: normally, X errors are reported asynchronously;
                      that is, you will receive the error a while after causing it.
                      To debug your program, run it with the --sync command line
                      option to change this behavior. You can then get a meaningful
                      backtrace from your debugger if you break on the gdk_x_error() function.)
                    Tried with dummy text in the foxmsg, again first time okay, second error

                    Solved by simply removing the message which would probably cause some users to
                    contact me wanting to know if okay.
-------------------------------------------------------------------------------}
procedure TSysInfo.GetDrives(DriveList : TStringList);
const
  mapperstr = '/mapper/';
var
  Dname : string; //device name, e.g. sda
  i : byte;
begin
  for i:= 0 to PartList.Count - 1 do
      if pos(devstr,PartList[i]) = 1 then
         begin
           Dname:= copy(PartList[i],1,pos(delim,PartList[i])-1);
           Dname:= replacestr(Dname,devstr,'');
           if pos(mapperstr,PartList[i]) = 0 then DriveList.Add(Dname);
           //else FoxMsg(rVirtTitleStr,[rVirtDevStr,devstr+Dname],mtWarning,[mbOK]);
         end;
end; //end procedure TSysInfo.GetDrives
{-------------------------------------------------------------------------------
FUNCTION TSYSINFO.PARTCOUNT
Counts the number of partitions on a drive and returns the count as the result.

Parameters:
    Device - the drive to count partitions on, in the form sda.

The function searches the array of drive names in sfIndex to get the index into sfList. It then counts
the lines in the sfdisk output starting /dev/.

Used by:

V1.0 - 17 Dec 2021
-------------------------------------------------------------------------------}
function TSysInfo.PartCount(Device : string) : byte;
var
  i : byte;
  Index : shortint;
begin
  PartCount:= 0; //default
  Index:= -1; //failed state
  for i:= 0 to length(sfIndex) - 1 do if sfIndex[i] = Device then Index:= i;
  if Index >= 0 then
     for i:= 0 to sfList[Index].Count - 1 do
         if pos(devstr,sfList[Index][i]) = 1 then PartCount+= 1;
end; //end function TSysInfo.PartCount
{-------------------------------------------------------------------------------
FUNCTION TSYSINFO.GETMODEL (drive)
Returns the model name for the drive specified in Device. The model name is in the parted output in
the form

     /dev/sda:240GB:scsi:512:512:gpt:ATA WDC WDS240G2G0A-:;

where the model name is ATA WDC WDS240G2G0A-

The string is split into a string array using : as the delimiter and the name extracted.

Parameters:
    Device - the drive where the model name is required, passed in the form sda

Used by:

V1.0 - 17 Dec 2021
V1.2 - 18 Jan 2022, was failing to match on a restore, reason model extracted here was 'SABRENT ' while
                    restore was searching for 'SABRENT' and failing, trailing space was the problem.
                    Fixed with trim.
-------------------------------------------------------------------------------}
function TSysInfo.getModel(Device : string) : string;
var
  i : byte;
  Index : shortint;
  Fields : TStringDynArray;
begin
  Index:= -1; //failed state
  getModel:= ''; //failed state
  for i:= 0 to PartList.Count - 1 do
      if pos(devstr + Device,PartList[i]) = 1 then
         begin
           Index:= i;
           break;
         end;
  if Index >= 0 then
     begin
       //extract the parameters out of the string
       Fields:= SplitString(PartList[Index],delim);
       getModel:= trim(Fields[6]);
     end;
end; //end function TSysInfo.getModel
{-------------------------------------------------------------------------------
FUNCTION TSYSINFO.GETLOGSECTORSIZE (drive)
Returns the logical sector size in bytes for the drive specified in Device. The sector size is in the
parted output in a line like:

     /dev/sda:240GB:scsi:512:512:gpt:ATA WDC WDS240G2G0A-:;

where the first 512 is the logical sector size and the second 512 is the physical sector size (not used
in foxclone).

The string is split into a string array using : as the delimiter and the name extracted.

Parameters:
    Device - the drive where the sector size is required, passed in the form sda

Used by:

V1.0 - 17 Dec 2021
-------------------------------------------------------------------------------}
function TSysInfo.getLogSectorSize(Device : string) : integer;
var
  i : byte;
  Index : shortint;
  Fields : TStringDynArray;
begin
  Index:= -1; //failed state
  getLogSectorSize:= 512; //default
  for i:= 0 to PartList.Count - 1 do
      if pos(devstr + Device,PartList[i]) = 1 then
         begin
           Index:= i;
           break;
         end;
  if Index >= 0 then
     begin
       //extract the parameters out of the string
       Fields:= SplitString(PartList[Index],delim);
       if Fields[3] <> '' then getLogSectorSize:= strtoint(Fields[3]);
     end;
end; //end function TSysInfo.getLogSectorSize
{-------------------------------------------------------------------------------
FUNCTION TSYSINFO.GETPTABLETYPE (drive)
Returns the partition table type for the drive specified in Device. It is in the parted output in a line
like:

     /dev/sda:240GB:scsi:512:512:gpt:ATA WDC WDS240G2G0A-:;

where the partition table type is gpt.

Accepted partition table types in foxclone are msdos, gpt, loop and unknown.

The string is split into a string array using : as the delimiter and the ptable type extracted.

Parameters:
    Device - the drive where the sector size is required, passed in the form sda

Used by:

V1.0 - 17 Dec 2021
-------------------------------------------------------------------------------}
function TSysInfo.getPTableType(Device : string) : string;
const
  unknownstr = 'unknown';
var
  i : byte;
  Index : shortint;
  Fields : TStringDynArray;

begin
  Index:= -1; //failed state
  getPTableType:= unknownstr; //failed state
  for i:= 0 to PartList.Count - 1 do
      if pos(devstr + Device,PartList[i]) = 1 then
         begin
           Index:= i;
           break;
         end;
  if Index >= 0 then
     begin
       //extract the parameters out of the string
       Fields:= SplitString(PartList[Index],delim);
       getPTableType:= Fields[5];
     end;
end; //end function TSysInfo.getPTableType
{-------------------------------------------------------------------------------
FUNCTION TSYSINFO.GETSERIAL (drive)
Returns the serial number for the drive specified in Device. It is extracted from the output of:

     pkexec lsblk -no serial,name | grep " sda"

The leading space before the device name is important, without it the output is:

1851BF802454   sda
               ├─sda1
               ├─sda2
               ├─sda3
               ├─sda4
               ├─sda5
               ├─sda6
               └─sda7

Parameters:
    Device - the drive where the serial number is required, passed in the form sda

note - did look at lsblk -iPo name,serial /dev/sda, this still generated output for sda1, sda2, etc.
so would still need grepping.

Used by:

V1.0 - 17 Dec 2021
-------------------------------------------------------------------------------}
function TSysInfo.getSerial(Device : string) : string;
begin
  getSerial:= getstr('pkexec lsblk -no serial,name | grep " ' + Device + '"');
  getSerial:= trim(replacestr(getSerial,Device,''));
end;//end function TSysInfo.getSerial
{-------------------------------------------------------------------------------
FUNCTION TSYSINFO.GETFIRSTPARTITION (drive)
Returns the sector number for the first partition on a drive. It is found by parsing the sfdisk dump
for the drive where the lines for each partition start with:

     /dev/sda1 : start=        2048, size=

All partitions are parsed because the partitions on the drive may not be ordered by part number and
sfdisk may not list the partitions in the order they appear on the drive.

This information is required because on a bootable drive with an msdos partition table, the second
stage bootloader lives in the gap between the end of the partition table (the first sector) and
start of the first partition.

Parameters:
    Device - the drive where the start of the first partition is required, passed in the form sda

Used by:

V1.0 - 17 Dec 2021
-------------------------------------------------------------------------------}
function TSysInfo.getFirstPartition(Device : string) : integer;
const
  startstr = 'start=';
var
  Index : shortint;
  i : byte;
  PartStart : string;
  StartSector : int64;
begin
  //find the part table
  Index:= FindPTable(Device);
  if Index = -1 then //no part table
     begin
       getFirstPartition:= 0;
       exit;
     end;
  getFirstPartition:= 9999;
  for i:= 0 to sfList[Index].Count - 1 do
      if pos(devstr + Device,sfList[Index][i]) = 1 then
         begin
           PartStart:= copy(sfList[Index][i],1,pos(comma,sfList[Index][i])-1); //copy to 1st comma
           PartStart:= trim(copy(PartStart,pos(startstr,PartStart)+6,99));
           StartSector:= StrToInt64(PartStart);
           if StartSector < getFirstPartition then getFirstPartition:= StartSector;
         end;
end; //end function TSysInfo.getFirstPartition
{-------------------------------------------------------------------------------
FUNCTION TSYSINFO.GETDRIVESIZE (drive)
Returns the drive size in blocks for. It is returned as the output from:

      cat /sys/class/block/sda/size

The output is in sectors and has to be converted to blocks (1024 bytes).

Parameters:
    Device - the drive where the size is required, passed in the form sda

Used by:

V1.0 - 17 Dec 2021
-------------------------------------------------------------------------------}
function TSysInfo.getDriveSize(Device : string; SectorSize : integer) : int64;
var
  SizeStr : string;
begin
  getDriveSize:= 0;
  SizeStr:= getStr('cat /sys/class/block/' + Device + '/size');
  if SizeStr <> '' then getDriveSize:= StrToInt64(SizeStr);
  getDriveSize:= SectorToBlock(getDriveSize,SectorSize);
end; //end function TSysInfo.getDriveSize
{-------------------------------------------------------------------------------
FUNCTION TSYSINFO.GETDRIVEUUID (drive)
Returns the UUID of a drive. It is one of the lines in the sfdisk output:

      label-id: 56BB38CB-6756-42F4-B9FA-EDF012B5078C

Parameters:
    Device - the drive where the UUID is required, passed in the form sda

Note - https://www.spinics.net/lists/util-linux-ng/msg13628.html - is this still an issue?
Not used in foxclone.

Will not be reported for a drive with no partition table.

Used by:

V1.0 - 17 Dec 2021
-------------------------------------------------------------------------------}
function TSysInfo.getDriveUUID(Device : string) : string;
const
  labelstr = 'label-id:';
var
  Index : shortint;
  i : byte;
begin
  //find the part table
  Index:= FindPTable(Device);
  if Index = -1 then //no part table
     begin
       getDriveUUID:= '';
       exit;
     end;
  for i:= 0 to sfList[Index].Count - 1 do
      if pos(labelstr,sfList[Index][i]) = 1 then
         begin
           getDriveUUID:= trim(replacestr(sfList[Index][i],labelstr,''));
           break;
         end;
end; //end function TSysInfo.getDriveUUID
{-------------------------------------------------------------------------------
FUNCTION TSYSINFO.GETFIRSTLBA (drive)
Returns the first logical block address on a drive. It is one of the lines in the sfdisk output:

      first-lba: 34

It is only applicable to drives with a gpt partition table and the units are sectors, so 34 = 17k in
on the drive (with 512byte sectors). The partition table on a gpt drive is the first 34 sectors on the
drive.

Parameters:
    Device - the drive where firstLBA is required, passed in the form sda

Used by:

V1.0 - 17 Dec 2021
-------------------------------------------------------------------------------}
function TSysInfo.getFirstLBA(Device : string) : shortint;
const
  firstLBAstr = 'first-lba:';
var
  Index : shortint;
  FirstLBA : string;
  i : byte;
begin
  getFirstLBA:= 0;
  //find the part table
  Index:= FindPTable(Device);
  if Index = -1 then //no part table
     begin
       getFirstLBA:= 0;
       exit;
     end;
  for i:= 0 to sfList[Index].Count - 1 do
      if pos(firstLBAstr,sfList[Index][i]) = 1 then
         begin
           FirstLBA:= trim(replacestr(sfList[Index][i],firstLBAstr,''));
           getFirstLBA:= strtoint(FirstLBA);
           break;
         end;
end; //end function TSysInfo.getFirstLBA
{-------------------------------------------------------------------------------
FUNCTION TSYSINFO.GETLASTLBA (drive)
Returns the last logical block address on a drive. It is one of the lines in the sfdisk output:

      last-lba: 468877278

It is only applicable to drives with a gpt partition table and the units are sectors.
On a gpt drive a backup copy of the partition table is kept at the end of the drive so lastLBA is
typically drive size - 34. Together with firstLBA it defines the space on the drive for partitions.

Parameters:
    Device - the drive where the lastLBA is required, passed in the form sda

Used by:

V1.0 - 17 Dec 2021
-------------------------------------------------------------------------------}
function TSysInfo.getLastLBA(Device : string) : int64;
const
  lastLBAstr = 'last-lba:';
var
  Index : shortint;
  LastLBA : string;
  i : byte;
begin
  getLastLBA:= 0;
  //find the part table
  Index:= FindPTable(Device);
  if Index = -1 then //no part table
     begin
       getLastLBA:= 0;
       exit;
     end;
  for i:= 0 to sfList[Index].Count - 1 do
      if pos(lastLBAstr,sfList[Index][i]) = 1 then
         begin
           LastLBA:= trim(replacestr(sfList[Index][i],lastLBAstr,''));
           getLastLBA:= strtoint64(LastLBA);
           break;
         end;
end; //end function TSysInfo.getLastLBA
{-------------------------------------------------------------------------------
FUNCTION TSYSINFO.FINDPTABLE
A 'helper' function, looks up the drive passed as a parameter in the sfIndex array and returns the
index in the array where it is found.

Index is then used in other functions to access the list of sfdisk output in the TStringList sfList.

Parameters:
    Device - the drive where the index is required, passed in the form sda

Used by:

V1.0 - 17 Dec 2021
-------------------------------------------------------------------------------}
function TSysInfo.FindPTable(Device : string) : shortint;
var
  i : byte;
begin
  FindPTable:= -1; //failed state
  for i:= 0 to length(sfIndex) - 1 do
      if pos(Device,sfIndex[i]) = 1 then
         begin
           FindPTable:= i;
           break;
         end;
end; //end function TSysInfo.FindPTable
{-------------------------------------------------------------------------------
PROCEDURE TSYSINFO.GETPARTS
Finds the sfdisk output for the drive specified by Device and returns a TStringList populated with the
partition names, e.g.

    sdd1
    sdd2
    sdd3

Parameters:
    PartNames - a TStringList populated with the partition names for drive specified
    Device - the drive to return the list of partitions for, e.g. sda

Used by:

V1.0 - 18 Dec 2021
-------------------------------------------------------------------------------}
procedure TSysInfo.GetParts(PartNames : TStringList; Device : string);
var
  Index : shortint;
  i : byte;
  PartStart : string;
begin
  //find the part table
  Index:= FindPTable(Device);
  if Index = -1 then //no part table
     begin
       PartNames.Clear; //empty any existing content
       exit;
     end;
  //find the partitions in the sfdisk output, lines start /dev/sda1 : start=
  for i:= 0 to sfList[Index].Count - 1 do
      if pos(devstr + Device,sfList[Index][i]) = 1 then
         begin
           //strip out the remainder of the line
           PartStart:= trim(copy(sfList[Index][i],1,pos(delim,sfList[Index][i])-1)); //copy to delim
           //strip out /dev/
           PartStart:= replacestr(PartStart,devstr,'');
           PartNames.Add(PartStart);
         end;
end; //end procedure TSysInfo.GetParts
{-------------------------------------------------------------------------------
PROCEDURE TSYSINFO.GETFLAGS (partition)
Returns the flags set on a partition as a text string. The function parses the output from parted -lm

     2:556MB:661MB:105MB:fat32:EFI system partition:boot, esp;
                                                    ^^
Parameters:
    DeviceP - the drive containing the partition, this includes the part prefix, sata device, e.g. sda.
              For a nvme drive it would be like nvme0n1p.
    Number - partition number, but as a text string, e.g. '1'.

Used by:

V1.0 - 24 Dec 2021
-------------------------------------------------------------------------------}
function TSysInfo.getFlags(DeviceP : string; Number : string) : string;
const
  EOLstr = ';'; //terminating character in parted output for a partition at the end of a line
var
  i : byte;
  Index : shortint;
  Fields : TStringDynArray;
begin
  getFlags:= ''; //default
  //find the start of the entry for the drive in the parted output
  Index:= FindDrive(DeviceP);
  if Index > 0 then
     for i:= Index to PartList.Count - 1 do
         begin
           if pos(Number,PartList[i]) = 1 then //have found the partition string
              begin
                Fields:= SplitString(PartList[i],delim);
                getFlags:= replacestr(Fields[6],EOLstr,'');
                exit;
              end;
           if PartList[i] = exitstr then exit; //run out of partitions for this drive
         end;
end; //end function TSysInfo.getFlags
{-------------------------------------------------------------------------------
PROCEDURE TSYSINFO.GETGPTNAME (partition)
Returns the partition name for a partition. The function parses the output from parted -lm

     2:556MB:661MB:105MB:fat32:EFI system partition:boot, esp;
                               ^^
It is only applicable drives with a GPT partition table. Foxclone does nothing with the partition name
other than displaying it in the relevant checklist boxes.

Parameters:
    DeviceP - the drive containing the partition, this includes the part prefix, sata device, e.g. sda.
              For a nvme drive it would be like nvme0n1p.
    Number - partition number, but as a text string, e.g. '1'.

Used by:

V1.0 - 24 Dec 2021
-------------------------------------------------------------------------------}
function TSysInfo.getGPTname(DeviceP : string; Number : string) : string;
var
  i : byte;
  Index : shortint;
  Fields : TStringDynArray;
begin
  getGPTname:= ''; //default
  //find the start of the entry for the drive in the parted output
  Index:= FindDrive(DeviceP);
  if Index > 0 then
     if pos('gpt',PartList[Index]) > 0 then //partition name is only applicable to a gpt part table
        for i:= Index to PartList.Count - 1 do
            begin
              if pos(Number,PartList[i]) = 1 then //have found the partition string
                 begin
                   Fields:= SplitString(PartList[i],delim);
                   getGPTname:= Fields[5];
                   exit;
                 end;
              if PartList[i] = exitstr then exit; //run out of partitions for this drive
            end;
end; //end function TSysInfo.getGPTname
{-------------------------------------------------------------------------------
PROCEDURE TSYSINFO.GETMOUNT (partition)
Returns the existing mount point for a partition if it is already mounted. It runs

   lsblk -iPo uuid,mountpoint /dev/sda4

the output is:

   UUID="fb689910-e942-46e1-b03d-e7032e0e36bb" MOUNTPOINT="/run/timeshift/backup"

Previously had problems with lsblk getting the label (returning empty when there was a label on the
partition). So every use of lsblk returns two parameters, this seemed to fix the problem.

For a partition on a loop device (there is no partition table) foxclone sets the part number = '0'.
This has to be stripped out for lsblk as the partition is defined by the device name, e.g. sda - there
are no partition numbers.

Parameters:
    DeviceP - the drive containing the partition, this includes the part prefix, sata device, e.g. sda.
              For a nvme drive it would be like nvme0n1p.
    Number - partition number, but as a text string, e.g. '1'.

Used by:

V1.0 - 24 Dec 2021
-------------------------------------------------------------------------------}
function TSysInfo.getMount(DeviceP : string; Number : string) : string;
const
  mountstr = 'MOUNTPOINT=';
  dquote = '"';
  swapstr = '[SWAP]';
begin
  if Number = '0' then Number:= ''; //it is a partition on a loop device
  getMount:= getStr('lsblk -iPo uuid,mountpoint /dev/' + DeviceP + Number);
  //fix error where repeating the output on a faulty part - getstr was returning two lines of output
  //both identical?? so strip out everything from the line feed.
  if pos(LF,getMount) > 0 then getMount:= copy(getMount,1,pos(LF,getMount)-1);
  getMount:= copy(getMount,pos(mountstr,getMount),99);
  getMount:= replacestr(getMount,mountstr,'');
  getMount:= replacestr(getMount,dquote,'');
  if getMount = swapstr then getMount:= '';
end; //end function TSysInfo.getMount
{-------------------------------------------------------------------------------
PROCEDURE TSYSINFO.GETPSTART (partition)
Returns the start sector for a partition. It parses the output from sfdisk (partition table dump).
sfdisk output for a partition is:

    /dev/sde2 : start=   291317760, size=   333824000, type=5

For a partition on a loop device there is no partition table, so the function returns zero. With no
partition table, the first (and only) partition on a loop device starts at sector 0.

Parameters:
    DeviceP - the drive containing the partition, this includes the part prefix, sata device, e.g. sda.
              For a nvme drive it would be like nvme0n1p.
    Number - partition number, but as a text string, e.g. '1'.

Used by:

V1.0 - 24 Dec 2021
V1.1 - 23 Oct 2023 problem on a drive with more than 9 partitions, the test on sda1 would also match
                   sda10, throwing an error because when it parses the line for sda10 which looks like:
                        /dev/sda10 : start=   302692352
                   it strips out sda1 which leaves
                        0 : start=   302692352
                   which subsequently is parsed to
                       0    302692352
                   which generates an error when converting to int.
                   Fixed by testing on sda1 + space. A break added to the for loop.
-------------------------------------------------------------------------------}
function TSysInfo.getPStart(DeviceP : string; Number : string) : int64;
const
  space = ' ';
  startstr = ': start=';
var
  Index : shortint;
  i : byte;
  PartStart : string;
  Device : string;
begin
  //find the part table
  Device:= replacestr(DeviceP,'p','');
  Index:= FindPTable(Device);
  if Index = -1 then //no part table
     begin
       getPStart:= 0; //empty any existing content
       exit;
     end;
  //find the partition in the sfdisk output, lines start /dev/sda1 : start=
  for i:= 0 to sfList[Index].Count - 1 do
      if pos(devstr + DeviceP + Number + space,sfList[Index][i]) = 1 then
         begin
           //copy to the comma
           PartStart:= copy(sfList[Index][i],1,pos(comma,sfList[Index][i])-1);
           PartStart:= replacestr(PartStart,devstr+DeviceP+Number,''); //strip out the device
           PartStart:= trim(replacestr(PartStart,startstr,'')); //strip out the start string
           getPStart:= StrToInt64(PartStart);
           break;
         end;
end; //end function TSysInfo.getPStart
{-------------------------------------------------------------------------------
PROCEDURE TSYSINFO.GETPSIZE (partition)
Returns the size of a partition in 1k blocks. It parses the output from sfdisk (partition table dump).
sfdisk output for a partition is:

    /dev/sde2 : start=   291317760, size=   333824000, type=5

Size shown in sfdisk is in sectors so this is converted to blocks once extracted from the string.

For a partition on a loop device there is no partition table, so the function returns zero. With no
partition table, the partition size is the same as the device. The size of a partition on loop device
must be determined external to this function.

Parameters:
    DeviceP - the drive containing the partition, this includes the part prefix, sata device, e.g. sda.
              For a nvme drive it would be like nvme0n1p.
    Number - partition number, but as a text string, e.g. '1'.
    SectorSize - sector size in bytes, used to convert sectors to blocks.

Used by:

V1.0 - 24 Dec 2021
V1.1 - 23 Oct 2023 see getPStart for error, with more than 9 partitions, will match on sda1 and sda10.
                   Space added to get match correct.
-------------------------------------------------------------------------------}
function TSysInfo.getPsize(DeviceP : string; Number : string; SectorSize : integer) : int64;
const
  sizestr = 'size=';
  space = ' ';
var
  Index : shortint;
  i : byte;
  PartSize : string;
  Device : string;
begin
  //find the part table
  Device:= replacestr(DeviceP,'p','');
  Index:= FindPTable(Device);
  if Index = -1 then //no part table
     begin
       getPsize:= 0; //empty any existing content
       exit;
     end;
  //find the partition in the sfdisk output, lines start /dev/sda1 : start=
  for i:= 0 to sfList[Index].Count - 1 do
      if pos(devstr + DeviceP + Number + space,sfList[Index][i]) = 1 then
         begin
           //copy from the 'size='
           PartSize:= copy(sfList[Index][i],pos(sizestr,sfList[Index][i]),99);
           PartSize:= copy(PartSize,1,pos(comma,PartSize)-1); //strip out everything after comma
           PartSize:= trim(replacestr(PartSize,sizestr,'')); //strip out the size str
           getPsize:= SectorToBlock(StrToInt64(PartSize),SectorSize);
           break;
         end;
end; //end procedure TSysInfo.getPsize
{-------------------------------------------------------------------------------
PROCEDURE TSYSINFO.GETPTYPE (partition)
This is only applicable to partitions on a drive with a msdos/legacy partition table, where the types
may be:

    primary
    extended
    logical

For a GPT partition table, all partitions are primary partitions.

Unfortunately, the machine-readable output from parted -lm does not show the partition type.
Parted -l must be run, meaning that parted has to be run twice, with the parted -l output stored in
the FullPartList tstringlist. Output is of the form:

    Number  Start   End    Size   Type      File system  Flags
     1      1049kB  149GB  149GB  primary   ext4
     2      149GB   320GB  171GB  extended
     5      149GB   320GB  171GB  logical   ext4

Note the leading space on the partition number when less than 10.

Parameters:
    DeviceP - the drive containing the partition, this includes the part prefix, sata device, e.g. sda.
              For a nvme drive it would be like nvme0n1p.
    Number - partition number, but as a text string, e.g. '1'.

Used by:

V1.0 - 24 Dec 2021
-------------------------------------------------------------------------------}
function TSysInfo.getPtype(DeviceP : string; Number : string) : string;
const
  Diskdevstr = 'Disk /dev/';
  numstr = 'Number';
  typestr = 'Type';
  msdosstr = 'msdos';
  primarystr = 'primary';
var
  i : byte;
  Index : shortint = -1;
  TypeCol : byte = 0;
  Device : string;
begin
  Device:= replacestr(DeviceP,'p','');
  if getPTableType(Device) <> msdosstr then
     begin
       getPtype:= primarystr;
       exit;
     end;
  //find the start of the entry for the drive in the parted output begins Disk /dev/sdb
  for i:= 0 to FullPartList.Count - 1 do
      if pos(Diskdevstr + Device,FullPartList[i]) = 1 then
         begin
           Index:= i;
           break;
         end;
  if Index = -1 then exit;
  //now find the heading line, need to find the column for "Type"
  for i:= Index+1 to FullPartList.Count - 1 do //note skipping the index line
      begin
        if pos(numstr,FullPartList[i]) = 1 then
           begin
             TypeCol:= pos(typestr,FullPartList[i]);
             Index:= i; //to avoid testing the lines again
             break;
           end;
        if pos(Diskdevstr,FullPartList[i]) = 1 then exit; //found the next drive
      end;
  //need to pad out the number if part number is less than 10
  if strtoint(Number) < 10 then Number:= Space + Number;
  //now extract the partition type
  for i:= Index to FullPartList.Count - 1 do //find the partition
      if pos(Number,FullPartList[i]) = 1 then
         begin
           getPtype:= copy(FullPartList[i],TypeCol,99);
           if pos(Space,getPtype) > 0 then getPtype:= trim(copy(getPtype,1,pos(space,getPtype)))
           else getPtype:= trim(getPtype);
           exit;
         end;
end; //end function TSysInfo.getPtype
{-------------------------------------------------------------------------------
PROCEDURE TSYSINFO.GETUSEDBLKS (partition)
Returns the used blocks on a partition of a known type.

The partition must be mounted for the function to return a valid result. It parses the output from df:

    Filesystem                  1K-blocks       Used Available Use% Mounted on
    udev                          7997180          0   7997180   0% /dev
    tmpfs                         1608956       3472   1605484   1% /run
    /dev/sdd1                    32549628   23031308   7841832  75% /

If the partition cannot be found, the function returns the size of the partition.

Parameters:
    DeviceP - the drive containing the partition, this includes the part prefix, sata device, e.g. sda.
              For a nvme drive it would be like nvme0n1p.
    Number - partition number, but as a text string, e.g. '1'.
    SectorSize - sector size in bytes, used to convert sectors to blocks and needed in the call to
                 getPsize.

Used by:

V1.0 - 24 Dec 2021
-------------------------------------------------------------------------------}
function TSysInfo.getUsedBlks(DeviceP : string; Number : string; SectorSize : integer) : int64;
var
  dfOutput : string;
begin
  //check for a partition on a loop device
  if Number = '0' then Number:= ''; //for a partition on a loop device
  dfOutput:= GetStr('df | grep ^/dev/' + DeviceP + Number);
  getUsedBlks:= 0;
  //this will need amended for a part on a loop device where part number = '0'
  dfOutput:= trim(replacestr(dfOutput,devstr + DeviceP + Number,''));
  //first number left is size in 1k blocks get it just in case
  dfOutput:= trim(copy(dfOutput,pos(space,dfOutput),99));
  //first number is now used blocks
  dfOutput:= trim(copy(dfOutput,1,pos(space,dfOutput)-1));
  //if nothing then an error
  if dfOutput = '' then
     begin
       FoxMsg(rWarningStr,[rUsedErrStr +DeviceP + Number + rUsedErrStr3,rUsedErrStr2],mtWarning,mbOK);
       getUsedBlks:= getPsize(DeviceP,Number,SectorSize);
     end
  else getUsedBlks:= strtoint64(dfOutput);
end; //end function TSysInfo.getUsedBlks
{-------------------------------------------------------------------------------
PROCEDURE TSYSINFO.GETPSFTYPE (partition)
Returns the partition type from the sfdisk output. Output from sfdisk is of the form:

For a msdos/legacy partition table:
/dev/sde1 : start=        2048, size=   291315712, type=83

For a gpt partition table:
/dev/sda5 : start=    79755264, size=    30720000, type=0FC63DAF-8483-4772-8E79-3D69D8477DE4, uuid=5E0AED86-3EC9-47F5-A8A0-6615C20E1F61

Recorded but not used in foxclone.

Parameters:
    DeviceP - the drive containing the partition, this includes the part prefix, sata device, e.g. sda.
              For a nvme drive it would be like nvme0n1p.
    Number - partition number, but as a text string, e.g. '1'.

Used by:

V1.0 - 24 Dec 2021
V1.1 - 23 Oct 2023 see getPStart for error, testing will match sda1 and sda10 on a drive with more than
                   9 partitions. Spaced added to criteria.
-------------------------------------------------------------------------------}
function TSysInfo.getPSFtype(DeviceP : string; Number : string) : string;
const
  typestr = 'type=';
  space = ' ';
var
  Index : shortint;
  i : byte;
  Device : string;
begin
  //find the part table
  getPSFtype:= '';
  Device:= replacestr(DeviceP,'p','');
  Index:= FindPTable(Device);
  if Index = -1 then exit; //no part table
  //find the partition in the sfdisk output, lines start /dev/sda1 : start=
  for i:= 0 to sfList[Index].Count - 1 do
      if pos(devstr + DeviceP + Number + space,sfList[Index][i]) = 1 then
         begin
           //copy from the 'type='
           getPSFtype:= copy(sfList[Index][i],pos(typestr,sfList[Index][i]),99);
           //strip out everything after comma, not present on sfdisk output on a legacy part table
           if pos(comma,getPSFtype) > 0 then getPSFtype:= copy(getPSFtype,1,pos(comma,getPSFtype)-1);
           getPSFtype:= trim(replacestr(getPSFtype,typestr,'')); //strip out the type str
           break;
         end;
end; //end procedure TSysInfo.getPSFtype
{-------------------------------------------------------------------------------
PROCEDURE TSYSINFO.GETFS (partition)
Returns the filesystem on the partition. It parses the parted -lm output to extract the filesystem
where the information is of the form:

    1:1049kB:34.0GB:34.0GB:ext4::boot;
                           ^^
Although partclone supports a number of apple filesystems, took a decision to not support them in
foxclone. Apple is closed source and makes it difficult/impossible for linux to communicate with
apple devices, so why should linux support it. Support for reiserfs also removed (read about the dev).

Parameters:
    DeviceP - the drive containing the partition, this includes the part prefix, sata device, e.g. sda.
              For a nvme drive it would be like nvme0n1p.
    Number - partition number, but as a text string, e.g. '1'.

When checking for exfat, blkid cmd returns
     andy@T432 ~ $ blkid | grep /dev/sde1
     /dev/sde1: LABEL="Ventoy" UUID="3B73-CC18" TYPE="exfat" PTTYPE="dos" PARTUUID="2fa36df4-01"
Checking with lsblk
     andy@T432 ~ $ lsblk -iPo fstype,name | grep NAME=\"sdb2\"
     FSTYPE="exfat" NAME="sdb2"

Used by:

V1.0 - 24 Dec 2021
V1.1 - 03 Jan 2022, parted does not support exfat, but other utilities do, so if parted does not
                    recognise the filesystem check with blkid to find out if it is exfat.
V1.2 - 10 Jan 2022, problems with blkid when used in ventoy. It will not return output even when used
                    with pkexec. Run in a terminal, first time no output, run with pkexec or sudo then
                    output, then in foxclone okay. Changed to lsblk to get filesystem for exfat.
V1.3 - 11 Jan 2022, no support for fat16, old/obsolete and only appears with ventoy, causes a message as
                    when it cannot mount the fat16 part to get used space.
V1.4 - 28 Jan 2022, booting from a usb stick (not ventoy) foxclone trying to read the efi part and
                    not able to mount (efi part will be already mounted), fs will be empty in parted
                    so have to trap it in lsblk and set unknown.
-------------------------------------------------------------------------------}
function TSysInfo.getFS(DeviceP : string; Number : string) : string;
const
  SwapStr = 'swap'; //swap partition
  UnknownStr = 'unknown';
  Fat32Str = 'fat32';
  Fat16Str = 'fat16'; //remove support for fat16, used on ventoy usb
  FatStr = 'vfat';
  ReiserStr = 'reiserfs'; //unsupported filesystem
  HfsStr = 'hfs'; //unsupported filesystem
  HfsPlusStr = 'hfsplus'; //unsupported filesystem
  lsblkstr = 'lsblk -iPo fstype,name | grep NAME=\"[ref]\"';
  refstr = '[ref]';
  fstypestr = 'FSTYPE="';
var
  i : byte;
  Index : shortint;
  Fields : TStringDynArray;
begin
  getFS:= UnknownStr; //default

  //writeln('get fs called...');
  if Number = '0' then Number:= '1'; //for a loop device
  //find the start of the entry for the drive in the parted output
  Index:= FindDrive(DeviceP);
  if Index > 0 then
     for i:= Index to PartList.Count - 1 do
         begin
           if pos(Number,PartList[i]) = 1 then //have found the partition string
              begin
                Fields:= SplitString(PartList[i],delim);
                getFS:= Fields[4];

                //writeln('fiels 4 =',getFS);

                case getFS of //decided not to support reiserfs or apple fs, saves a few MB in the iso
                     ''         : begin //it might be exfat, check with lsblk
                                    getFS:= getstr(replacestr(lsblkstr,refstr,DeviceP + Number));
                                    getFS:= replacestr(getFS,fstypestr,'');
                                    getFS:= copy(getFS,1,pos('"',getFS)-1);
                                    //if still empty or vfat set unknown
                                    if (getFS = '') or (getFS = 'vfat') then getFS:= UnknownStr;
                                  end;
                     FatStr     : getFS:= Fat32Str;
                     Fat16Str   : getFS:= UnknownStr;
                     ReiserStr  : getFS:= UnknownStr;
                     HfsStr     : getFS:= UnknownStr;
                     HfsPlusStr : getFS:= UnknownStr;
                end;
                if pos(SwapStr,getFS) > 0 then getFS:= Swapstr; //e.g. linux-swap(v1)
                exit;
              end;
           if PartList[i] = exitstr then exit; //run out of partitions for this drive
         end;
end; //end function TSysInfo.getFS
{-------------------------------------------------------------------------------
PROCEDURE TSYSINFO.GETLABEL (partition)
Returns the label for a partition if one exists. It runs

   lsblk -iPo uuid,label /dev/sda4

the output is:

   UUID="bfe48857-036a-4e87-a62c-7e45e90da204" LABEL="home"

Previously had problems with lsblk getting the label (returning empty when there was a label on the
partition). So every use of lsblk returns two parameters, this seemed to fix the problem.

For a partition on a loop device (there is no partition table) foxclone sets the part number = '0'.
This has to be stripped out for lsblk as the partition is defined by the device name, e.g. sda - there
are no partition numbers.

Parameters:
    DeviceP - the drive containing the partition, this includes the part prefix, sata device, e.g. sda.
              For a nvme drive it would be like nvme0n1p.
    Number - partition number, but as a text string, e.g. '1'.

Used by:

V1.0 - 24 Dec 2021
-------------------------------------------------------------------------------}
function TSysInfo.getLabel(DeviceP : string; Number : string) : string;
const
  labelstr = 'LABEL=';
  dquote = '"';
begin
  if Number = '0' then Number:= ''; //for a partition on a loop device
  getLabel:= getStr('lsblk -iPo uuid,label /dev/' + DeviceP + Number);
  //fix error where repeating the output on a faulty part - it generated two lines of output
  if pos(LF,getLabel) > 0 then getLabel:= copy(getLabel,1,pos(LF,getLabel)-1);
  getLabel:= copy(getLabel,pos(labelstr,getLabel),99);
  getLabel:= replacestr(getLabel,labelstr,'');
  getLabel:= replacestr(getLabel,dquote,'');
end; //end function TSysInfo.getLabel
{-------------------------------------------------------------------------------
PROCEDURE TSYSINFO.GETUUID (partition)
Returns the UUID for a partition. It runs

   lsblk -iPo label,uuid /dev/sda4

the output is:

   LABEL="" UUID="5250141E50140AFD"

Previously had problems with lsblk getting the label (returning empty when there was a label on the
partition). So every use of lsblk returns two parameters, this seemed to fix the problem.

For a partition on a loop device (there is no partition table) foxclone sets the part number = '0'.
This has to be stripped out for lsblk as the partition is defined by the device name, e.g. sda - there
are no partition numbers.

Parameters:
    DeviceP - the drive containing the partition, this includes the part prefix, sata device, e.g. sda.
              For a nvme drive it would be like nvme0n1p.
    Number - partition number, but as a text string, e.g. '1'.

Used by:

V1.0 - 24 Dec 2021
-------------------------------------------------------------------------------}
function TSysInfo.getUUID(DeviceP : string; Number : string) : string;
const
  UUIDstr = 'UUID=';
  dquote = '"';
begin
  if Number = '0' then Number:= ''; //for a partition on a loop device
  getUUID:= getStr('lsblk -iPo label,uuid /dev/' + DeviceP + Number);
  //fix error where repeating the output on a faulty part
  if pos(LF,getUUID) > 0 then getUUID:= copy(getUUID,1,pos(LF,getUUID)-1);
  getUUID:= copy(getUUID,pos(UUIDstr,getUUID),99);
  getUUID:= replacestr(getUUID,UUIDstr,'');
  getUUID:= replacestr(getUUID,dquote,'');
end; //end function TSysInfo.getUUID
{-------------------------------------------------------------------------------
PROCEDURE TSYSINFO.GETBOOTFLAG (partition)
Only applicable to drives with a msdos/legacy partition table. Output from sfdisk is of the form:

     /dev/sdc1 : start=  3863818240, size=    40960938, type=83, bootable

Recorded but not used directly in foxclone.

Parameters:
    DeviceP - the drive containing the partition, this includes the part prefix, sata device, e.g. sda.
              For a nvme drive it would be like nvme0n1p.
    Number - partition number, but as a text string, e.g. '1'.

Used by:

V1.0 - 24 Dec 2021
V1.1 - 23 Oct 2023 see getPStart for error, search criteria will match sda1 and sda10, space added.
-------------------------------------------------------------------------------}
function TSysInfo.getBootFlag(DeviceP : string; Number : string) : string;
const
  labelstr = 'label: gpt'; //part table type in sfdisk output
  flagstr = 'bootable';
  space = ' ';
var
  Index : shortint;
  i : byte;
  Device : string;
begin
  //find the part table
  getBootFlag:= '';
  Device:= replacestr(DeviceP,'p','');
  Index:= FindPTable(Device);
  if Index = -1 then exit; //no part table
  //find the partition in the sfdisk output, lines start /dev/sda1 : start=
  for i:= 0 to sfList[Index].Count - 1 do
      begin
        if pos(labelstr,sfList[Index][i]) = 1 then exit; //flag is only applicable to legacy part table
        if pos(devstr + DeviceP + Number + space,sfList[Index][i]) = 1 then
           if pos(flagstr,sfList[Index][i]) > 0 then
              begin
                getBootFlag:= flagstr;
                break;
              end;
      end;
end; //end procedure TSysInfo.getBootFlag
{-------------------------------------------------------------------------------
PROCEDURE TSYSINFO.FINDDRIVE
A helper function, searches the output from parted -lm for the drive specified and returns the index
of the line in the parted output where it occurs.

Parameters:
    DeviceP - the drive containing the partition, this includes the part prefix, sata device, e.g. sda.
              For a nvme drive it would be like nvme0n1p.

Used by:

V1.0 - 24 Dec 2021
-------------------------------------------------------------------------------}
function TSysInfo.FindDrive(DeviceP: string) : shortint;
var
  Device : string;
  i : byte;
begin
  Device:= replacestr(DeviceP,'p',''); //strip out the part prefix if present
  FindDrive:= -1; //fault condition
  for i:= 0 to PartList.Count - 1 do
      if pos(devstr + Device,PartList[i]) = 1 then
         begin
           FindDrive:= i;
           break;
         end;
end; //end function TSysInfo.FindDrive
{-------------------------------------------------------------------------------
PROCEDURE TSYSINFO.COPYPARTEDLM
Used by debug, the procedure adds the output from parted -lm (the machine readable version) to the
TStringList passed as a parameter.

Parameters:
    LogList - the TStringList to add the output of parted -lm to.

Used by:

V1.0 - 29 Dec 2021
-------------------------------------------------------------------------------}
procedure TSysInfo.CopyPartedLM(LogList : TStringList);
var
  i : byte;
begin
  for i:= 0 to PartList.Count - 1 do
      if PartList[i] <> '' then LogList.Add(PartList[i]);
end; //end procedure TSysInfo.CopyPartedLM
{-------------------------------------------------------------------------------
PROCEDURE TSYSINFO.COPYPARTED
Used by debug, the procedure adds the output from parted --list (the human readable version) to the
TStringList passed as a parameter. Needed as this is the only source of info on whether a partition is
extended or logical.

Parameters:
    LogList - the TStringList to add the output of parted -l to.

Used by:

V1.0 - 29 Dec 2021
-------------------------------------------------------------------------------}
procedure TSysInfo.CopyParted(LogList : TStringList);
var
  i : byte;
begin
  for i:= 0 to FullPartList.Count - 1 do LogList.Add(FullPartList[i]);
end; //end procedure TSysInfo.CopyParted
{-------------------------------------------------------------------------------
PROCEDURE TSYSINFO.COPYPTABLES
Used by debug, the procedure adds the output all the sfdisk dump output on each drive - this is a
text file that defines the partition table. Note - no partition table, e.g. a loop device, then no
partition table output from sfdisk.

Parameters:
    LogList - the TStringList to add the output of sfdisk --dump to.

Used by:

V1.0 - 29 Dec 2021
-------------------------------------------------------------------------------}
procedure TSysInfo.CopyPtables(LogList : TStringList);
var
  i,j : byte;
begin
  for i:= 0 to sfList.Count - 1 do
      begin
        for j:= 0 to sfList[i].Count - 1 do if sfList[i][j] <> '' then LogList.Add(sfList[i][j]);
        LogList.Add(Separator2);
      end;
end; //end procedure TSysInfo.CopyPtables
end.

