{ FoxClone
  Copyright (C) 2021 Andy Hardwick
  This source is free software; you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free
  Software Foundation; either version 2 of the License, or (at your option)
  any later version.
  This code is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
  details.
  A copy of the GNU General Public License is available on the World Wide Web
  at <http://www.gnu.org/copyleft/gpl.html>. You can also obtain it by writing
  to the Free Software Foundation, Inc., 51 Franklin Street - Fifth Floor,
  Boston, MA 02110-1335, USA.
}
{-------------------------------------------------------------------------------
UNIT - PARTITIONS

The data structures used in foxclone to describe a system are:

    TSystem - a TObjectList of all the drives in a PC.

    TDrive - one for each drive in the system, defines the data held on a drive, e.g. name, size,
    partition table type and the methods to access/change those data.

    TPartList - one of the data attributes of a TDrive is a TPartList, a list of all the partitions
    on a drive.

    TPart - one for each partition on a drive, defines the data held on a partition, e.g. size,
    file system, label, used space and the methods to access/change those data.

This unit defines the data structures and methods for a TPartList and TPart.

V1.0 - 06 Mar 2021
V1.1 - 26 Apr 2024 did have unsupporteFS defined here, but moved to SupportedFS in utilities, better
                   to check for supported filesystems than unsupported.
-------------------------------------------------------------------------------}
unit partitions;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, Contnrs, Dialogs, process, Types, utilities, StrUtils, il8nstrings, settings,
  fdialog, info;

type
  TPart = class(TObject)
   // const
    //  unsupportedFS : string = 'unknown.crypto_LUKS.LVM2_member.zfs.zfs_member';

    private
      pBoot : boolean;     //is the partition mounted as /
      pDrive : string; //partition ref of the parent drive with prefix, e.g. sda or nvme0n1p
      pNumber  : string;  //a string but just a number, e.g. '1', note '0' for a part on a loop device
      pMountpt : string;   //where it mounts
      pStart   : Int64;    //where the partition starts in sectors
      pSector : integer; //sector size in bytes, copied from the parent drive
      pSize    : Int64;   //size of partition in blocks
      pUsed    : Int64;   //used blocks in partition
      pType    : string;  //msdos only, e.g. primary, extended, logical
      psfType  : string;  //as reported by sfdisk in the Type= output
      fs       : string;  //e.g. ntfs, ext4, or not = unknown if not recognised
      pName    : string;  //gpt only
      pLabel   : string;  //if it has one
      OS       : string;  //operating system populated by OSprober
      flags    : string;  //e.g. boot, msftdata
      sfFlags  : string;  //as reported by sfdisk
      UUID     : string;  //needed for swap parts
      pSelected : boolean; //selected partition to operate on, e.g. backup, restore
      EFIpart : boolean;  //set if this is an EFI partition, EFI parts are always backed up
    public
      property OpSys : string read OS write OS;
      property boot : boolean read pboot write pboot;
      property FileSystem : string read fs;
      property PartType : string read ptype;
      property selected : boolean read pSelected write pSelected;
      property EFIpartition : boolean read EFIpart;
      property Used : Int64 read pused;
      property Size : Int64 read pSize write pSize;
      property Drive : string read pDrive write pDrive;
      property Number : string read pNumber write pNumber;
      property StartSector : int64 read pStart;
      property SectorSize : integer read pSector write pSector;
      property PartMountPt : string read pMountpt;
      property PartLabel : string read plabel;
      property Pflags : string read flags; //needed to find if bios_grub partition 27/9/23

      constructor create;
      procedure getPartInfo;
      procedure Print(var OutF : Textfile);
      function ListStr : string; //a string for part checklistboxes
      function ShortStr : string; //a strong for the progress form
      function HintStr(isSource : boolean) : string; //hint string for part checklistboxes
      function UserSelect : boolean;
      function UMount : boolean;
      function Mount(ReadWrite : boolean; mountpoint : string = '') : string;
      function FoxMount : boolean; //returns true if partition mounted by foxclone
      procedure ParseSFdisk(sfPartStr : string; dSector : integer);
      procedure ParseBackFile(PTableType,PartStr, BlockStr : string);
      function BackStr(BackFname : String) : string;
      function BackFileStr(PartTable : string) : string;
      function RestoreStr(BackFname : String; DestDevice : string) : string;
      function MakeSwap(DestDrive : string) : boolean;
      function CloneStr(DestDrive : string) : string;
      function VerifyStr(BackFname : String) : string;
  end;

  TPartList = class(TObjectList)
    private
      function GetItems(aIndex: Integer) : TPart;
      procedure SetItems(aIndex: Integer; aPart : TPart);
    public
      function FindName(PartStr : string): integer;
      property Items[Index: Integer]: TPart read GetItems write SetItems; default;
  end;
var
  Partition: TPart;

implementation
{-------------------------------------------------------------------------------
FUNCTION TPARTS.GETITEMS
A wrapper so that instead of accessing an element in a TDrive with:
  TDrive(MyDrives[i]).mountpt)
It can be accessed with:
  MyDrives[i].mountpt

It replaces the default TObjectList.Items[Index: Integer] property of a TObjectList

Used by:

V1.0 - 24 Feb 2021
-------------------------------------------------------------------------------}
function TPartList.GetItems(aIndex: Integer) : TPart;
begin
  result:= TPart(get(aIndex));
end; //end function TPartList.GetItems
{-------------------------------------------------------------------------------
PROCEDURE TPARTS.SETITEMS
A wrapper so that instead of accessing an element in a TDrive with:
  TDrive(MyDrives[i]).mountpt)
It can be accessed with:
  MyDrives[i].mountpt

It replaces the default TObjectList.Items[Index: Integer] property of a TObjectList

  Used by:

V1.0 - 24 Feb 2021
-------------------------------------------------------------------------------}
procedure TPartList.SetItems(aIndex: Integer; aPart: TPart);
begin
  put(aIndex,aPart);
end; //end procedure TPartList.SetItems
{-------------------------------------------------------------------------------
FUNCTION TPARTLIST.FINDBYNAME
Seaches the system to find the matching partition.

Parameters:
    PartStr : string of the form 'sda1 and other text' - this is the subject of the search

The function searches for a match = pnumber in PartStr. It returns the index of the drive or if not
found returns -1.

Only a small number of partitions, so searching from start to finish not a big overhead.

Used by:

V1.0 - 27 Feb 2021
V1.1 - 10 Apr 2021, i changed from byte to shortint
V1.1 - 24 Oct 2023, on a drive with +10 partitions searching on sda1 (PartRef) in a string containing
                    sda10 will match and cause errors downstream.
                    Adding a space to partref to fix problem.
-------------------------------------------------------------------------------}
function TPartList.FindName(PartStr : string) : integer;
const
  space = ' ';
var
  i : shortint;
  found : boolean = false;
  PartRef : string;
begin
  for i:= 0 to self.Count - 1 do
      begin
        PartRef:= self[i].pdrive + self[i].pnumber + space;
        if pos(PartRef,PartStr) = 1 then
           begin
             found:= true;
             FindName:= i;
             break;
           end;
      end;
  if not found then FindName:= -1;
end; //end TPartList.FindbyName
{-------------------------------------------------------------------------------
CONSTRUCTOR TPART.CREATE
Simply initialises all the variables in TPart.

Used by:

V1.0 - 24 Feb 2021
V1.1 - 07 Mar 2021 - modified as a result of changing TPart definition to include all information
                    from parted and sfdisk
V1.2 - 07 Mar 2021 - BlkAdjust added
V1.3 - 29 Mar 2021 - foxmount added
-------------------------------------------------------------------------------}
constructor TPart.create;
begin
  inherited create;
  pBoot:= false;
  pDrive:= '';
  pNumber:= '';
  pMountpt:= '';
  pStart:= 0;
  pSize:= 0;
  pUsed:= 0;
  pSector:= 0;
  pType:= 'primary';
  psfType:= '';
  fs:= '';
  pName:= '';
  pLabel:= '';
  OS:= '';
  Flags:= '';
  sfFlags:= '';
  UUID:= '';
  pSelected:= false;
  EFIpart:= false;
end; //end constructor TPart.create
{-------------------------------------------------------------------------------
PROCEDURE TPART.GETPARTINFO
Adds partition information to a TPart.

Rewritten to use the info unit and the SysInfo object. GetPartInfo populates a TPart by calling the
methods in SysInfo. For information that is parsed from the output of parted or sfdisk, this
information is already in memory as TStringLists, so a lot faster.

Used by:

V1.0 - 27 Feb 2021
V1.1 - 04 Mar 2021 modified to take account of a partition on a loop device, i.e. created by disks
V1.2 - 06 Mar 2021 modified to use all info from sfdisk and change default unit to sectors NOT blocks
V1.3 - 07 Mar 2021 df reports used as 1k blocks, converted to sectors
                   fix to get size on a loop device - sfdisk does not generate output
V1.4 - 23 Mar 2021 changed so that pused = 0 on an extended partition so that when calculating the
                   space used for backup, the extended part doesn't double count
V1.5 - 29 Mar 2021 foxmount removed as local var, now part of the TPart definition and used elsewhere
V1.6 - 02 Apr 2021 foxmount removed, now checks that a part is mounted in the default mount folder
V1.7 - 20 Oct 2021 comments in header changed to clarify if sectors or blocks on psize and pused.
                   They are sectors.
V1.8 - 28 Oct 2021 psize and pused are now in blocks and blkadjust removed. psize is determined from
                   sfdisk so size is sectors, converted to blocks.
V1.9 - 09 Nov 2021 Problem with the part label, sure it was working. Uses lsblk -nio label /dev/sdb1
                   this is returning an empty string, but test in a terminal and it works?? Also
                   lsblk is also used to get the UUID and mountpoint and these work okay. No idea what
                   the problem is?? Changed to get label, UUID and mountpoint in one hit and this seems
                   to be working. See comments in code
V1.10 - 11 Nov 2021 more problems with lsblk. No UUIDs or labels for sda, but working on other drives.
                    Found that if just ONE partition was mounted on sda then it worked (partitions are
                    mounted on the other drives in the system). Again working in a terminal not in
                    foxclone. Finally tried executing lsblk with pkexec and it worked (had to define a
                    policy for lsblk). No idea why, maybe because termcmd runs commands with bash?
                    Suspect this problem may not arise in the production version as foxclone will be
                    running as root to avoid write issues on partitions not owned by the user.
V1.11 - 14 Nov 2021 on a drive that is screwed up, e.g. ptable restored but unsuccessful, then mounting
                    partitions on that drive to get used space will fail, but this is not being trapped.
                    Fixed by testing the output from df for an empty string.
V1.12 - 29 Dec 2021 completely rewritten.
V1.13 - 30 Dec 2021 DefaultFolder instead of MountFolder, it was leaving all parts mounted after getting
                    used blocks
V1.14 - 24 Jan 2022 When booting from a Ventoy stick, the iso is in an exFAT partition which Ventoy
                    mounts. Trying to mount the partition to get used space results in an error.
                    Now sets the filesystem to unknown if the label = Ventoy.
-------------------------------------------------------------------------------}
procedure TPart.getPartInfo;
const
  msresstr = 'Microsoft reserved partition';
  unknownstr = 'unknown';
  swapstr = 'swap'; //swap partition
  Ventoystr = 'Ventoy';
var
  UsedMount : string; //the mount point used to determne used blocks
begin
  //flags
  flags:= SysInfo.getFlags(pdrive,pnumber);
  //writeln(pDrive + pnumber,' Flags=',flags);
  //EFIpart set if this is an EFI partition
  if (pos('esp',flags) > 0) and (pos('boot',flags) > 0) then EFIpart:= true;
  //pName
  pName:= SysInfo.getGPTname(pdrive,pnumber);
  //pLabel if it has one - was getting it later, now earlier so can check for ventoy
  pLabel:= SysInfo.getLabel(pdrive,pnumber);
  //writeln(pDrive + pnumber,' partname=',pName);
  //pMountpt where it mounts if already mounted before running foxclone
  pMountpt:= SysInfo.getMount(pdrive,pnumber);
  //writeln(pDrive + pnumber,' mount pt=',pMountpt);
  //pBoot
  if pMountpt = '/' then pBoot:= true;
  //pStart where the partition starts in sectors
  pStart:= SysInfo.getPStart(pdrive,pnumber);
  //writeln(pDrive + pnumber,' partstart=',pStart);
  //pSize size of partition in blocks
  if pNumber <> '0' then pSize:= SysInfo.getPsize(pDrive,pNumber,pSector); //check for loop device
  //writeln(pDrive + pnumber,' part size=',pSize,space,SizeStr(pSize));
  //pType  msdos only, e.g. primary, extended, logical. All parts are primary on a gpt drive
  pType:= SysInfo.getPtype(pdrive,pnumber);
  //writeln(pDrive + pnumber,' part type=',pType);
  //psfType as reported by sfdisk, e.g. type=DE94BBA4-06D1-4D40-A16A-BFD50179D6AC
  psfType:= SysInfo.getPSFtype(pdrive,pnumber);
  //writeln(pDrive + pnumber,' sfdisk part type=',psfType);
  //fs filesystem e.g. ntfs, ext4, or not = unknown if not recognised
  fs:= SysInfo.getFS(pdrive,pnumber);
  if pName = msresstr then fs:= unknownstr; //fix issue found on desktop
  //writeln(pDrive + pnumber,' filesystem=',fs);
  //pUsed used sectors in partition, can use tune2fs for linux fs, but get from df when mounted
  //so a common approach for all filesystems.
  //do not want to try and mount the exfat Ventoy partition, it is already mounted and will generate
  //an error. By setting to an unknown fs it will not get mounted and used = size.
  if pLabel = VentoyStr then fs:= UnknownStr;
  if fs = UnknownStr then pUsed:= pSize;
  if pType = 'extended' then pUsed:= 0; //getting fs will have set it to unknown
  //for an extended part and swap part used=0, for an unknown filesystem pused=psize, for everything
  //else need to mount the part to get used blocks from df
  if (fs <> UnknownStr) and (fs <> SwapStr) and (pType <> 'extended') then
     begin
       if pMountpt = '' then UsedMount:= Mount(false) else UsedMount:= pMountpt;
       //if okay then get used else = psize
       if UsedMount <> '' then pUsed:= SysInfo.getUsedBlks(pdrive,pnumber,pSector) else pUsed:= pSize;
     end;
  //writeln(pDrive + pnumber,' part used=',pUsed,space,SizeStr(pUsed));
  if pos(MountFolder,UsedMount) > 0 then UMount;
  //pLabel if it has one
  pLabel:= SysInfo.getLabel(pdrive,pnumber);
  //writeln(pDrive + pnumber,' label=',pLabel);
  //UUID needed for swap parts
  UUID:= SysInfo.getUUID(pdrive,pnumber);
  //writeln(pDrive + pnumber,' UUID=',UUID);
  //sfFlags  : string;  //as reported by sfdisk dont think used in foxclone
  sfFlags:= SysInfo.getBootFlag(pdrive,pnumber); //only applies to msdos part table
  //writeln(pDrive + pnumber,' DOS bootflag=',sfFlags);
  //OS operating system populated by OSprober
end;
{-------------------------------------------------------------------------------
FUNCTION TPART.PRINT
Prints partition information to a terminal or a file.

Used by:

V1.0 - 26 Feb 2021
V1.1 - 07 Mar 2021, changed as a result of ntfslastsector no longer in the type definition and
                    saving sizes as sectors
V1.2 - 10 Apr 2021, removed parameters passed (device + prefix) as now in the TPart definition as pDrive
V1.3 - 28 Oct 2021, blkadjust removed, psize and pused are in blocks, output to textfile
-------------------------------------------------------------------------------}
procedure Tpart.Print(var OutF : Textfile);
const
  delim = '; ';
begin
  writeln(OutF,separator2);
  if pboot then write(OutF,'BOOT PARTITION ') else
     if EFIpart then write(OutF,'EFI PARTITION ') else write(OutF,'PARTITION ');
  writeln(OutF,pDrive + pnumber,delim);
  if pos('extended',pType) = 0 then //pType <> 'extended' then
     begin
       write(OutF,'fs: ',fs,'; ');
       if psize = 0 then writeln(OutF,'size: ERROR partition size reported as zero',delim) else
          begin
            write(OutF,'size: ',FormatFloat('#,##0',pSize),' blocks (',SizeStr(pSize),')',delim);
            writeln(OutF,'used: ',FormatFloat('#,##0',pUsed),' blocks (',FormatFloat('#0%',pused/psize*100),')',delim);
          end;
       write(OutF,'type: ',pType,'; ');
       if plabel <> '' then write(OutF,'label: ',plabel,delim);
       if pname <> '' then write(OutF,'name: ',pname,delim);
       if flags <> '' then writeln(OutF,'Flags: ',flags,delim) else writeln;
       if pmountpt <> '' then writeln(OutF,'mountpoint: ',pmountpt,delim);
       writeln(OutF,'UUID: ',UUID,delim);
       if OS <> '' then writeln(OutF,'OS: ',OS,delim);
     end
  else
     begin
       write(OutF,'type: ',pType,delim);
       writeln(OutF,'size: ',FormatFloat('#,##0',psize),' blocks (',SizeStr(psize),')',delim);
     end;
end; //end procedure TPart.Print
{-------------------------------------------------------------------------------
FUNCTION TPART.LISTSTR
Returns a partition string for display in checklist boxes.

Used by:

V1.0 - 26 Feb 2021
V1.1 - 07 Mar 2021 changed as a result of saving sectors not blocks
V1.2 - 28 Oct 2021 blkadjust removed, psize is now in blocks
-------------------------------------------------------------------------------}
function TPart.ListStr : string;
const
  ExtendStr = 'extended';
  PrimaryStr = 'primary';
  delim = ' • '; //output to terminal would not show, but happy in listbox, copied from character map
                 //U+2022 BULLET
begin
  ListStr:= pdrive + pnumber + delim;
  if ptype <> ExtendStr then
     begin
       ListStr+= fs;
       if plabel <> '' then ListStr+= ' [' + pLabel + ']' + delim else ListStr+= delim;
       ListStr+= SizeStr(pSize) + ' (' + FormatFloat('#0%',pused/psize*100)
                 + ' ' + rUsedStr + ')';
       if flags <> '' then if EFIpart then ListStr+= delim + rEFIpartStr
                              else ListStr+= delim + rFlagsStr + ': ' + flags;
       if pname <> '' then ListStr+= delim + pname;
       if ptype <> PrimaryStr then ListStr+= delim + ptype;
       if OS <> '' then ListStr+= delim + OS;
     end
  else
     begin
       ListStr+= ptype + delim;
       ListStr+= SizeStr(pSize);
     end;
end; //end function TPart.ListStr
{-------------------------------------------------------------------------------
FUNCTION TPART.SHORTSTR
Returns a partition string for display in the progress form

Used by:

V1.0 - 11 Oct 2021
V1.1 - 28 Oct 2021 BlkAdjust removed
-------------------------------------------------------------------------------}
function TPart.ShortStr : string;
const
  delim = ' • ';
  ExtendStr = 'extended';
begin
  ShortStr:= pdrive + pnumber + delim;
  if ptype = ExtendStr then ShortStr+= ptype + delim else ShortStr+= fs + delim;
  ShortStr+= SizeStr(pSize);
end; //end function TPart.ShortStr
{-------------------------------------------------------------------------------
FUNCTION TPART.HINTSTR
Returns a partition string for display in checklist boxes as a hint.

Parameters:
    iSource - added to distinguish between a source partition and a destination partition. When a source
              and part is extended, swap or EFI comment is added to the hintstr that it is autobacked
              up, when false this is not added.

Used by:

V1.0 - 26 Feb 2021
V1.1 - 07 Mar 2021 BlkAdjust on sizes as a result of saving sectors not blocks
V1.2 - 24 Mar 2021 isSource added as parameter
V1.3 - 30 Apr 2021 isSource removed parameter due to changes to a common part hint string
V1.4 - 28 Oct 2021 blkadjust removed, psize pused in blocks
V1.5 - 05 Jan 2022 size now reported in MB, user thought showing blocks was too detailed.
-------------------------------------------------------------------------------}
function TPart.HintStr(isSource : boolean) : string;
const
  ExtendStr = 'extended';
  LF2 = LF + LF;
  delim = ': ';
//var
//   AddStr : string = '';
begin
  //if isSource then AddStr:= LF + rAutoBackStr;
  HintStr:= pdrive + pnumber;
  if EFIpart then HintStr+= ' ' + rEFIpartStr;
  if ptype <> ExtendStr then
     begin
       if OS <> '' then HintStr+= ' ' + OS;
       HintStr+= LF2 + rFileSystemStr + delim + fs;
       HintStr+= LF2 + rSizeStr + delim + FormatMB(pSize);
       HintStr+= LF2 + rUsedUCStr + delim + FormatMB(pUsed);
       HintStr+= ' (' + FormatFloat('#0%',pused/psize*100) + ')';
       if plabel <> '' then HintStr+= LF2 + rLabelStr + delim + plabel;
       if pname <> '' then HintStr+= LF2 + rNameStr + delim + pname;
       HintStr+= LF2 + rTypeStr + delim + ptype;
       if flags <> '' then HintStr+= LF2 + rFlagsStr + delim + flags;
     end
  else
     begin
       HintStr+= LF2 + rTypeStr + delim + ptype;
       HintStr+= LF2 + rSizeStr + delim + FormatMB(pSize);
     end;
end; //end function TPart.HintStr
{-------------------------------------------------------------------------------
FUNCTION TPART.USERSELECT
Returns true if a partition can be selected/de-selected by the user. EFI, swap and extended partitions
cannot be selected by the user.

Used by:

V1.0 - 31 Mar 2021
V1.1 - 12 Aug 2023, bios_grub flag added so that a bios_grub partition is always backed up and cannot
                    be selected by the user.
-------------------------------------------------------------------------------}
function TPart.UserSelect : boolean;
const
  ExtendStr = 'extended';
  SwapStr = 'swap';
  FlagStr = 'bios_grub';
begin
  if EFIpart or (pType = ExtendStr) or (fs = SwapStr) or (flags = FlagStr) then UserSelect:= false
  else UserSelect:= true;
end; //end function TPart.UserSelect
{-------------------------------------------------------------------------------
FUNCTION TPART.UMOUNT
Unmounts a partition. If the partition has been mounted by foxclone = MountFolder is the start
of the mount point, then it also removes the mount folder.

issues - using /dev/sdaX to unmount or using the mount folder. when mounted in more than one location
what does it do????

Used by:

V1.0 - 07 Apr 2021
V1.1 - 03 Nov 2021, if checking the mount point returned an empty string, the return from the function
                    was indeterminate, UMount was not set, fixed. Changed message dialog to foxmsg.
V1.2 - 12 Nov 2021, pkexec added to lsblk command
V1.3 - 08 Jan 2022, was getting an error trying to umount a swap partition, check and exit.
-------------------------------------------------------------------------------}
function TPart.UMount : boolean;
const
  bufsize = 128; // Buffer size for reading output
  swapstr = 'swap';
var
  PartRef : string;
  MountPt : string;
  umntProcess : TProcess;
  Buffer : string = '';
  BytesRead   : byte;
begin
  UMount:= true; //default is success
  //if a swap partition nothing to do
  if fs = swapstr then exit;
  //first check if a loop device
  if pNumber = '0' then
     begin
       //have to check if we have a loop device on an nvme or sd card and strip out the last p
       if pos('p',pdrive) = length(pdrive) then PartRef:= copy(pdrive,1,length(pdrive)-1)
       else PartRef:= pDrive;
     end
  else PartRef:= pDrive + pNumber;
  //get current mount point
  MountPt:= trim(getStr('pkexec lsblk -nio mountpoint /dev/' + PartRef));
  //check if mounted
  if MountPt <> '' then
     begin
       //umount the partition
       setlength(Buffer,bufsize);
       umntProcess:= TProcess.create(nil);
       // -l is lazy umount, in testing not all partitions were unmounting
       termCmd('pkexec umount -l ' + '/dev/' + PartRef,umntProcess);
       //output from the command is on stderr
       BytesRead:= umntProcess.stderr.Read(buffer[1],length(buffer));
       if BytesRead > 0 then
          begin
            setlength(buffer,BytesRead);
            FoxMsg(rErrorStr,[rUmountFailStr + ' /dev/' + PartRef,buffer],mtError,[mbOK]);
            UMount:= false;
          end;
       umntProcess.free;
       if UMount then if pos(MountFolder,MountPt) = 1 then rmdir(PartRef); //remove mount folder
     end
end; //end function TPart.UMount
{-------------------------------------------------------------------------------
FUNCTION TPART.MOUNT
Mounts a partition and returns the mount point as a string. Returns an empty string if the mount failed.
Failure also includes failure to create the mount folder.

If the partition is already mounted it just returns the mount point, if not it mounts the partition.

Mount has to check if the partition is on a loop device (pnumber = '0') and adjust the mount point and
mount command accordingly.

It creates the mount folder if needed in the default mount folder as the partition number, e.g.
/mnt/foxclone/sdd1.

Note - does not use getStr for the terminal command as this would loop several times on failure -
where getStr returns an empty string if the partition is not already mounted.

Parameters:

    ReadWrite - when set true mounts the partition read/write, otherwise mounts it read only.

    Mountpoint - a string, if supplied uses this as the mount point. When not provided uses the
    default mount folder and partition reference, e.g. sda2.

Used by:

V1.0 - 05 Apr 2021
V1.1 - 21 Dec 2021, message dialog changed to foxmsg
-------------------------------------------------------------------------------}
function TPart.Mount(ReadWrite : boolean; MountPoint : string = '') : string;
const
  bufsize = 128; // Buffer size for reading output
var
  PartRef : string;
  Access : string; //write access or read only
  MntCmd : string;
  mntProcess : TProcess;
  buffer     : string = '';
  BytesRead  : integer;     //from the stream
begin
  //check for a loop device, have '0' as pNumber and need to mount it as sdX not sdXY
  if pNumber = '0' then
     begin
       //have to check if we have a loop device on an nvme or sd card and strip out the last p
       if pos('p',pdrive) = length(pdrive) then PartRef:= copy(pdrive,1,length(pdrive)-1)
       else PartRef:= pDrive;
     end
  else PartRef:= pDrive + pNumber;
  Mount:= trim(getStr('pkexec lsblk -nio mountpoint /dev/' + PartRef));
  if Mount = '' then //not mounted
     begin
       if MountPoint = '' then MountPoint:= MountFolder + PartRef;
       //check passed parameters
       if ReadWrite then Access:= '' else Access:= '-r '; // -r = read-only
       MntCmd:= 'pkexec mount ' + Access + ' /dev/' + PartRef + ' ' + MountPoint;
       //writeln('mount command=',MntCmd);
       //create mount folder if needed
       if mkdir(MountPoint) > -1 then //mount folder created or usable
          begin
            setlength(buffer,bufsize);
            mntProcess:= TProcess.create(nil);
            termCmd(MntCmd,mntProcess);
            BytesRead:= mntProcess.output.Read(buffer[1],length(buffer));
            if BytesRead > 0 then
               begin
                 setlength(buffer,BytesRead);
                 FoxMsg(rWarningStr,[rMountFailStr + ' /dev/' + PartRef,buffer],mtError,[mbOK]);
                 Mount:= '';
               end
            else Mount:= MountPoint;
            mntProcess.free;
          end
     end;
end; //end function TPart.Mount
{-------------------------------------------------------------------------------
FUNCTION TPART.FOXMOUNT
Checks if a partition was mounted by foxclone and returns true if it was

Does this by checking the mount point, if the mount point starts with the default mount folder =
/mnt/foxclone, then it was mounted by foxclone.

Has to check if a loop device is mounted and adjust the mount point accordingly.

Used by:

V1.0 - 05 Apr 2021
V1.1 - 12 Nov 2021, pkexec added to lsblk command
-------------------------------------------------------------------------------}
function TPart.FoxMount : boolean; //returns true if partition mounted by foxclone
var
  PartRef : string;
begin
  //check for a loop device, if '0' is pNumber and need to treat the partition as sdX not sdXY
  if pNumber = '0' then //we have a loop device on an nvme or sd card and strip out the last p
     begin
       if pos('p',pdrive) = length(pdrive) then PartRef:= copy(pdrive,1,length(pdrive)-1)
       else PartRef:= pDrive;
     end
  else PartRef:= pDrive + pNumber;
  FoxMount:= pos(MountFolder,trim(getStr('pkexec lsblk -nio mountpoint /dev/' + PartRef))) = 1;
end; //end function TPart.FoxMount
{-------------------------------------------------------------------------------
PROCEDURE TPART.PARSESFDISK
Parses a partition line in an sfdisk output to extract
       pnumber - this is stored as sda1, it will be later stripped of sda to just leave '1'
       pstart - start sector
       psize - partition size in sectors
       uuid gpt only
       name gpt

Procedure as a method of TPart to avoid having to define a read/write property for ALL the variables in
TPart.

Parameters:
    sfPartStr - a partition string from sfdisk, format differs slightly between an msdos and gpt
    partition table

    GPT
    /dev/sda1 : start=        2048, size=     1083392, type=DE94BBA4-06D1-4D40-A16A-BFD50179D6AC, uuid=DA48E761-66C9-4D01-A653-FC0A72E5F656, name="Basic data partition", attrs="RequiredPartition"

    MSDOS
    /dev/sda1 : start=        2048, size=    66404352, type=83, bootable

    dSector - new, integer, the sector size of the drive used to populate psector.

Used by:

V1.0 - 09 Apr 2021
V1.1 - 14 Apr 2021, amended to use a TStringDynArray with SplitString = easier than a lot of copy
V1.2 - 28 Oct 2021, changes consequent on deleting blkadjust, procedure now populates psector from
                    dsector passed as a parameter. psize converted to blocks
-------------------------------------------------------------------------------}
procedure TPart.ParseSFdisk(sfPartStr : string; dSector : integer);
const
  delim = ',';
  DevStr ='/dev/';
  StartStr = 'start=';
  SizeStr = 'size=';
  TypeStr = 'type=';
  UUIDstr = 'uuid=';
  NameStr = 'name=';
  //there is potentially an "attr=" field for gpt, not parsed
var
  ParamList : TStringDynArray;
begin
  //sector size
  pSector:= dSector;
  ParamList:= SplitString(sfPartStr,delim);
  //ParamList[0] is both part number, e.g. sda1 and the start sector
  //pnumber - the device name is stripped out by a later procedure
  ParamList[0]:= replacestr(ParamList[0],DevStr,'');
  pNumber:= trim(copy(ParamList[0],1,pos(':',ParamList[0])-1));
  //pStart
  ParamList[0]:= copy(ParamList[0],pos(':',ParamList[0])+1,99);
  ParamList[0]:= trim(replacestr(ParamList[0],StartStr,''));
  pStart:= StrToInt64(ParamList[0]);
  //ParamList[1] = pSize
  ParamList[1]:= trim(replacestr(ParamList[1],SizeStr,''));
  pSize:= SectorToBlock(StrToInt64(ParamList[1]),pSector);
  //ParamList[2] = type, uuid-like for gpt a number for msdos
  //not used at the moment in foxclone
  psfType:= trim(replacestr(ParamList[2],TypeStr,''));
  //from here, only applicable to gpt
  //ParamList[3] = uuid
  if Length(ParamList) > 3 then UUID:= trim(replacestr(ParamList[3],UUIDstr,''));
  //ParamList[4] = pName (partition name)
  if Length(ParamList) > 4 then
     begin
       pName:= trim(replacestr(ParamList[4],NameStr,''));
       pName:= replacestr(pName,'"',''); //get rid of the double quotes
     end;
end; //end procedure TPart.ParseSFdisk
{-------------------------------------------------------------------------------
PROCEDURE TPART.PARSEBACKFILE
Parses a partition and a blocks line in a backup file to extract:
       fs = filesystem
       label
       pName = partition name, gpt only
       flags if present
       OS if present
       pUsed - in the backup file is in blocks, converted to sectors

Parameters:
    PTableType - msdos or gpt, used to determine if the field in the backup file refers to either the
    partition name (gpt) or partition type (msdos) = primary, logical or extended.

    PartStr - varies slightly between msdos and gpt.

    BlockStr - of the form 'Blocks:sda6,3248364'.

Example formats:
GPT
  Partition:sda1 : ntfs : Basic data partition : 542MB : 73% used [Recovery] : Flags-hidden, diag
  Blocks:sda1,395468
  Partition:sda2 : fat32 : EFI system partition : 98.3MB : 32% used : Flags-boot, esp
  Blocks:sda2,31597
  Partition:sda3 : <unknown> : Microsoft reserved partition : 16.8MB : 100% used : Flags-msftres
  Blocks:sda3,16800
  Partition:sda4 : ntfs : Basic data partition : 39.2GB : 41% used : Flags-msftdata
  Blocks:sda4,15931204
  Partition:sda5 : ext4 : <none> : 15.1GB : 41% used [root] *Linux Mint 19.2 Tina*
  Blocks:sda5,6120028
  Partition:sda6 : ext4 : <none> : 20.0GB : 16% used [home]
  Blocks:sda6,3248364
  Partition:sda7 : swap : d785dbd8-2558-4993-ac1d-ba09202f60d0
  Blocks:sda7,0
MSDOS
  Partition:sda1 : ntfs : primary : 593MB : 70% used [System Reserved] *Windows 10*
  Blocks:sda1,414492
  Partition:sda2 : ntfs : primary : 59.8GB : 38% used
  Blocks:sda2,22599652
  Partition:sda3 : ext4 : primary : 10.0GB : 27% used : Flags-boot *Bodhi Linux 5*
  Blocks:sda3,2732208
  Partition:sda4 : extended
  Blocks:sda4,0
  Partition:sda5 : ext4 : logical : 12.0GB : 1% used
  Blocks:sda5,69340
  Partition:sda6 : swap : 8a2a0bf9-f6e6-49ee-9eaa-c66a2ccd5cbf
  Blocks:sda6,0

The procedure sets selected=true to indicate that the backup does have an image for the partition. Not
strictly true for extended or swap parts.

Used by:

V1.0 - 12 Apr 2021
V1.1 - 14 May 2021, it was not setting the flag for the EFI partition, it does now
V1.2 - 26 May 2021, was not setting selected = false for partitions not in the backup
V1.3 - 28 Oct 2021, blkadjust removed, pused is in blocks
V1.4 - 09 Nov 2021, changed test on pname to searching on pos to match new format
-------------------------------------------------------------------------------}
//why is sda7 on uefi 20191110 showing a differnt uuid for sda7= swap in sfdisk and backup
procedure TPart.ParseBackFile(PTableType,PartStr, BlockStr : string);
const
  delim = ':';
  SwapStr = 'swap';
  GPTstr = 'gpt';
  DOSstr = 'msdos';
  FlagsStr = 'Flags-';
  BlkStr = 'Blocks:';
  ExtendStr = 'extended';
  LogicalStr = 'logical';
  NoneStr = 'none';
  OldUnknownStr = '<unknown>';
  UnknownStr = 'unknown';
  Fat32Str = 'fat32';
var
  TempStr : string;
  ParamList : TStringDynArray;
  //i : byte;
begin
  Selected:= true; //i.e. there is a backup image for the partition
  ParamList:= SplitString(PartStr,delim);
  //for i:= 0 to Length(ParamList) - 1 do writeln('Param ',i,' =',ParamList[i]);
  //ParamList[0] is always 'Partition'
  //ParamList[1] is the the partition ref, e.g. sda1
  //ParamList[2] is the filesystem or swap or extended
  fs:= trim(ParamList[2]);
  if Length(ParamList) > 3 then //for extended part no more entries
     begin
       //ParamList[3] is either the swap UUID, part name (gpt) or logical/primary (msdos)
       if fs = SwapStr then UUID:= trim(ParamList[3]) else
          if PTableType = GPTstr then pName:= trim(ParamList[3]) else pType:= trim(ParamList[3]);
       //ParamList[4] is size, not needed, got from sfdisk
       //ParamList[5] is % used, not needed, but may contain the label or OS
       if pos('[',ParamList[5]) > 0 then //there is a label
          begin
            TempStr:= copy(ParamList[5],pos('[',ParamList[5])+1,99);
            TempStr:= copy(TempStr,1,pos(']',TempStr)-1);
            pLabel:= TempStr;
          end;
       if pos('*',ParamList[5]) > 0 then //there is an operating system
          begin
            TempStr:= copy(ParamList[5],pos('*',ParamList[5])+1,99);
            TempStr:= copy(TempStr,1,pos('*',TempStr)-1);
            OS:= TempStr;
          end;
     end;
  if Length(ParamList) = 7 then //there are some flags
     begin
       //first take out the operating system if present
       if pos('*',ParamList[6]) > 0 then //there is an operating system
          begin
            TempStr:= copy(ParamList[6],pos('*',ParamList[6])+1,99);
            TempStr:= copy(TempStr,1,pos('*',TempStr)-1);
            OS:= TempStr;
          end;
       //get the flags
       if pos('*',ParamList[6]) > 0 then TempStr:= copy(ParamList[6],1,pos('*',ParamList[6])-1)
       else TempStr:= ParamList[6];
       Flags:= trim(replacestr(TempStr,FlagsStr,''));
       //writeln('back file ',fs,'flags =',flags);
       if (pos('esp',Flags) > 0) and (pos('boot',flags) > 0) and (fs = Fat32Str) then EFIpart:= true;
       //if EFIpart then writeln('found efi part in back file');
     end;
  //get used blocks and convert to sectors
  //pnumber is populated by TDrive.ReadsfDisk
  TempStr:= replacestr(BlockStr,BlkStr + pDrive + pNumber + ',','');
  pUsed:= StrToInt64(TempStr); //blocks
  //pUsed:= pUsed * BlkAdjust;   //sectors
  if pUsed > pSize then pUsed:= pSize; //fix for some old backups and MS reserved parts
  //tidy up
  if fs = OldUnknownStr then fs:= UnknownStr;
  if pos(NoneStr,pName) > 0 then pName:= '';
  //need to check swap on a msdos drive
  if (fs = SwapStr) and (PTableType = DOSstr) and (StrToInt(pNumber) > 4) then pType:= LogicalStr;
  //sort out the extended part
  if fs = ExtendStr then
     begin
       pType:= fs;
       fs:= '';
     end;
  //check if the partition was backed up, this will have the line Blocks:sda6,0, so pused = 0
  //blocks are not recorded for an extended part or a swap part
  if (pused = 0) and (pType <> ExtendStr) and (fs <> SwapStr) then Selected:= false;
end; //end procedure TPart.ParseBackFile
{-------------------------------------------------------------------------------
FUNCTION TPART.VERIFYSTR
Generates a terminal command line to run partclone to verify an image file.
For extended or swap partitions the function returns an empty string. This can be tested by the
calling procedure.

Note that VerifyStr is called on the partitions in RDrive.

The command line is determined by the backup settings for compression and split. Example terminal
commands are:

compress no split (seComp - the default):
cat 20191107.sda1.img.gz | unpigz -c | partclone.chkimg -s - -L 20191107.sda1.img.chk

compress and split (seBoth):
cat 20211116.sda1.img.gz.* | unpigz -c | partclone.chkimg -s - -L 20211116.sda1.img.chk

uncompressed split (seSplit):
cat 20210202.sda1.img.* | partclone.chkimg -s - -L 20210202.sda1.img.chk

uncompressed no split (seNone):
partclone.chkimg -s 20211116.sda1.img -L 20211116.sda1.img.chk

The logfile name is the same for all options, .chk is appended to the image filename. The logfile
is saved in the original backup folder.

Note - it does not test for an unknown filesystem, you cannot check an image file created with
partclone.dd and this is filtered out before this function is called.

Note2 - the calling procedure what is it needs to check the unsupported filesystems.


Parameters:
    BackFname - the full pathname for backup file, key is that this has .backup as the extension and
    must be stripped out and that the filename may have spaces in it - which must be escaped.

Used by:

V1.0 - 05 Dec 2021
-------------------------------------------------------------------------------}
function TPart.VerifyStr(BackFname : String) : string;
const
  partcloneStr = 'partclone.chkimg -s - -L [logfile]';
  logStr = '[logfile]';
  chkExtnStr = '.chk';
  backExtn = '.backup';
  dotStr = '.';
  imgStr = '.img';
  pigzStr = ' | unpigz -c | ';
  catStr = 'cat ';
  SplitStr = '.* ';
  PipeStr = ' | ';
  CompStr = '.gz';
  swapStr = 'swap';
  extendStr = 'extended';
  DashStr = ' - ';
var
  LogFile : string;
  SourceFile : string;
begin
  //return an empty string if the partition is a swap part or extended part, nothing for partclone to do
  if (fs = swapStr) or (ptype = extendStr) then
     begin
       VerifyStr:= '';
       exit;
     end;
  //escape any spaces in the filename
  BackFname:= replacestr(BackFname,Space,ESCspace);
  //build the logfile name
  LogFile:= replacestr(BackFname,backExtn,dotStr + pDrive + pnumber + imgStr + chkExtnStr);
  SourceFile:= replacestr(BackFname,backExtn,dotStr + pDrive + pnumber + imgStr);
  case RSettings of
       seNone  : begin
                   VerifyStr:= replacestr(partcloneStr,Dashstr,Space+ SourceFile + Space);
                   VerifyStr:= replacestr(VerifyStr,logStr,LogFile);
                 end;
       seComp  : begin
                   SourceFile:= SourceFile + CompStr;
                   VerifyStr:= catStr + SourceFile + pigzStr + partcloneStr;
                   VerifyStr:= replacestr(VerifyStr,logstr,LogFile);
                 end;
       seSplit : begin
                   SourceFile:= SourceFile + SplitStr;
                   VerifyStr:= catStr + SourceFile + PipeStr + partcloneStr;
                   VerifyStr:= replacestr(VerifyStr,logstr,LogFile);
                 end;
       seBoth  : begin
                   SourceFile:= SourceFile + CompStr + SplitStr;
                   VerifyStr:= catStr + SourceFile + pigzStr + partcloneStr;
                   VerifyStr:= replacestr(VerifyStr,logstr,LogFile);
                 end;
  end;
end; //end function TPart.VerifyStr
{-------------------------------------------------------------------------------
FUNCTION TPART.RESTORESTR
Generates a terminal command line to run partclone for a restore or clone from file. For extended or
swap partitions the function returns an empty string. This can be tested by the calling procedure.

Note that RestoreStr is called on the partitions in RDrive NOT the partitions on the target drive
(set acTarget).

The command line is determined by the backup settings for compression and split. Example terminal
commands are:

compress no split (seComp - the default):
cat /20191110.sda6.img.gz | unpigz -c | pkexec partclone.ext4 -L /tmp/foxclone/20211116.sda6.log -r -F -s - -O /dev/sda6

split no compress (seSplit):
cat 20210202.sda4.img.* | pkexec partclone.ntfs -L /tmp/foxclone/20211116.sda4.log -r -F -s - -O /dev/sda4

no split no compress (seNone):
pkexec partclone.ext4 -L /tmp/foxclone/20211116.sda1.log -r -F -s 20211116.sda1.img -O /dev/sda1

compress and split (seBoth):
cat 20211116.sda1.img.gz.* | unpigz -c | pkexec partclone.ext4 -L /tmp/foxclone/20211116.sda1.log -r -F -s - -O /dev/sda1

cat /media/andy/looppart/compressed\ split/20220103.sda1.img.gz.* pkexec partclone.ntfs -L /tmp/foxclone/20220103.sda1.log -r -F -s - -O /dev/sda1

The logfile name is the same for all options and is dumped in the tmp folder. The logfile name uses
the destination partition NOT the source partition.

Note for an unknown filesystem, i.e. using dd to restore, the options '-r -F' are not used in partclone.

Parameters:
    BackFname - the full pathname for backup file, key is that this has .backup as the extension and
    must be stripped out and that the filename may have spaces in it - which must be escaped.

    DestDevice - between taking a backup and doing a restore the target drive may have changed, e.g.
    backup taken with drive on sda, but it is on sdb when the restore is done. As RestoreStr is called
    on RDrive, DestDevice contains the destination drive, e.g. sde and is used to build the output
    string, e.g. sde1. It must contain the part prefix, e.g. nvme0n1p not nvme0n1.

Used by:

V1.0 - 16 Nov 2021
V1.1 - 04 Dec 2021, was escaping spaces on both logfile and imgfile, now just on BackFname
V1.2 - 03 Jan 2022, seBoth cmd string wrong, missing unpigz string
V1.3 - 17 Apr 2024, modified to allow for a LUKS partition identified as crypto_LUKS
V1.4 - 19 Apr 2024, modified for LVM
V1.4 - 20 Apr 2024, changed to generic test on unsupported filesystems
V1.5 - 26 Apr 2024, changed to checking for supported filesystems where SupportedFS is defined in
                    utilities.
-------------------------------------------------------------------------------}
function TPart.RestoreStr(BackFname : String; DestDevice : string) : string;
const
  pkexecStr = 'pkexec partclone.[fs] -L [logfile] -r -F -s [source] -O ';
  logStr = '[logfile]';
  logExtnStr = '.log';
  backExtn = '.backup';
  dotStr = '.';
  fsStr = '[fs]';
  imgStr = '.img';
  SourceStr = '[source]';
  pigzStr = ' | unpigz -c | ';
  catStr = 'cat ';
  devStr = '/dev/';
  SplitStr = '.* ';
  PipeStr = ' | ';
  CompStr = '.gz';
  AddOptionsStr = '-r -F ';
  ddmodeStr = 'dd';
  swapStr = 'swap';
  extendStr = 'extended';
var
  LogFile : string;
  pkexecCmd : string;
  ImgFile : string;
  OutStr : string;
begin
  //return an empty string if the partition is a swap part or extended part, nothing for partclone to do
  if (fs = swapStr) or (ptype = extendStr) then
     begin
       RestoreStr:= '';
       exit;
     end;
  //escape any spaces in the filename
  BackFname:= replacestr(BackFname,Space,EscSpace); //escape any spaces
  //build the output string
  OutStr:= devStr + DestDevice + pNumber;
  //build the logfile name
  Logfile:= DefaultFolder + copy(BackFname,rpos('/',BackFname)+1,99);
  Logfile:= replacestr(Logfile,backExtn,dotStr + DestDevice + pNumber + logExtnStr);
  //build image file name
  Imgfile:= replacestr(BackFname,backExtn,dotStr + pDrive + pNumber + imgStr);
  //build the pkexec command
  pkexecCmd:= replacestr(pkexecStr,logStr,Logfile); //insert log filename
  if (pos(fs,SupportedFS) = 0) then
     begin
       pkexecCmd:= replacestr(pkexecCmd,fsStr,ddmodeStr);
       pkexecCmd:= replacestr(pkexecCmd,AddOptionsStr,''); //strip them out
     end
  else pkexecCmd:= replacestr(pkexecCmd,fsStr,fs); //insert filesystem into partclone command
  if RSettings = seNone then pkexecCmd:= replacestr(pkexecCmd,SourceStr,Imgfile)
  else pkexecCmd:= replacestr(pkexecCmd,SourceStr,'-');
  //build the command line
  case RSettings of
       seNone  : RestoreStr:= pkexecCmd + OutStr; //tested
       seComp  : RestoreStr:= catStr + Imgfile + CompStr + pigzStr + pkexecCmd + OutStr; //tested
       seSplit : RestoreStr:= catStr + Imgfile + SplitStr + PipeStr + pkexecCmd + OutStr; //tested
       seBoth  : RestoreStr:= catStr + Imgfile + CompStr + SplitStr + pigzStr + pkexecCmd + OutStr; //tested
  end;
end; //end TPart.RestoreStr
{-------------------------------------------------------------------------------
FUNCTION TPART.CLONESTR
Generates a terminal command line to run partclone for a clone. For extended or swap partitions the
function returns an empty string. This can be tested by the calling procedure.

Simplest of all the partclone strings, there are only two forms:

pkexec partclone.ext4 -F -b --logfile /tmp/foxclone/20211126.sdd5.log -s /dev/sda5 -O /dev/sdd5
for a known file system
or
pkexec partclone.dd -F --logfile /tmp/foxclone/20211126.sdd3.log -s /dev/sda3 -O /dev/sdd3
for an unknown filesystem, the only difference being -b

Options:
    -F : Force progress.
    -b : Local device to device copy on-the-fly, source and output both are device.

    then why omit on dd, try it.

The function must be called on the drive set acSource, the target is specified in DestDrive.

Parameters:
    DestDrive - in the form sda, where this must include the part prefix


Used by:

V1.0 - 26 Nov 2021
V1.1 - 02 Dec 2021, initially thought that the -b option was okay with dd, it's not, modified to remove.
V1.2 - 18 Apr 2024, modified to cater for a LUKS encrypted partition
V1.3 - 20 Apr 2024, changed to generic test for unsupported filesystems
V1.4 - 26 Apr 2024, changed to checking for supported filesystems where SupportedFS is defined in
                    utilities.
-------------------------------------------------------------------------------}
function TPart.CloneStr(DestDrive : string) : string;
const
  pkexecStr = 'pkexec partclone.[fs] -F -b --logfile [logfile] -s /dev/[source] -O /dev/[dest]';
  ddmodeStr = 'dd'; //used for unknown filesystems, uses partclone.dd
  ddSwitchStr = ' -b'; //switch not used when using partclone.dd
  logStr = '[logfile]';
  logExtnStr = '.log';
  dotStr = '.';
  fsStr = '[fs]';
  SourceStr = '[source]';
  DestStr = '[dest]';
  swapStr = 'swap';
  extendStr = 'extended';
var
  Logfile : string; //takes the form 20211126.sda1.log and stored in the default folder /tmp/foxclone
begin
  //return an empty string if the partition is a swap part or extended part, nothing for partclone to do
  if (fs = swapStr) or (ptype = extendStr) then
     begin
       CloneStr:= '';
       exit;
     end;
  //build the logfile name
  Logfile:= DefaultFolder + FormatDateTime('yyyymmdd',now) + dotStr + Destdrive + pNumber + logExtnStr;
  //insert logfile name
  CloneStr:= replacestr(pkexecStr,logStr,Logfile);
  //filesystem
  if (pos(fs,SupportedFS) = 0) then
     begin
       CloneStr:= replacestr(CloneStr,fsStr,ddmodeStr);
       CloneStr:= replacestr(CloneStr,ddSwitchStr,''); // -b option not applicable to partclone.dd
     end
  else CloneStr:= replacestr(CloneStr,fsStr,fs);
  //source partition
  CloneStr:= replacestr(CloneStr,SourceStr,pDrive + pNumber);
  //destination partition
  CloneStr:= replacestr(CloneStr,DestStr,DestDrive + pNumber);
end; //end function TPart.CloneStr
{-------------------------------------------------------------------------------
FUNCTION TPART.BACKSTR
Generates a terminal command line to run partclone for a backup. For extended or swap partitions the
function returns an empty string. This can be tested by the calling procedure.

Parameters:
    BackFname - the full pathname for backup file, key is that this has .backup as the extension and
    must be stripped out and that the filename may have spaces in it - which must be escaped.

The command line is determined by the backup settings for compression and split. Example terminal
commands are:

Default compress no split (seComp)
pkexec partclone.ntfs -c -F -s /dev/sda1 -L "/media/20211020.sda1-log.txt"  | pigz -c -1 > "/media/20211020.sda1.img.gz"

no compression no split (seNone)
pkexec partclone.ntfs -c -F -s /dev/sda1 -L "/media/20211020.sda1-log.txt"  -O "/media/20211020.sda1.img"

Split and compress (seBoth)
pkexec partclone.ntfs -c -F -s /dev/sda1 -L "/media/20211020.sda1-log.txt"  | pigz -c -1 | split -d -b 2048M - "/media/20211020.sda1.img.gz."

Split no compress (seSplit)
pkexec partclone.ntfs -c -F -s /dev/sda1 -L "/media/20211020.sda1-log.txt"  | split -d -b 2048M - "20211020.sda1.img."
                             dotStr = '.'; //separator and added at the end of split image filenames

Unknown filesystems use partclone.dd - this is tested for and in addition the "-c" option is stripped
out of the options for partclone. Example:

pkexec partclone.dd -F -s /dev/sda3 -L "/media/20211022.sda3-log.txt"  | pigz -c -1 > "/media/20211022.sda3.img.gz"

The logfile name is the same for all options.

Used by:

V1.0 - 23 Oct 2021
V1.1 - 09 Nov 2021, got seBoth wrong, missing .gz
V1.2 - 17 Apr 2024, failing on LUKS where fs is reported as crypto_LUKS not unknown. Now checks for
                    crypto_LUKS
V1.3 - 19 Apr 2024, modified for LVM where the fs is shown as LVM2_member, tests on LVM because not sure
                    how other partitions part of the logical volume will show
V1.3 - 20 Apr 2024, changed to generic test on unsupported filesystems
V1.4 - 26 Apr 2024, changed to test on supported filesystems where SupportedFS is defined in utilities.
-------------------------------------------------------------------------------}
function TPart.BackStr(BackFname : String) : string;
const
  pkexecStr = 'pkexec partclone.[fs] -c -F -s [device] -L [logfile]';
  fsStr = '[fs]';
  deviceStr ='[device]';
  logStr = '[logfile]';
  devStr = '/dev/';
  backExtn = '.backup';
  logExtn = '-log.txt';
  pigzStr = ' | pigz -c -1 > ';
  redirectStr = '> ';
  splitStr = ' | split -d -b 2048M - ';
  imgStr = '.img';
  compStr = '.gz'; //compressed image file
  dotStr = '.'; //separator and added at the end of split image filenames
  swapStr = 'swap';
  extendStr = 'extended';
  ddmodeStr = 'dd'; //used for unknown filesystems, uses partclone.dd
  ddSwitchStr = ' -c'; //switch not used when using partclone.dd
  noCompStr = ' -O ';
var
  LogFile : string;
  ImgFile : string;
begin
  //return an empty string if the partition is a swap part or extended part, nothing for partclone to do
  if (fs = swapStr) or (ptype = extendStr) then
     begin
       BackStr:= '';
       exit;
     end;
  //escape any spaces in the backup filename
  BackFname:= replacestr(BackFname,Space,EscSpace);
  //generate the logfile name from the backup filename
  Logfile:= replacestr(BackFname,backExtn,dotStr + pdrive + pnumber + logExtn);
  //generate initial partclone command
  BackStr:= replacestr(pkexecStr,deviceStr,devStr + pdrive + pnumber);
  BackStr:= replacestr(BackStr,logStr,Logfile); //add in the logfile name
  //sort out the filesystem
  //writeln('filesystem=',fs);
  if (pos(fs,SupportedFS) = 0) then
     begin
       BackStr:= replacestr(BackStr,fsStr,ddmodeStr);
       BackStr:= replacestr(BackStr,ddSwitchStr,''); //strip out the -c switch from partclone cmd
     end
  else BackStr:= replacestr(BackStr,fsStr,fs);
  //setup initial image filename
  ImgFile:= replacestr(BackFname,backExtn,dotStr + pdrive + pnumber + imgStr);
  //if Defaults.BackSettings = seComp then writeln('compression no split');
  //build the command line
  case Defaults.BackSettings of
       seComp : begin //the default, compression no split
                  ImgFile+= compStr;
                  BackStr+= pigzStr + ImgFile;
                end;
       seNone : BackStr+= noCompStr + ImgFile; //no compresson no split
       seBoth : begin //compression and split
                  ImgFile+= compStr + dotStr;
                  BackStr+= pigzStr;
                  BackStr:= replacestr(BackStr,redirectStr,''); //get rid of "> "
                  BackStr+= splitStr + ImgFile;
                end;
       seSplit : begin //no compression with split
                   ImgFile+= dotStr;
                   BackStr+= splitStr + ImgFile;
                 end;
  end;
  DebugInfo(false,'TPart.BackStr',BackStr);
end; //end function TPart.BackStr
{-------------------------------------------------------------------------------
PROCEDURE TPART.BACKFILESTR
Generates a string for the backup file following the format of previous foxclone versions.

Parameters:
    PartTable - either msdos or gpt, output is different dependent on the ptable.

Used by:

V1.0 - 08 Nov 2021
-------------------------------------------------------------------------------}
function TPart.BackFileStr(PartTable : string) : string;
const
  delim = ' : ';
begin
  BackFileStr:= 'Partition:';
  BackFileStr+= pDrive + pNumber;
  if (fs <> 'swap') and (pType <> 'extended') then
     begin
       BackFileStr+= delim + fs;
       if PartTable = 'msdos' then BackFileStr+= delim + pType
       else BackFileStr+= delim + pName;
       BackFileStr+= delim + SizeStr(pSize);
       BackFileStr+= delim + FormatFloat('#,##0',pUsed/pSize*100)+ '% used';
       if pLabel <> '' then BackFileStr+= ' [' + pLabel + ']';
       if flags <> '' then BackFileStr+= delim + 'Flags-' + flags;
       if OS <> '' then BackFileStr+= ' *' + OS + '*';
     end;
  if pType = 'extended' then BackFileStr+= delim + pType;
  if fs = 'swap' then
     begin
       BackFileStr+= delim + fs;
       BackFileStr+= delim + UUID;
     end;
end; //end function TPart.BackFileStr
{-------------------------------------------------------------------------------
PROCEDURE TPART.MAKESWAP
Restores a swap partition. Failure is not fatal, but the function returns false on error so that
foxclone can flag it. The user can always recreate it after the event.

Parameters:
  DestDrive - because makeswap is called on RDrive, the destination may not be the same, e.g. the backup
  may have been done on sda, but with additional drives plugged in, it could be something else, e.g.
  sdd for the restore. DestDrive contains the destination drive in the format sda, for nvme drives it
  MUST include the part prefix, e.g. nvme0n1p.

Used by:

V1.0 - 19 Nov 2021
V1.1 - 20 Nov 2021 changed to a boolean function, returns false on an error
-------------------------------------------------------------------------------}
function TPart.MakeSwap(DestDrive : string) : boolean;
const
  bufsize = 128;
  ErrorStr1 = 'error';
  ErrorStr2 = 'cannot';
var
  cmdstr, buffer, Swapstr : string;
  cmdProcess : TProcess;
  BytesRead : integer;
begin
  MakeSwap:= true; //assume success
  cmdstr:= 'pkexec mkswap -U ' + UUID + ' /dev/' + DestDrive + pNumber;
  //writeln('make swap cmd=',cmdstr);
  //writeln('make swap UUID=',UUID);
  Swapstr:= '';
  setlength(buffer,bufsize);
  cmdProcess:= TProcess.Create(nil);
  termCmd(cmdstr,cmdProcess);
  repeat
    setlength(Buffer,bufsize);
    BytesRead:= cmdProcess.StdErr.Read(Buffer[1],length(Buffer)); //this works
    setlength(Buffer,BytesRead);
    Swapstr+= Buffer;
  until BytesRead = 0;
  //writeln('make swap output=',swapstr);
  //mkswap outputs on failure include "error" and "cannot"
  if (pos(ErrorStr1,Swapstr) > 0) or (pos(ErrorStr2,swapstr) > 0) then
     begin
       FoxMsg(rWarningStr,[rSwapErrStr,SwapStr],mtWarning,[mbOK]);
       MakeSwap:= false;
     end;
  cmdProcess.free;
end; //end procedure TPart.MakeSwap

end.

