{ FoxClone
  Copyright (C) 2019 - 2021 2020 Andy Hardwick
  This source is free software; you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free
  Software Foundation; either version 2 of the License, or (at your option)
  any later version.
  This code is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
  details.
  A copy of the GNU General Public License is available on the World Wide Web
  at <http://www.gnu.org/copyleft/gpl.html>. You can also obtain it by writing
  to the Free Software Foundation, Inc., 51 Franklin Street - Fifth Floor,
  Boston, MA 02110-1335, USA.
}

{-------------------------------------------------------------------------------
UNIT STARTUPFORM
This is the initial progress window displayed when starting foxclone - to give the user some
indication that something is happening as reading all drives/partitions can take some time.

It is accessed the OnActivate event = TMainFm.ReadDrives for the main form.

Changes:

V1.0 - 12 Jan 2020
V1.1 - 24 Sep 2020
V1.2 - 27 Mar 2021 foxclone50 version with amendments from original
-------------------------------------------------------------------------------}
unit startupform;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, Forms, Controls, ComCtrls, StdCtrls, Graphics, Dialogs, DateUtils, strutils, GTKLabel,
  il8nstrings, drives, partitions, backup, restore, clone, verify, utilities, fdialog, info;

type
  TUpdateType = (udrive, upart, uOS);

  TStartupFm = class(TForm)
  const
    dtotal = 20; //relative weight to each item in the progress bar where max = 100
    ptotal = 60;
    ostotal = 20;
  private
    UserClose : boolean;
    TotalDrives : byte; //total drives in the system
    TotalParts : byte; //total partitions in the system
    ProgLbl: TLabelGTK;
    ProgBar : TProgressBar;
    DriveLbl, DriveStatusLbl : TLabelGTK;
    PartLbl, PartStatusLbl : TLabelGTK;
    OSLbl, OSstatusLbl : TLabelGTK;
    procedure CloseStartup(Sender: TObject; var CloseAction: TCloseAction);
    procedure UpdateProgress(UType : TUpdateType; index : byte);
  public
    constructor Create(TheOwner: TComponent);  override;
    procedure ReadSystem;
  end;


implementation
{-------------------------------------------------------------------------------
TSTARTUPFM.CONSTRUCTOR
Constructor creating all the objects on the startup form - this is displayed when foxclone starts
and is reading the drives and partitions in the system.

Note - all labels are TLabelGTK, but the status labels showing [x..y] are right aligned with a common
fixed width. Using setCaption would override the width, so caption updates on the status labels are
done through the normal caption:= something.

V1.0 - 03 Mar 2021
V1.1 - 27 Mar 2021, UserClose added
V1.2 - 24 Dec 2021, progress labels on drives and partitions right aligned with the prog bar, not
                    align right did not work.
-------------------------------------------------------------------------------}
constructor TStartupFm.Create(TheOwner: TComponent);
const
  bHeight = 150;  //main window initial height
  hMargin = 20; //horizontal margin for objects
  vMargin = 10; //vertical margin for objects
  wStatus = 50; //width of the status labels
var
  LabelWidth : integer;
begin
  inherited CreateNew(TheOwner,1);
  UserClose:= false;
  BorderStyle:= bsSingle; //don't allow the form to resize
  formstyle:=fsStayonTop; //works
  Font.Size:= 10; //fix for the startup form
  Caption:= rReadSystem; //caption for the title bar
  Position:=poScreenCenter;
  OnClose:= @CloseStartup;
  //get the width of the widest label so that the status labels can align
  LabelWidth:= GetTextWidth(rOpSysStr,font);
  if GetTextWidth(rDriveStr,font) > LabelWidth then LabelWidth:= GetTextWidth(rDriveStr,font);
  if GetTextWidth(rPartitionStr,font) > LabelWidth then LabelWidth:= GetTextWidth(rPartitionStr,font);
  LabelWidth+= hMargin; //make it a bit wider
  //make the form twice as wide as labelwidth
  SetInitialBounds(0,0,LabelWidth * 2,bHeight); //defines initial size
  ProgLbl:= TLabelGTK.Create(Self);
  with ProgLbl do
       begin
         parent:= Self;
         Width:= Self.ClientWidth - 2 * hMargin;
         AnchorSide[akTop].Side:= asrTop;
         AnchorSide[akTop].Control:= Self;
         BorderSpacing.top:= vMargin;
         AnchorSide[akLeft].Side:= asrLeft;
         AnchorSide[akLeft].Control:= Self;
         BorderSpacing.Left:= hMargin;
         setCaption(rReadDrives);
       end;
  ProgBar:= TProgressBar.Create(Self);
  with ProgBar do
       begin
         Parent:= Self;
         Width:= Self.ClientWidth - 2 * hMargin;
         AnchorSide[akTop].Side:= asrBottom;
         AnchorSide[akTop].Control:= ProgLbl;
         BorderSpacing.top:= vMargin;
         AnchorSide[akLeft].Side:= asrLeft;
         AnchorSide[akLeft].Control:= ProgLbl;
         BorderSpacing.Left:= 0;
         Height:= 12;
         BarShowText := false;
         Smooth:= True;
         Min:= 0;
         Max:= 100;
         Position:= 0;
       end;
  DriveLbl:= TLabelGTK.Create(Self);
  with DriveLbl do
       begin
         parent:= Self;
         AnchorSide[akTop].Side:= asrBottom;
         AnchorSide[akTop].Control:= ProgBar;
         BorderSpacing.top:= vMargin;
         AnchorSide[akLeft].Side:= asrLeft;
         AnchorSide[akLeft].Control:= ProgLbl;
         BorderSpacing.Left:= 0;
         setCaption(rDriveStr);
       end;
  DriveStatusLbl:= TLabelGTK.Create(Self);
  with DriveStatusLbl do
       begin
         parent:= Self;
         AnchorSide[akTop].Side:= asrBottom;
         AnchorSide[akTop].Control:= ProgBar;
         BorderSpacing.top:= vMargin;
         AnchorSide[akLeft].Side:= asrLeft;
         AnchorSide[akLeft].Control:= DriveLbl;
         BorderSpacing.Left:= ProgBar.Width - wStatus;
         Width:= wStatus;
         Alignment:= taRightJustify;
         Caption:= '[0..0]'; //cannot use setcaption as it will alter the width
       end;
  PartLbl:= TLabelGTK.Create(Self);
  with PartLbl do
       begin
         parent:= Self;
         AnchorSide[akTop].Side:= asrBottom;
         AnchorSide[akTop].Control:= DriveLbl;
         BorderSpacing.top:= vMargin;
         AnchorSide[akLeft].Side:= asrLeft;
         AnchorSide[akLeft].Control:= ProgLbl;
         BorderSpacing.Left:= 0;
         setCaption(rPartitionStr);
       end;
  PartStatusLbl:= TLabelGTK.Create(Self);
  with PartStatusLbl do
       begin
         parent:= Self;
         AnchorSide[akTop].Side:= asrTop;
         AnchorSide[akTop].Control:= PartLbl;
         BorderSpacing.top:= 0;
         AnchorSide[akLeft].Side:= asrLeft;
         AnchorSide[akLeft].Control:= DriveLbl;
         BorderSpacing.Left:= ProgBar.Width - wStatus;
         Width:= wStatus;
         Alignment:= taRightJustify;
         Caption:= '[0..0]';
       end;
  OSLbl:= TLabelGTK.Create(Self);
  with OSLbl do
       begin
         parent:= Self;
         AnchorSide[akTop].Side:= asrBottom;
         AnchorSide[akTop].Control:= PartLbl;
         BorderSpacing.top:= vMargin;
         AnchorSide[akLeft].Side:= asrLeft;
         AnchorSide[akLeft].Control:= ProgLbl;
         BorderSpacing.Left:= 0;
         setCaption(rOpSysStr);
       end;
  OSstatusLbl:= TLabelGTK.Create(Self);
  with OSstatusLbl do
       begin
         parent:= Self;
         AnchorSide[akTop].Side:= asrTop;
         AnchorSide[akTop].Control:= OSLbl;
         BorderSpacing.top:= 0;
         AnchorSide[akLeft].Side:= asrLeft;
         AnchorSide[akLeft].Control:= OSLbl;
         BorderSpacing.Left:= ProgBar.Width - wStatus;
         Width:= wStatus;
         Alignment:= taRightJustify;
         OSstatusLbl.Caption:= '[?]';
       end;
end; //end constructor TStartupFm.Create
{-------------------------------------------------------------------------------
PROCEDURE TSTARTUPFM.UPDATEPROGRESS
A rewrite of the original procedure updateFM.

Updates the startup form as drives and partitions are read.

Parameters:
    index  - either the number of drives read or partitions read, used to update x in
    [x..total] on the drive and partition status lines.

    UType - used in the case statement to determine if updating drive, partition or OS information.

V1.0 - 29 Dec 2021
-------------------------------------------------------------------------------}
procedure TStartupFm.UpdateProgress(UType : TUpdateType; index : byte);
const
  Template = '[x..y]';
  Complete = '   ✔';
begin
  case UType of
       udrive : with DriveStatusLbl do begin
                     Caption:= replacestr(Template,'x',inttostr(index));
                     Caption:= replacestr(Caption,'y',inttostr(TotalDrives));
                     ProgBar.Position:= ProgBar.Position + 1; // += doesnt work
                     if index = TotalDrives then DriveLbl.setCaption(DriveLbl.Caption + Complete);
                end;
       upart  : with PartStatusLbl do begin
                     Caption:= replacestr(Template,'x',inttostr(index));
                     Caption:= replacestr(Caption,'y',inttostr(TotalParts));
                     ProgBar.Position:= ProgBar.Position + 1;
                     if index = TotalParts then PartLbl.setCaption(PartLbl.Caption + Complete);
                end;
       uOS :    with OSstatusLbl do begin
                     Caption:= replacestr(Caption,'?',inttostr(index));
                     //setCaption(Caption);
                     ProgBar.Position:= ProgBar.Max;
                     OSLbl.setCaption(OSLbl.Caption + Complete);
                     ProgLbl.setCaption(rCompleteStr);
                end;
  end;
  Application.ProcessMessages;
end; //end procedure TStartupFm.UpdateProgress
{-------------------------------------------------------------------------------
PROCEDURE TSTARTUPFM.READSYSTEM
Rewritten to reduce the time taken to read the system. Uses the info unit and SysInfo to read the major
chunks of info - parted and sfdisk output into TStringLists.

The order that info is read is determined by the need to provide visual feedback to the user that
something is happening.

Sequence
   finding drives, creates SysInfo which runs parted (twice - using lm to get machine readable
   output and using --list to get human readable - the only way to find out if a partition is extended or
   logical.

   finding partitions, then runs SysInfo.runSFdisk which runs sfdisk on every drive to get the partition
   table.

   add drives to MyDrives and then run getinfo on drives to populate with info, updating the progress
   form on drives read.

   add partitions and then run getinfo on partitions to populate, updating progress form on parts read.

   find the boot drive.

   run osprober to find any installed OSes on the drives.

   populate box1 on all the tabs.

   short delay before closing the form.

V1.0 - 04 Mar 2021
V1.1 - 26 Mar 2021, restore tab functions added
V1.2 - 01 Jun 2021, clone tab functions added
V1.3 - 09 Dec 2021, running in the VM getting a list index out of bounds. if statement added to the
                    for loop on number of partitions.
V1.4 - 29 Dec 2021, a complete rewrite
-------------------------------------------------------------------------------}
procedure TStartupFm.ReadSystem;
var
  i, j : shortint;
  Finish : TDateTime;
  aList : TStringList; //used for a list of drives and a list of partitions
  PartsRead : byte = 0; //used to display parts counter on the progress form
begin
  if assigned(MyDrives) then FreeAndNil(MyDrives); //free up TSystem and start again
  if assigned(SysInfo) then FreeAndNil(SysInfo); //free up TSysInfo and start again
  ProgBar.Position:= 0;
  ProgBar.Max:= 2; //initial start to show findng drives & parts
  ProgLbl.setCaption(rFindDrives);
  Application.ProcessMessages;
  SysInfo:= TSysInfo.Create; //runs parted on creation
  TotalDrives:= SysInfo.NoDrives;
  UpdateProgress(uDrive,0);
  ProgLbl.setCaption(rFindParts);
  Application.ProcessMessages; //in case a /dev/mapper device has triggered a message
  SysInfo.runSFdisk; //get partition tables with sfdisk
  TotalParts:= SysInfo.NoPartitions;
  UpdateProgress(uPart,0);
  //set the progress bar allocate 25% to os-prober, takes a while to run
  ProgBar.Max:= TotalDrives + TotalParts + round((TotalDrives + TotalParts) * 0.25);
  MyDrives:= TSystem.Create;
  aList:= TStringList.Create; //to hold a list of device names
  SysInfo.GetDrives(aList);   //get the drives, a tstringlist of device names, e.g. sda, sdb, etc.
  ProgLbl.setCaption(rReadDrives);
  Application.ProcessMessages; //in case a /dev/mapper device has triggered a message
  //create a tdrive for each drive and add to the list of drives
  for i:= 0 to aList.Count - 1 do
      begin
        Drive:= TDrive.create;
        //populate device, e.g. sda
        Drive.MountPt:= aList[i];
        //add the part prefix
        if pos('sd',aList[i]) <> 1 then Drive.Prefix:= 'p'; //for nvme & sd cards
        //get the number of partitions on the drive
        Drive.TotalParts:= SysInfo.PartCount(Drive.MountPt);
        //create a list of partitions
        Drive.Partitions:= TPartList.create;
        MyDrives.Add(Drive);
      end;
  aList.Free;
  //now populate the drives with drive info
  for i:= 0 to MyDrives.Count - 1 do
      begin
        MyDrives[i].GetDriveInfo;
        //update the form
        UpdateProgress(uDrive,i+1);
      end;
  //add TParts to the drives
  ProgLbl.setCaption(rReadParts);
  Application.ProcessMessages;
  for i:= 0 to MyDrives.Count - 1 do //cycle through each drive adding partitions
      begin
        aList:= TStringList.Create; //to hold a list of partition names
        SysInfo.GetParts(aList,MyDrives[i].MountPt);   //get a list of parts for the drive
        if aList.Count > 0 then for j:= 0 to aList.Count - 1 do
           begin //add partref = mountpt and prefix, add psector size from the Tdrive
             Partition:= TPart.Create;
             Partition.Drive:= MyDrives[i].MountPt + MyDrives[i].Prefix;
             Partition.Number:= replacestr(aList[j],MyDrives[i].MountPt + MyDrives[i].Prefix,'');
             Partition.SectorSize:= MyDrives[i].DriveSector;
             MyDrives[i].Partitions.Add(Partition);
           end
        else //either a blank drive or a drive as a loop device with a single partition
           begin
              if MyDrives[i].pTable = 'loop' then //need to check if blank or loop,
                 begin
                   MyDrives[i].TotalParts:= 1; //searching parts on the part table will return 0
                   Partition:= TPart.Create;
                   Partition.Drive:= MyDrives[i].MountPt + MyDrives[i].Prefix;
                   Partition.Number:= '0'; //part no = 0 is how foxclone knows it is on a loop device
                   Partition.SectorSize:= MyDrives[i].DriveSector;
                   Partition.Size:= MyDrives[i].DriveSize; //normally from sfdisk dont have one on loop
                   MyDrives[i].Partitions.Add(Partition);
                 end;
           end;
      end;
  //now populate each partition using PartsRead as the counter to display on the form
  for i:= 0 to MyDrives.Count - 1 do
      if MyDrives[i].TotalParts > 0 then for j:= 0 to MyDrives[i].TotalParts - 1 do
         begin
           MyDrives[i].Partitions[j].getPartInfo;
           //update the form
           PartsRead+= 1;
           UpdateProgress(uPart,PartsRead);
         end;
  //find operating systems
  ProgLbl.setCaption(rReadOS);
  Application.ProcessMessages;
  i:= MyDrives.OSprobe;
  UpdateProgress(uOS,i);
  //find the boot drive and set the flag in the relevant TDrive
  MyDrives.FindBoot;
  //add drives to checkbox1 on the backup tab, need to add other tabs as well
  BackTab.PopulateBox1;
  RestoreTab.PopulateBox1;
  CloneTab.PopulateBox1;
  VerifyTab.PopulateBox1;

  UserClose:= true;
  //put a small delay in before the form closes
  Finish:= now;
  repeat until SecondsBetween(now,Finish) > 1;
end; //end TStartupFm.UpdateFm
{-------------------------------------------------------------------------------
PROCEDURE TSTARTUPFM.CLOSESTARTUP
The onclose event handler - if the user clicks on the close window icon on the form heading.

Tried a procedure that would end the startup form and close foxclone but didnt seem to want to work or
fell into the too difficult category. So settled on the user not being able to close the startup form.

Did look at OnCloseQuery but this seems to be written around closing the main form and terminating the
application, e.g. saving files, program state, etc.

A 'global' variable UserClose has been defined in the TStartupFm definition, this is set false by
the constructor and is only set true at the end of TStartupFm.ReadSystem.

Parameters:
    CloseAction can be
         caNone      Do nothing
         caHide      The form is hidden
         caFree      The form is destroyed
         caMinimize  The form is minimized

V1.0 - 27 Mar 2021
V1.1 - 11 May 2021, FoxDialog replaced with FoxMsg
-------------------------------------------------------------------------------}
procedure TStartupFm.CloseStartup(Sender: TObject; var CloseAction: TCloseAction);
begin
  if UserClose then CloseAction:= caFree else
    begin
      FoxMsg(rStartCaptionStr,[rStartMsgStr],mtInformation,[mbOk]);
      CloseAction:= caNone;
    end;
end; //end procedure TStartupFm.CloseStartup

end.

