{ FoxClone
  Copyright (C) 2021 Andy Hardwick
  This source is free software; you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free
  Software Foundation; either version 2 of the License, or (at your option)
  any later version.
  This code is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
  details.
  A copy of the GNU General Public License is available on the World Wide Web
  at <http://www.gnu.org/copyleft/gpl.html>. You can also obtain it by writing
  to the Free Software Foundation, Inc., 51 Franklin Street - Fifth Floor,
  Boston, MA 02110-1335, USA.
}
{-------------------------------------------------------------------------------
UNIT - VERIFY

The objects and methods for verify. The verify panel is a child of the restore panel. No additional
objects are required so the unit defines the methods required for verify - essentially selecting which
backup files should be checked and running partclone verify on them.

V1.0 - 22 Jun 2021
V1.1 - 22 Apr 2024, Box2CK added, a copy of the same procedure in restore. Previously it was using the
                    procedure in restore. Trouble is it then insists on checking that the source
                    drive is present. This is not needed for verify.
-------------------------------------------------------------------------------}
unit verify;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, CheckLst, StdCtrls, Forms, Controls, dialogs, strutils, DateUtils, Graphics,
  restore, basepanel, backup, clone, GTKlabel, drives, partitions, utilities,
  il8nstrings, fdialog, settings;

type
  TVerify = class(TRestore)
    private
      procedure Box1CK(Sender: TObject);
      procedure Box2CK(Sender: TObject);  //new 22 apr 24
      procedure Box3CK(Sender: TObject);
      procedure NextBtnCK(Sender: TObject);
    public
      constructor Create(TheOwner: TComponent);
  end;

var
  VerifyTab : TVerify;

implementation
uses partclone;
{-------------------------------------------------------------------------------
CONSTRUCTOR TVERIFY.CREATE
Creates the verify panel and adds in the properties, functions & procedures specific to clone.

verifytab is an amended copy of restoretab.

Used by:

V1.0 - 22 Jun 2021
-------------------------------------------------------------------------------}
constructor TVerify.Create(TheOwner: TComponent);
begin
  inherited Create(TheOwner);
  Box1.BoxType:= veBox1;
  Box2.BoxType:= veBox2;
  Box3.BoxType:= veBox3;
  TitleLbl.setCaption(rVerifyTitleStr);
  ActionImg.Picture.LoadFromFile(VerifyImgFile);
  //re-using box4 label as a prompt to select backup files to verify
  with Box4Lbl do begin
       setCaption(rVerifyBox4Str);
       AnchorSide[akTop].Side:= asrCenter;
       AnchorSide[akTop].Control := Box3;
       Enabled:= true;
       Visible:= false;
  end;
  freeandnil(LBox4); //not needed in verify
  with NextBtn do begin
       Caption:= rVerifyTabStr;
       AnchorSide[akTop].Side:= asrBottom;
       AnchorSide[akTop].Control := Box3;
  end;
  //event handlers
  Box1.OnClick:= @Box1CK;
  Box2.OnClick:= @Box2CK;    //new 22 apr 24
  Box3.OnClick:= @Box3CK;
  Box3.OnClickCheck:= @Box3CheckCK; //fix for bug in lazarus, not firing onclick on second checkbox click
  NoteBtn.OnClick:= @NoteCK;
  NextBtn.OnClick:= @NextBtnCK;
end; //end constructor TVerify.Create
{-------------------------------------------------------------------------------
PROCEDURE TVERIFY.BOX1CK
Event handler for clicks on box1.

Modified version of TRestore.Box1CK.

Modifications are:
   resets restore tab
   no listbox 4.

Only a single drive may be selected.

On a valid click that selects a drive:
   sets that drive as the source for the backup file
   populates box2 with the partitions on the drive

If the user clicks on blank space - nothing happens.

Note -  clicking on either the check box or the text triggers this procedure, but if clicking on text it
does not change the state of the checkbox, clicking on the checkbox does change its state. Went round
in circles a few times trying to sort out the logic of this. The initial code on saving the checkbox
state and then resetting it works, but not altogether sure how.

Used by:

V1.0 - 20 Jun 2021
-------------------------------------------------------------------------------}
procedure TVerify.Box1CK(Sender: TObject);
var
  Source : shortint;
  dindex : shortint;
begin
  //clear all other tabs
  BackTab.ClearAll;
  RestoreTab.ClearAll;
  CloneTab.ClearAll;
  MyDrives.unSetAll(acNone);
  Source:= Box1.ItemIndex;
  if Source < 0 then exit;
  clearBox(2);       //clear all the other boxes
  clearBox(3);
  ShowDateAndNote(false); //make backup date and note button invisible
  NextBtn.Enabled:= false; //reset the save to file button
  Box1.CheckAll(cbUnchecked); //only one can be selected
  Box2Lbl.Enabled:= true;
  Box1.Checked[Source]:= not Box1.CheckState[Source];
  Box1.CheckState[Source]:= Box1.Checked[Source];
  if Box1.Checked[Source] then
     begin
       dindex:= MyDrives.FindName(Box1.Items[Source]);
       if dindex = -1 then SystemError(12,Box1.Items[Source]); //close foxclone on error
       MyDrives[dindex].setAction:= acSource; //set as the drive containing the backup
       PopulateBox2; //populate box2 with partitions in MyDrives[dindex]
     end;
end; //end procedure TVerify.Box1CK
{-------------------------------------------------------------------------------
PROCEDURE TVERIFY.BOX3CK
The event handler for OnClick for box3.

Multiple partitions may be selected, clicking invokes three checks:

     does the partition exist in the backup set - if not backed up cannot be selected.

     is the partition an extended partition - cannot be selected.

     is the partition a swap partition - cannot be selected.

Used by:

V1.0 - 22 Jun 2021
V1.1 - 13 Oct 2021, const space moved to utilities
V1.2 - 04 Dec 2021, box3ck allows you to select an unknown partition type, fixed.
V1.3 - 22 Apr 2024, changes due to unsupported filesystems like LUKS not trapped in previous code. A
                    bug introduced when V50 was changed to find exFAT parts which are not found
                    with parted.
V1.4 - 26 Apr 2024, moved to checked for supported filesystems and defined that in Utilities. Better
                    to check for supported filesystems than unsupported filesystems.
-------------------------------------------------------------------------------}
procedure TVerify.Box3CK(Sender: TObject);
const
  ExtendStr = 'extended';
  SwapStr = 'swap';
  PartStr = '[partition]';
  FSstr ='[fs]';
var
  Source, i : byte;
  pIndex : shortint;
  Msg1 : string;
begin
  //writeln('box3 clicked on item ',Box3.ItemIndex);
  if Box3.ItemIndex < 0 then exit; //need to add >= box.count
  Source:= Box3.ItemIndex;
  //a checkbox click may have flipped the state so reset (observed in testing)
  Box3.Checked[Source]:= Box3.CheckState[Source];
  with Box3 do //a set of if statements each exiting the procedure
     begin
       //find the partition - it should match Source, but bolt and braces
       pIndex:= RDrive.Partitions.FindName(Items[Source]);
       if pIndex = -1 then SystemError(23,Items[Source]);
       //was the partition backed up?
       if not RDrive.Partitions[pIndex].selected then //the partition was not backed up
          begin
            Checked[Source]:= false;
            CheckState[Source]:= false; //these should be no change
            with RDrive.Partitions[pIndex] do Msg1:= replacestr(rNoVerifyBackupStr,PartStr,Drive + Number);
            FoxMsg(rNoBackupStr,[Msg1 + Space + rNoVerifyStr],mtInformation,[mbOK]);
            exit;
          end;
       //is it an extended partition?
       if pos(ExtendStr,Items[Source]) > 0 then //extended part follows setting of the part table
          begin
            Checked[Source]:= false;
            CheckState[Source]:= false;
            FoxMsg(rExttitle,[rExtMsg + Space + rNoVerifyStr],mtInformation,[mbOK]);
            exit;
          end;
       //is it a swap partition?
       if pos(SwapStr,Items[Source]) > 0 then //extended part follows setting of the part table
          begin
            Checked[Source]:= false;
            CheckState[Source]:= false;
            FoxMsg(rSwapStr,[rSwapMsgStr + Space + rNoVerifyStr],mtInformation,[mbOK]);
            exit;
          end;
       //is it an unknown filesystem
       if pos(RDrive.Partitions[pIndex].FileSystem,SupportedFS) = 0 then
          begin
            Checked[Source]:= false;
            CheckState[Source]:= false;
            with RDrive.Partitions[pIndex] do
               begin
                 //foxmsg changed to include the filesystem not supported
                 Msg1:= replacestr(rUnknownStr,PartStr,Drive + Number);
                 Msg1:= replacestr(Msg1,FSstr,FileSystem);
               end;
            FoxMsg(rUnsupportedStr,[Msg1,rNoVerifyStr],mtInformation,[mbOK]);
            exit;
          end;
       //a standard partition, state can be changed
       Checked[Source]:= not CheckState[Source];
       CheckState[Source]:= Checked[Source];
       //reset the verify button
       NextBtn.Enabled:= false;
       //and set the verify button
       for i:= 0 to Count - 1 do if Checked[i] then
          begin
            NextBtn.Enabled:= true;
            exit;
          end
     end
end; //end procedure TVerify.Box3CK
{-------------------------------------------------------------------------------
PROCEDURE TVERIFY.BOX3CK
The event handler for clicking on the next button, just calls runfoxclone.

Used by:

V1.0 - 22 Jun 2021
-------------------------------------------------------------------------------}
procedure TVerify.NextBtnCK(Sender: TObject);
begin
  runFoxClone(acVerify,RestFname);
end;

{-------------------------------------------------------------------------------
PROCEDURE TVERIFY.BOX2CK
Event handler for clicks on box2. This is a copy modified of the procedure in restore, changed so that
if does not check for the source drive. The verify may be done on another machine.

Does a lot.

Only a single partition may be selected.

Selecting a valid partition opens up the file open dialog to get the backup file.

Opens the backup file and parses it (and the associated sfdisk file).
Populates box3 with the contents of the backup file.
Searches MyDrives for the drive using the model name (and serial number if present) in the backup file.
Sets the drive found in MyDrives as acTarget
Populates box4 (which is a list box not checklist box) with the contents of the target drive.

Used by:

V1.0 - 29 Mar 2021
V1.1 - 05 May 2021  check added to see if the target drive has moved, e.g. sda to sdb
V1.2 - 11 May 2021, FoxDialog replaced with FoxMsg
V1.3 - 28 May 2021, modified to use checkstate
V1.4 - 20 Jun 2021, modified to work with verify, it should be able to make it work with clone from
                    file as well
V1.5 - 08 Sep 2021, nextbtn enabled.
V1.6 - 12 Nov 2021, sets selected=false for all parts on the source drive and then sets selected=true
                    for the part containing the backup. This so that the part can be later unmounted
                    after the restore if foxclone mounted it.
V1.7 - 04 Dec 2021, modification so that the nextbtn is only enabled on restore not verify
V1.8 - 24 Jan 2022, on verify if the source drive has moved, e.g. sda to sdb it raises the warning
                    message that the drive to restore to has changed - not wanted. Cancelling generates
                    a sigsegv. If statement on checking mount point changed to check box type as well.
V1.9 - 26 Jan 2022, Was mounting an unknown filesystem, modified so that it does nothing.
V1.10 - 22 Apr 2024 modified to remove those elements of TRestore.Box2CK that are required for restore,
                    this includes requiring the source drive to be present for the verify.
-------------------------------------------------------------------------------}
procedure TVerify.Box2CK(Sender: TObject);
const
  UnknownStr = 'unknown';
var
  Source : shortint;
  dIndex, pIndex : shortint;
  i : byte;
begin
  writeln('tverify box2 clicked');
  ClearBox(3);
  if assigned(LBox4) then ClearBox(40); //40 = tlist box where 4 = tchecklistbox, if needed for verify
  NextBtn.Enabled:= false; //reset
  ShowDateAndNote(false); //make backup date and note button invisible
  Source:= Box2.ItemIndex;
  if Source < 0 then exit;
  DebugInfo(true,'TVerify.Box2CK',Box2.Items[Source]);
  Box2.CheckAll(cbUnchecked); //only one can be selected
  Box2.Checked[Source]:= not Box2.CheckState[Source];
  Box2.CheckState[Source]:= Box2.Checked[Source];
  if Box2.Checked[Source] then
     begin
       dIndex:= MyDrives.getDrive(acSource); //find the drive containing the backup
       if dIndex = -1 then SystemError(14,'TRestore.Box2CK'); //close foxclone on error
       //get the partition
       pindex:= MyDrives[dindex].Partitions.FindName(Box2.Items[Source]);
       if pindex = -1 then SystemError(15,Box2.Items[Source]);
       //check the filesystem is not unknown, if it is do nothing
       //this should be changed to use unsupportedFS in partitions but what happens if the user
       //has installed foxclone in a system using LVM or LUKS and the partition is mounted in that
       //system?? leave unchanged.
       if MyDrives[dIndex].Partitions[pIndex].FileSystem = UnknownStr then
         begin
           Box2.Checked[Source]:= false;
           Box2.CheckState[Source]:= false;
           exit;
         end;
       //check the partition is not extended, swap or EFI
       if not MyDrives[dIndex].Partitions[pIndex].UserSelect then
         begin
           Box2.Checked[Source]:= false;
           Box2.CheckState[Source]:= false;
           exit;
         end;
       //get and check for restore filename, populates RestFname
       if not GetFile(dIndex,pIndex) then
          begin
            Box2.Checked[Source]:= false;
            Box2.CheckState[Source]:= false;
            exit;
          end;
       //parse the backup file into RDrive
       if not ReadBackupFile then
          begin
            Box2.Checked[Source]:= false;
            Box2.CheckState[Source]:= false;
            exit;
          end;
       //having checked and got the backup, set the partition selected so that restore can find it later
       //this is so that if foxclone mounted the source partition it can be unmounted.
       for i:= 0 to MyDrives[dIndex].TotalParts - 1 do MyDrives[dIndex].Partitions[i].selected:= false;
       MyDrives[dIndex].Partitions[pIndex].selected:= true;
       PopulateBox3;

       ShowDateAndNote(true);
       Box4Lbl.Visible:= true; //verify only using it for something else
      end;
end; //end TVerify.Box2CK


end.

